#include "applications_helper.h"


#define DOMAIN_CROSSING_TRAFFIC "crossing"
#define DOMAIN_ELEVATORS "elevator"
#define DOMAIN_TRIANGLE "triangle"

ConfigReader load_config_reader(std::string config_path)
{
    ConfigReader config_reader;
    config_reader.init(config_path + "rexd_config.cfg", config_path);
    return config_reader;
}

Scenario load_scenario(const ConfigReader& config_reader, const std::string& rules_file)
{
    std::string real_rules_file;
    if (rules_file == "") {
        real_rules_file = config_reader.get_path("ppddl_domain_file");
    }
    else {
        real_rules_file = rules_file;
    }
    return Scenario(real_rules_file, config_reader.get_path("simulator_ppddl_current_problem_file"), config_reader);
}

Scenario load_scenario_ground_truth(const ConfigReader& config_reader)
{
    return load_scenario(config_reader, config_reader.get_path("ppddl_domain_file_ground_truth"));
}

TransitionGroup load_transitions(const ConfigReader& config_reader, Scenario& scenario)
{
    TransitionGroup transitions =  read_transitions_from_file(config_reader.get_path("transitions_backup_path"), 
                                                              scenario.get_state().ppddl_object_manager, 
                                                              scenario.get_state().get_constants(), 
                                                              config_reader.get_variable<bool>("variable_maintains_to_next_state"));
    scenario.domain.index_transitions(transitions);
    return transitions;
}

bool do_rules_have_a_chance_of_change_the_state(const RuleSet& grounded_rules, const State& state)
{
    bool has_a_chance_of_change_the_state = false;
    BOOST_FOREACH(RulePtr rule_ptr, grounded_rules) {
        GroundedRulePtr grounded_rule_ptr = boost::make_shared<GroundedRule>(*static_cast<const GroundedRule*>(&(*rule_ptr)));
        if (grounded_rule_ptr->satisfiesPreconditions(state)) {
            has_a_chance_of_change_the_state = true;
        }
    }
    return has_a_chance_of_change_the_state;
}

TypedAtom get_random_grounding(const TypedAtom& action, const PPDDLObjectManager& ppddl_object_manager)
{
    TypedAtom grounded_action(action);
    BOOST_FOREACH(PPDDL::Object& param, grounded_action.params) {
        std::vector<std::string> possible_objects = ppddl_object_manager.get_type_objects(param.type);
        param.name = *utils::get_random_element<std::vector<std::string>::const_iterator>(possible_objects.begin(),
                                                                                        possible_objects.end());
    }
    return grounded_action;
}

State generate_random_state(const ConfigReader& config_reader, const Scenario& initial_scenario)
{
    PredicateGroup<TypedAtom> constants = initial_scenario.get_state().get_constants();
    PredicateGroup<TypedAtom> non_constants = initial_scenario.get_state().get_non_constants();
    Domain non_constant_domain(config_reader);
    non_constant_domain.index_predicates(non_constants);
    non_constant_domain.index_objects(initial_scenario.get_state().ppddl_object_manager.get_ppddl_objects_per_type());
    
    State full_non_constant_state(non_constant_domain.get_predicate_group_no_closed_world(non_constants, true), 
                                  initial_scenario.get_state().ppddl_object_manager);
    State full_new_state;
    full_new_state.ppddl_object_manager = initial_scenario.get_state().ppddl_object_manager;
    BOOST_FOREACH(const Predicate<TypedAtom>& predicate, full_non_constant_state) {
        if ( (config_reader.get_variable<std::string>("domain") == DOMAIN_CROSSING_TRAFFIC) && 
                (predicate.name == "robot-at") ) // negate all robot-at (only one can be true)
        {
            // hacks for crossing traffic domain, force only one robot-at predicate
            Predicate<TypedAtom> new_predicate(predicate, false);
            full_new_state.insert(new_predicate);
        }
        else if ( (config_reader.get_variable<std::string>("domain") == DOMAIN_ELEVATORS) && 
                (predicate.name == "elevatorAtFloor") ) // negate all elevator-at (only one per elevator can be true)
        {
            // hacks for crossing traffic domain, force only one robot-at predicate
            Predicate<TypedAtom> new_predicate(predicate, false);
            full_new_state.insert(new_predicate);
        }
        else if ( (config_reader.get_variable<std::string>("domain") == DOMAIN_TRIANGLE) && 
                (predicate.name == "vehicleAt") ) // negate all elevator-at (only one per elevator can be true)
        {
            // hacks for crossing traffic domain, force only one robot-at predicate
            Predicate<TypedAtom> new_predicate(predicate, false);
            full_new_state.insert(new_predicate);
        }
        else if (rand() > (RAND_MAX/2)) {
            full_new_state.insert(predicate);
        }
        else {
            full_new_state.insert(predicate.get_negated());
        }
    }
    
    if (config_reader.get_variable<std::string>("domain") == DOMAIN_CROSSING_TRAFFIC) {
        // hacks for crossing traffic domain, force only one robot-at predicate
        PredicateGroup<TypedAtom>::const_iterator singleton_predicate;
        while ( (singleton_predicate = utils::get_random_element<PredicateGroup<TypedAtom>::const_iterator>(full_new_state.begin(), full_new_state.end()))->name != "robot-at");
        full_new_state.update_predicate(singleton_predicate->get_negated());
    }
    else if (config_reader.get_variable<std::string>("domain") == DOMAIN_ELEVATORS) {
        // hacks for elevators domain, only one position per elevator
        PredicateGroup<TypedAtom>::const_iterator singleton_predicate;
        while ( (singleton_predicate = utils::get_random_element<PredicateGroup<TypedAtom>::const_iterator>(full_new_state.begin(), full_new_state.end()))->name != "elevatorAtFloor");
        full_new_state.update_predicate(singleton_predicate->get_negated());
    }
    else if (config_reader.get_variable<std::string>("domain") == DOMAIN_TRIANGLE) {
        // hacks for elevators domain, only one position per elevator
        PredicateGroup<TypedAtom>::const_iterator singleton_predicate;
        while ( (singleton_predicate = utils::get_random_element<PredicateGroup<TypedAtom>::const_iterator>(full_new_state.begin(), full_new_state.end()))->name != "vehicleAt");
        full_new_state.update_predicate(singleton_predicate->get_negated());
    }
    
    full_new_state.insert(constants.begin(), constants.end());
    full_new_state.set_constants(constants);
    return full_new_state;
        
}

TransitionGroup generate_random_transitions(const ConfigReader& config_reader, const Scenario& scenario, const uint number_transitions)
{
    TransitionGroup resulting_transitions(scenario.get_state().ppddl_object_manager);
    uint n_total = scenario.get_non_special_action_list().size() * number_transitions * 2;
    uint n_action = 0;
    std::vector<TypedAtom> action_list = scenario.get_non_special_action_list();
    BOOST_FOREACH(TypedAtom action, action_list) {
        LOG(INFO) << "Generating transitions for action " << action;
        TransitionGroup transitions_change(scenario.get_state().ppddl_object_manager);
        TransitionGroup transitions_no_change(scenario.get_state().ppddl_object_manager);
        
        while ( (transitions_change.size_historial() < number_transitions) || 
                ( (transitions_change.size_historial() + transitions_no_change.size_historial()) < (2*number_transitions) ) ) 
        {
            Scenario new_scenario(scenario);
            State full_new_state = generate_random_state(config_reader, scenario);
            new_scenario.set_state(full_new_state);
            TypedAtom grounded_action = get_random_grounding(action, scenario.get_state().ppddl_object_manager);
            
            // check if the rule has a chance of changing something (i.e. transitions are satisfied)
            RuleSet grounded_action_rules = scenario.get_rules().get_grounded_action_rules(grounded_action, full_new_state);
            bool has_a_chance_of_change_the_state = do_rules_have_a_chance_of_change_the_state(grounded_action_rules, full_new_state);
            
            new_scenario.transition_next_state_simple(grounded_action);
            
            Transition new_transition(full_new_state, grounded_action, new_scenario.get_state(), 0.0, config_reader.get_variable<bool>("variable_maintains_to_next_state"));
            if (has_a_chance_of_change_the_state) {
                    transitions_change.add_transition(new_transition);
            }
            else {
                if (transitions_no_change.size_historial()  < number_transitions) {
                    transitions_no_change.add_transition(new_transition);
                }
            }
            LOG_EVERY_N(100,INFO) << "Completed " << transitions_change.size_historial() << "|" << transitions_no_change.size_historial() << ".";
            utils::print_progress(n_action*(number_transitions*2) + transitions_change.size_historial() + transitions_no_change.size_historial(), n_total);
        }
        resulting_transitions.add_transitions(transitions_change);
        resulting_transitions.add_transitions(transitions_no_change);
        ++n_action;
    }
    
    return resulting_transitions;
}


