#include "apps/applications_helper.h"
#include "lfit/lfit_learner.h"
#include "rexd/pasula_learner.h"


#define LOGA(Level_, Logger_, ...)   \
    while (Log::Enabled(Level_)) {  \
      Log_Write(Level_, Logger_, __func__, __FILE__, __LINE__, __VA_ARGS__);  \
      break;  \
    }

void Log_Write(uint level,
               char const* logger,
               char const* function,
               char const* file,
               int line,
               std::string msg)
{
    std::cout << msg;
}


RuleSetPtr learn(std::string learner_type = "pasula", std::string action_string = "")
{
    struct timeval timer_start = utils::start_timer();
    LOG(INFO) << "Learning";
    ConfigReader config_reader = load_config_reader();
    Scenario scenario = load_scenario(config_reader);
    TransitionGroup transitions = load_transitions(config_reader, scenario);
    
    LearnerPtr learner;
    if (learner_type == "pasula") {
        learner = boost::make_shared<PasulaLearner>(PasulaLearner(config_reader));
    }
    else {
        learner = boost::make_shared<LFITLearner>(LFITLearner(config_reader));
    }
    
    RuleSetPtr rules_ptr;
    if ((config_reader.get_variable<bool>("learning")) && (transitions.size_set() > 0) ) {
        CINFO("planning") << START_COLOR_BLUE << "Learning" << END_COLOR;
        RuleSetPtr removed_rules_ptr, added_rules_ptr;
        if (action_string != "") {
            rules_ptr = RuleSetPtr(new RuleSet);
            TypedAtom action(action_string);
            rules_ptr = learner->learn_action_rules(*rules_ptr, 
                                                    transitions, 
                                                    action, 
                                                    removed_rules_ptr, 
                                                    added_rules_ptr);
        }
        else {
            if (learner_type == "pasula") {
                rules_ptr = learner->learn_quiet(transitions);
            }
            else {
                rules_ptr = learner->learn(transitions);
            }
        }
        CINFO("planning") << "Learned " << rules_ptr->size() << " rules.";
//         CINFO("planning") << "Learned " << *rules_ptr;
        rules_ptr->write_rules_to_file(config_reader.get_path("nid_rules_path"));
        // write PPDDL rules
        Scenario scenario(Scenario(*rules_ptr, 
                                   config_reader.get_path("simulator_ppddl_current_problem_file"), 
                                   config_reader));
        scenario.write_PDDL_domain("learned_rules.ppddl");
        
        LOG(INFO) << "Rules are:";
        LOG(INFO) << *rules_ptr;
    }
    else {
        CINFO("planning") << "Skipping learning (no transitions or no previous action)";
    }
    
    LOG(INFO) << "Time taken: " << utils::get_elapsed_time(timer_start);
    return rules_ptr;
}

RuleSetPtr read_and_parse_lfit_rules(char* path) {
    struct timeval timer_start = utils::start_timer();
    LOG(INFO) << "Reading and parsing LFIT rules";
    ConfigReader config_reader = load_config_reader();
    Scenario scenario = load_scenario(config_reader);
    TransitionGroup transitions = load_transitions(config_reader, scenario);
    
    ReaderLFIT reader_lfit(path);
    RuleSetPtr rules_ptr;
    reader_lfit.read_ruleset(rules_ptr, 
                             transitions, 
                             scenario.get_state().get_constants(), 
                             config_reader);
    
    LOG(INFO) << "Rules are:";
    LOG(INFO) << *rules_ptr;
    LOG(INFO) << "Time taken: " << utils::get_elapsed_time(timer_start);
    
    // write PPDDL rules
    scenario.set_rules(*rules_ptr);
//     RexdScenario rexd_scenario(scenario);
    scenario.write_PDDL_domain("learned_rules.ppddl");
    return rules_ptr;
}

void generate_domain_transitions(const uint number_transitions) {
    struct timeval timer_start = utils::start_timer();
    LOG(INFO) << "Generating transitions for using a learner afterwards";
    ConfigReader config_reader = load_config_reader();

    CINFO("planning") << "Using ground truth rules";
    Scenario scenario(load_scenario_ground_truth(config_reader));
    
    TransitionGroup resulting_transitions = generate_random_transitions(config_reader, scenario, number_transitions);
    
    resulting_transitions.write_pasula_transitions_to_file(config_reader.get_path("transitions_backup_path"));
    LOG(INFO) << "";
    LOG(INFO) << "Time taken: " << utils::get_elapsed_time(timer_start);
}

void evaluate_rules(const std::string& rules_file, const uint number_action_tests)
{
    int verbose = 0;
    LOG(INFO) << "Testing ruleset";
    ConfigReader config_reader = load_config_reader();
    Scenario scenario(load_scenario(config_reader, rules_file));
    Scenario scenario_ground_truth(load_scenario_ground_truth(config_reader));
    
    uint num_tests = 0;
    float total_difference = 0.0;
    std::vector<TypedAtom> action_list = scenario_ground_truth.get_non_special_action_list();
    BOOST_FOREACH(TypedAtom action, action_list) {
        uint n_states_without_changed_tested = 0;
        uint n_states_with_change_tested = 0;
        while ( (n_states_with_change_tested < number_action_tests) || 
                ( (n_states_without_changed_tested + n_states_with_change_tested) < (2*number_action_tests) ) ) 
        {
            Scenario new_scenario_ground_truth(scenario_ground_truth);
            State full_new_state = generate_random_state(config_reader, scenario_ground_truth);
            new_scenario_ground_truth.set_state(full_new_state);
            TypedAtom grounded_action = get_random_grounding(action, scenario_ground_truth.get_state().ppddl_object_manager);
            
            RuleSet grounded_action_rules_ground_truth = scenario_ground_truth.get_rules().get_grounded_action_rules(grounded_action, full_new_state);
            bool has_a_chance_of_change_the_state = do_rules_have_a_chance_of_change_the_state(grounded_action_rules_ground_truth, full_new_state);
            if (has_a_chance_of_change_the_state) {
                n_states_with_change_tested++;
            }
            else {
                n_states_without_changed_tested++;
            }
            
            if (has_a_chance_of_change_the_state || (n_states_without_changed_tested <= number_action_tests)) {
                // obtain changes of ground truth and normal rules
                CLOG_IF(verbose >=1, INFO, "trivial") << "---------------";
                CLOG_IF(verbose >=1, INFO, "trivial") << "Applying action: " << grounded_action;
                bool success; // unused
                float reward; // unused
                new_scenario_ground_truth.transition_next_state(grounded_action, success, reward);
                
                Transition transition(full_new_state, grounded_action, new_scenario_ground_truth.get_state(), 0.0, config_reader.get_variable<bool>("variable_maintains_to_next_state"));
                CLOG_IF(verbose >=3, INFO, "trivial") << transition;
                
                CLOG_IF(verbose >=2, INFO, "trivial") << "Changes are " << transition.get_changes();
                CLOG_IF(verbose >=2, INFO, "trivial") << "Ground truth rules (" << scenario_ground_truth.get_rules().size() << ")";
                Transitions::Coverage ground_truth_coverage = scenario_ground_truth.get_rules().covers_transition_effects(transition, scenario_ground_truth.get_state().ppddl_object_manager);
                CLOG_IF(verbose >=3, INFO, "trivial") << "Coverage: " << ground_truth_coverage;
                CLOG_IF(verbose >=2, INFO, "trivial") << "Tested rules (" << scenario.get_rules().size() << ")";
                Transitions::Coverage coverage = scenario.get_rules().covers_transition_effects(transition, scenario_ground_truth.get_state().ppddl_object_manager);
                CLOG_IF(verbose >=3, INFO, "trivial") << "Coverage: " << coverage;
                CLOG_IF(verbose >=1, INFO, "trivial") << "The difference is (" << ground_truth_coverage.get_coverage(true) << " - " << coverage.get_coverage(true)
                        << ") = " << (ground_truth_coverage.get_coverage(true) - coverage.get_coverage(true));
                total_difference += std::abs(ground_truth_coverage.get_coverage(true) - coverage.get_coverage(true));
                num_tests++;
            }
        }
    }
    LOG(INFO) << "Differences are " << total_difference/num_tests;
    LOG(INFO) << "Num tests are " << num_tests;
}

#define ACTION_NAME_ALL "action_name_all"
void check_rules_score(const std::string& action_name = ACTION_NAME_ALL)
{
    LOG(INFO) << "Checking scores";
    ConfigReader config_reader = load_config_reader();
    Scenario scenario(load_scenario(config_reader));
    TransitionGroup transitions = load_transitions(config_reader, scenario);
    
    lfit::MapHeadRules map_head_rules;
    lfit::RuleWithCoverageSet lfit_rules;
    BOOST_FOREACH(const RulePtr& rule_ptr, scenario.get_rules()) {
        uint n_total = rule_ptr->count_covering_transitions_preconditions(transitions);
        LOG(ERROR) << "Unitialized n_covered!!!!!!";
        uint n_covered = n_total;
        lfit::RuleWithCoverage new_lfit_rule(rule_ptr, lfit::Coverage(n_covered, n_total));
        lfit_rules.insert(new_lfit_rule);
    }
    
    LOG(INFO) << "Calculating log score";
    lfit_rules.get_score(transitions, config_reader, 2);
    
    LOG(INFO) << "Calculating not log score";
    lfit_rules.get_score(transitions, config_reader, 2);
}


void print_help()
{
    std::cout << "Usage: rexd_utils [option] [args]" << std::endl
              << "Options:" << std::endl
              << "\tlearn [learner] [action] \t- learns using current folder files" << std::endl
              << "\tlfit lfit_rules \t- reads and parses lfit rules" << std::endl
              << "\tscore [action]\t- gets score of all rules [or only action rules]" << std::endl
              << "\ttransitions <number_per_rule> \t- creates random transitions for learning" << std::endl
              << "\tevaluate_rules rules_file number_per_rule \t- Compares rule set agains ground truth" << std::endl;
}


int main(int argc, char** argv) {
    // generated random seed
    srand((unsigned)time(0));
    utils::init_loggers("logger.conf");

    if (argc < 2) {
        print_help();
    }
    else {
        std::string input_arg(argv[1]);
        if (input_arg == "learn") {
            if (argc == 3) {
                learn(argv[2]);
            }
            else if (argc > 3) {
                learn(argv[2],argv[3]);
            }
            else {
                learn();
            }
        }
        else if (input_arg == "lfit") {
            if (argc == 3) {
                read_and_parse_lfit_rules(argv[2]);
            }
            else {
                print_help();
            }
        }
        else if (input_arg == "score") {
            if (argc == 2) {
                check_rules_score();
            }
            else if (argc == 3) {
                check_rules_score(argv[2]);
            }
            else {
                print_help();
            }
        }
        else if (input_arg == "transitions") {
            if (argc == 3) {
                generate_domain_transitions(boost::lexical_cast<uint>(argv[2]));
            }
            else {
                print_help();
            }
        }
        else if (input_arg == "evaluate_rules") {
            if (argc == 4) {
                evaluate_rules(argv[2], boost::lexical_cast<uint>(argv[3]));
            }
            else {
                print_help();
            }
        }
        else {
            print_help();
        }
    }

    return 0;
}
