#include "ippc_planner.h"
#include "boost/filesystem/operations.hpp"

IppcPlanner::IppcPlanner(const ConfigReader& config_reader)
{
    noop_actions_allowed = config_reader.get_variable<bool>("noop_actions_allowed");
    
    // check existance of binary files
    planner_path = config_reader.get_binary_path("planner_path");
    if ( !boost::filesystem::exists( planner_path ) ) {
        CERROR("planning") << "Planner executable " << planner_path << " does not exist!";
    }
    
    // port
    planner_port = config_reader.get_variable<uint>("planner_port");
    if (planner_port == 0) { // auto selection
        planner_port = utils::get_socket_from_current_dir_hash(1);
    }
}


int IppcPlanner::estimate_distance_to_finish(const RexdScenario &rexd_scenario)
{
    std::vector< TypedAtom > planning_res;
    if (!plan_n_actions(rexd_scenario, planning_res, false)) {
        return 0;
    }
    std::cerr << "IppcProstPlanner::estimate_distance_to_finish: " << "Estimation: " << planning_res.size() << std::endl;
    return planning_res.size();
}


bool IppcPlanner::plan_n_actions(const RexdScenario& rexd_scenario, std::vector<TypedAtom>& plan_result, const bool thoroughly, const int verbose, const uint n_actions, const bool force_effect) const
{
    RexdScenario updated_scenario(rexd_scenario);
    TypedAtom::Opt current_action_opt;
    bool reached_goal = false;

    if (rexd_scenario.is_in_goal()) {
        if (verbose)
            CINFO("planning") << "Is goal. Not planning.";
        return true;
    }

    uint max_iterations = std::max(n_actions, rexd_scenario.get_planning_horizon());
    uint iteration = 0;
    bool finished_planning = false;
    while ( (iteration < max_iterations) && !finished_planning ) {
        if (verbose) {
            CINFO("planning") << "Plan action " << iteration << " with state: " << Predicates::transform_to_nontyped(updated_scenario.get_state());
            CINFO("planning") << "     and objects: " << updated_scenario.get_state().ppddl_object_manager;
        }
        else
            std::cout << "." << std::flush;
        
        // plan
        float current_reward;
        current_action_opt = this->plan(updated_scenario, current_reward, thoroughly, verbose);
        
        // apply current action
        if (current_action_opt) {
            plan_result.push_back(*current_action_opt);
            if (current_action_opt->is_special()) {
                if (*current_action_opt == MAKE_TYPED_ATOM_TEACHER) {
                    if (verbose) {
                        CINFO("planning") << "Failed: Teacher action planned.";
                    }
                    finished_planning = true;
                }
            }
            bool success; // not used
            float reward; // not used
            updated_scenario.transition_next_state(*current_action_opt, success, reward, force_effect);
            updated_scenario.decrease_planning_horizon();
        }
        else { // no solution -> finished
            finished_planning = true;
            reached_goal = true;
            if (verbose) {
                CINFO("planning") << "No more solutions.";
            }
        }

        // check goal
        if (updated_scenario.is_in_goal()) {
            finished_planning = true;
            reached_goal = true;
            if (verbose)
                CINFO("planning") << "Reached a goal!";
        }
        
        if (verbose >= 1) {
            LOG(INFO) << "Current reward is " << updated_scenario.get_reward();
        }
    }
    return reached_goal;
}


TypedAtom::Opt IppcPlanner::plan(const RexdScenario& rexd_scenario, float& planning_result, const bool thoroughly, const int verbose) const
{
    TypedAtom::Opt resulting_plan;

    if (verbose) {
        if (thoroughly)
            CINFO("planning") << "Planning full... " << NOENDL;
        else
            CINFO("planning") << "Planning fast... " << NOENDL;
    }

    std::streambuf *cout_buf;
    if (!verbose) {
        cout_buf = stream_quiet(std::cout);
    }
    resulting_plan = ippcPlan(rexd_scenario, thoroughly, verbose, planning_result);
    if (!verbose) {
        stream_restore(std::cout, cout_buf);
    }

    if (verbose) {
        if (resulting_plan) {
            CINFO("planning") << " Action: " << START_COLOR_BLUE << *resulting_plan << END_COLOR << " (" << planning_result << ")";
        }
        else {
            CINFO("planning") << " No plan";
        }
    }
    return resulting_plan;
}


TypedAtom::Opt IppcPlanner::ippcPlan(const RexdScenario& rexd_scenario, const bool thoroughly, const int verbose, float& planning_result) const
{
    write_problem_files(rexd_scenario);
    
    int pid = fork();
    if ( pid == -1 ) {
        perror("Cannot proceed. fork() error");
        exit(-1);
    }
    if (pid == 0)
    {
        // child process
        utils::initialize_random_seed(); // regenerate rand in the child process
        execute_planner_command(thoroughly, verbose);
        exit(0);
    }

    XMLServer_t *ippcserver;
    ippcserver = new XMLServer_t();
    ippcserver->start_session(planner_port);
    ippcserver->start_round();

    NonTypedAtom non_typed_action = ippcserver->get_action(rexd_scenario.get_state(), planning_result);
    TypedAtom action = rexd_scenario.get_state().ppddl_object_manager.get_typed_atom(non_typed_action);
    if ((!noop_actions_allowed) && (action == MAKE_TYPED_ATOM_NOOP_ACTION)) {
        action = TypedAtom("");
    }

    ippcserver->end_round();
    ippcserver->end_session();
    
    int status;
    wait(&status);
    delete ippcserver;

    if (utils::trim(action.name) == "") {
        planning_result = -100000.0;
        return TypedAtom::Opt();
    }

    return TypedAtom::Opt(action);
}

