/*
    <one line to give the program's name and a brief idea of what it does.>
    Copyright (C) 2013  <copyright holder> <email>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "auto_teacher.h"


AutoTeacher::AutoTeacher(const ConfigReader& config_reader_):
    stateless_scenario(config_reader_.get_path("ppddl_domain_file_ground_truth"), 
                       config_reader_.get_path("simulator_ppddl_current_problem_file"), 
                       config_reader_),
    complete_retries(0)
{
    // create planner
    planner = PlannerFactory::get_auto_teacher_planner(config_reader_);
}

float AutoTeacher::get_expected_action_success_probability(const State& state, const TypedAtom& action) const
{
    Scenario scenario(stateless_scenario);
    scenario.set_state(state);
    return planner->get_expected_action_success_probability(scenario, action, false);
}

TypedAtom::Opt AutoTeacher::get_action_simple(const State& state) const
{
    Scenario scenario(stateless_scenario);
    scenario.set_state(state);
    return planner->plan(scenario, true, false);
}

TypedAtom::Opt AutoTeacher::get_action(const State& state,
                                         const std::vector<TypedAtom>& action_list,
                                         const std::vector<std::string> actions_requiring_excuse)
{
    TypedAtom::Opt action_opt = TypedAtom::Opt();
    std::string action_string;
    CINFO("teacher") << "Auto Teacher!";

    uint auto_teacher_retries = 0;
    while (!action_opt) {
        action_opt = get_planned_action(state, action_list, actions_requiring_excuse);
        
        if ((!action_opt)) {
            if (stateless_scenario.config_reader.get_variable<bool>("auto_teacher_infinite") ) {
                infinite_autoteacher(action_opt, auto_teacher_retries);
            }
            else {
                action_opt = TypedAtom::Opt (MAKE_TYPED_ATOM_HELP); // read_new_action();
            }
        }
    }
    return action_opt;
}


TypedAtom::Opt AutoTeacher::get_planned_action(const State& state,
                                     const std::vector<TypedAtom>& action_list,
                                     const std::vector<std::string> actions_requiring_excuse)
{
    TypedAtom::Opt planned_action;
    Scenario scenario(stateless_scenario);
    scenario.set_state(state);
    RexdScenario rexd_scenario(scenario);
    rexd_scenario.set_vmin();

    TypedAtom::Opt teacher_plan = planner->plan(rexd_scenario, true, false);
    if (teacher_plan && (*teacher_plan != MAKE_TYPED_ATOM_TEACHER)) {
        TypedAtom action = *teacher_plan;
        planned_action = TypedAtom::Opt(action);
        TypedAtom symbolic_action = TypedAtom(action);
        symbolic_action.make_symbolic();

        // always add if it is a new action
        if (stateless_scenario.config_reader.get_variable<bool>("auto_teacher_forced")) {
            CINFO("teacher") << "Auto teacher: " << action << " - forced mode.";
        }
        else if (std::find(action_list.begin(), action_list.end(), symbolic_action)!=action_list.end()) {
            // not a new action
            // add action only if it is in actions requiring excuse
            if (std::find(actions_requiring_excuse.begin(), actions_requiring_excuse.end(), planned_action->name)==actions_requiring_excuse.end()) {
                // action not found in actions_requiring_excuse
                CINFO("teacher") << "Auto teacher got action: " << action << " but is not recommended by excuses. Requesting human teacher help.";
                CINFO("teacher") << "Waiting for new action: ";
                planned_action = TypedAtom::Opt();
            }
            else if (stateless_scenario.config_reader.get_variable<bool>("auto_teacher_excuse_autoexecute")) {
                CINFO("teacher") << "Action " << action << " matches excuses recommendation, executing it.";
            }
            else {
                CINFO("teacher") << "Action " << action << " matches excuses recommendation.";
                CINFO("teacher") << "Waiting for new action: ";
                planned_action = TypedAtom::Opt();
            }
        }
        else {
            CINFO("teacher") << "Action " << action << " is a new action. Selecting it as this unknown action seems neccesary for this task.";
        }
    }
    else { // no plan or teacher action obtained (which means reward < Vmin -> no plan)
        CINFO("teacher") << "Auto teacher failed (dead end?)";
        if (stateless_scenario.config_reader.get_variable<bool>("auto_teacher_autodead")) {
            planned_action = TypedAtom::Opt(MAKE_TYPED_ATOM_DEAD);
        }
        else {
            CINFO("teacher") << "Waiting for new action: ";
            planned_action = TypedAtom::Opt();
        }
    }
    return planned_action;
}


void AutoTeacher::infinite_autoteacher(TypedAtom::Opt& action_opt, uint& auto_teacher_retries)
{
    auto_teacher_retries ++;
    if (auto_teacher_retries > stateless_scenario.config_reader.get_variable<uint>("auto_teacher_infinite_retries")) {
        complete_retries++;
        if (complete_retries < stateless_scenario.config_reader.get_variable<uint>("auto_teacher_infinite_full_retries")) {
            action_opt = TypedAtom::Opt(MAKE_TYPED_ATOM_RETRY);
        }
        else {
            complete_retries = 0;
            action_opt = TypedAtom::Opt(MAKE_TYPED_ATOM_HELP);
        }
    }
}


