#include "failure_analyzer_tools.h"

#define DEBUG_EXCUSE_RULE_ALT 0


std::vector<RuleSet> FailureAnalyzerUtils::get_rules_that_overcome_planning_failure(const RuleSet rules, const PlanningFailureOpt& planning_failure_opt, const TransitionGroup& transitions) const
{
    if (!planning_failure_opt) {
        return std::vector<RuleSet>();
    }
    
    int DEBUG = DEBUG_EXCUSE_RULE_ALT;
    std::vector<RuleSet> extra_rulesets;

    double orig_score = rules.score(transitions, scenario.get_state().ppddl_object_manager);

    RuleSet rules_requiring_planning_failure;

    PlanningFailure symbolic_planning_failure(*planning_failure_opt);
    symbolic_planning_failure.make_symbolic();
    rules_requiring_planning_failure = rules.find_rules_with_a_precondition(symbolic_planning_failure);
    
    // TODO change/remove precondition
    CINFO("excuses") << "Changing " << rules_requiring_planning_failure.size() << " rules.";
    for ( RuleList::iterator rule_it=rules_requiring_planning_failure.begin() ; rule_it != rules_requiring_planning_failure.end(); ++rule_it ) {
        RuleSet rules_copy(rules);
        rules_copy.remove_rule(*rule_it);

        RuleSet new_rules = create_new_rule_candidates_with_planning_failure(*rule_it, symbolic_planning_failure, transitions);
        CINFO("excuses") << "Checking " << new_rules.size() << " new rules for action " << (*rule_it)->name << " with score ( " << orig_score << " ): " << NOENDL;
        if (DEBUG > 1)
            CINFO("excuses") << "Original rule was " << **rule_it;
        for ( RuleList::iterator new_rule_it=new_rules.begin() ; new_rule_it != new_rules.end(); ++new_rule_it ) {
            RuleSet new_generated_rules(rules_copy);
            new_generated_rules.push_back(*new_rule_it);
            double new_score = new_generated_rules.score(transitions, scenario.get_state().ppddl_object_manager);
            if (new_score != (-1. * std::numeric_limits<double>::max()) ) {
                if (DEBUG > 1)
                    CINFO("excuses") << "Got new rule with score " << new_score << " (orig score was " << orig_score << ")";
                if (DEBUG > 2)
                    CINFO("excuses") << **new_rule_it;
                if (new_score >= orig_score) {
                    CINFO("excuses") << "*" << NOENDL;
                    if (DEBUG > 0)
                        CINFO("excuses") << "Got new rule \"" << (*new_rule_it)->name
                                         << "\" with score " << new_score << " (orig score was " << orig_score << ")";
                    if (DEBUG > 1)
                        CINFO("excuses") << **new_rule_it;
                    extra_rulesets.push_back(new_generated_rules);
                }
                else {
                    CINFO("excuses") << "." << NOENDL;
                }
            }
        }
        CINFO("excuses");
    }

    return extra_rulesets;
}


RuleSet FailureAnalyzerUtils::create_new_rule_candidates_with_planning_failure(const RulePtr rule, const PredicateGroup<TypedAtom> symbolic_planning_failures, const TransitionGroup& transitions) const
{
    int DEBUG = DEBUG_EXCUSE_RULE_ALT;
    RuleSet result;
    if (DEBUG > 0)
        CINFO("excuses") << "Rule candidates for rule " << rule->name << " with symbolic planning_failures " << symbolic_planning_failures;
    
    if (DEBUG > 3)
        CINFO("excuses") << "Obtaining possible preconditions...";
    PredicateGroup<TypedAtom> new_preconditions;
    for ( TransitionSet::const_iterator transition_it = transitions.begin_set() ; transition_it != transitions.end_set(); ++transition_it ) {
        if (rule->covers_transition_preconditions(*transition_it, transitions.ppddl_object_manager) > 0.0) {
            new_preconditions.add_predicates(transition_it->get_prev_state());
        }
    }
    new_preconditions.erase(symbolic_planning_failures);
    if (DEBUG > 2)
        CINFO("excuses") << "Final possible preconditions are " << new_preconditions;

    if (DEBUG > 3)
        CINFO("excuses") << "Generating rules with new preconditions...";

    // TODO check if passing typed params is good!
    // WARNING PROBABLY SHOULD DO FILTERING BASED ON TYPES!
    PredicateGroup<TypedAtom> new_preconditions_symbolic = new_preconditions.get_symbolic_predicates_permutations(rule->params);

    if (DEBUG > 3)
        CINFO("excuses") << "Adding symbolic predicates " << new_preconditions_symbolic;
    for ( PredicateGroup<TypedAtom>::iterator sym_pred_it=new_preconditions_symbolic.begin() ; sym_pred_it != new_preconditions_symbolic.end(); ++sym_pred_it) {
        SymbolicRulePtr new_rule = boost::make_shared<SymbolicRule>(SymbolicRule(*rule));
        PredicateGroup<TypedAtom> new_rule_preconditions = new_rule->get_preconditions();
        new_rule_preconditions.erase(symbolic_planning_failures);
        if (!new_rule_preconditions.has_predicate(*sym_pred_it)) {
            if (DEBUG > 4)
                CINFO("excuses") << "Adding new rule candidate with precondition " << *sym_pred_it;
            new_rule_preconditions.insert(*sym_pred_it);
            new_rule->set_preconditions(new_rule_preconditions);
            result.push_back(new_rule);
        }
    }

    return result;
}

