/*
 * LFIT conflict solver
 * Copyright (C) 2015  David Martínez <dmartinez@iri.upc.edu>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#define PERFORMANCE_HACK_NUMBER_BEFORE_SWITCH_TO_GREEDY 10000

#include "lfit_conflict_solver.h"

using namespace lfit;

ConflictArrayCoveringGeneral::ConflictArrayCoveringGeneral(const RuleWithCoverageSet& ruleset)
{
    conflict_array.resize(ruleset.size());
    BOOST_FOREACH(std::vector<bool>& row, conflict_array) {
        row.resize(ruleset.size(), true); // default is all conflict (this value is only maintained for the diagonal in the matrix)
    }
    uint rule_idx1 = 0;
    for (RuleWithCoverageSet::const_iterator rule_it1 = ruleset.begin(); rule_it1 != ruleset.end(); ++rule_it1) {
        conflict_array[rule_idx1][rule_idx1] = true;
        RuleWithCoverageSet::const_iterator rule_it2 = rule_it1;
        uint rule_idx2 = rule_idx1 + 1;
        for (++rule_it2; rule_it2 != ruleset.end(); ++rule_it2) {
            
            if ( (static_cast<const TypedAtom>(*rule_it1->rule_ptr) == *rule_it2->rule_ptr) && 
                 (rule_it1->rule_ptr->get_preconditions().get_logical_intersects(rule_it2->rule_ptr->get_preconditions()))
               ) {
                conflict_array[rule_idx1][rule_idx2] = true;
                conflict_array[rule_idx2][rule_idx1] = true;
            }
            else {
                conflict_array[rule_idx1][rule_idx2] = false;
                conflict_array[rule_idx2][rule_idx1] = false;
            }
            ++rule_idx2;
        }
        ++rule_idx1;
    }
}

bool ConflictArrayCoveringGeneral::is_there_a_conflict(const int rule_idx1, const int rule_idx2)
{
    return conflict_array[rule_idx1][rule_idx2];
}

ConflictArrayCoveringTransitions::ConflictArrayCoveringTransitions(const RuleWithCoverageSet& ruleset, const TransitionGroup& transitions)
{
    coverage_table.resize(ruleset.size());
    size_t rule_idx = 0;
    BOOST_FOREACH(RuleWithCoverage rule_with_cov, ruleset) {
        std::vector< Transitions::Coverage > coverages = rule_with_cov.get_covered_transitions_full(transitions);
        BOOST_FOREACH(const Transitions::Coverage& coverage, coverages) {
            for (size_t change_idx = 0; change_idx < coverage.size(); ++change_idx) {
                coverage_table[rule_idx].push_back(coverage.get_change_coverage(change_idx));
            }
        }
        ++rule_idx;
    }
}

bool ConflictArrayCoveringTransitions::is_there_a_conflict(const int rule_idx1, const int rule_idx2)
{
    for (size_t transition_change_idx = 0; transition_change_idx < coverage_table[0].size(); ++transition_change_idx) {
        if ( (coverage_table[rule_idx1][transition_change_idx] > 0.0) &&
             (coverage_table[rule_idx2][transition_change_idx] > 0.0) ) 
        {
            return true;
        }
    }
    return false;
}


RuleConflictSolver::RuleConflictSolver(const RuleWithCoverageSet& ruleset, const TransitionGroup& transitions, const ConfigReader& config_reader):
    ruleset(ruleset)
{
    if (config_reader.get_variable<bool>("lfit_learner_conflicts_avoid_all")) {
        generate_conflict_array_full();
    }
    else {
        generate_conflict_array_transitions(transitions);
    }
}

void RuleConflictSolver::generate_conflict_array_full()
{
    conflict_array_ptr = ConflictArrayPtr(new ConflictArrayCoveringGeneral(ruleset));
}

void RuleConflictSolver::generate_conflict_array_transitions(const TransitionGroup& transitions)
{
    conflict_array_ptr = ConflictArrayPtr(new ConflictArrayCoveringTransitions(ruleset, transitions));
}

RuleWithCoverageSet RuleConflictSolver::transform_indexed_ruleset_to_standard(const IndexedRuleSet& indexed_ruleset) const
{
    RuleWithCoverageSet standard_ruleset;
    BOOST_FOREACH(const int rule_idx, indexed_ruleset) {
        RuleWithCoverageSet::iterator rule_it = ruleset.begin();
        std::advance(rule_it, rule_idx);
        standard_ruleset.insert(*rule_it);
    }
    return standard_ruleset;
}


RuleWithCoverageSet RuleConflictSolver::get_best_subset_non_conflicting_heuristic(const TransitionGroup& transitions, const ConfigReader& config_reader, const int verbose)
{
    float lfit_score_optimistic_value = config_reader.get_variable<float>("lfit_learner_score_optimistic_value");
    uint max_number_rules_per_iter = config_reader.get_variable<uint>("lfit_learner_conflicts_heuristic_max_rules_per_iter");
    uint max_number_iterations = config_reader.get_variable<uint>("lfit_learner_conflicts_heuristic_max_iterations");
    
    IndexedRuleSet best_ruleset;
    float best_score = -1 * std::numeric_limits<float>::max();
    std::set<ScoreAndIndexedRuleSet> possibly_valid_indexed_results;
//     std::priority_queue<ScoreAndIndexedRuleSet> possibly_valid_indexed_results;
    
    for (int rule_idx = 0; rule_idx < (int)ruleset.size(); ++rule_idx) {
        IndexedRuleSet new_indexed_candidate;
        new_indexed_candidate.insert(rule_idx);
        RuleWithCoverageSet new_ruleset_candidate = transform_indexed_ruleset_to_standard(new_indexed_candidate);
        float new_score = new_ruleset_candidate.get_score(transitions, config_reader, 0);
        if (new_score > best_score) {
            best_ruleset = new_indexed_candidate;
            best_score = new_score;
            if (verbose>=2) {
                CLOG_IF(verbose>=2, INFO, "trivial") << "---------";
                CLOG_IF(verbose>=2, INFO, "trivial") << "New best score with";
                new_ruleset_candidate.get_score(transitions, config_reader, verbose);
            }
        }
                        
        float optimistic_score = new_ruleset_candidate.get_score(transitions, config_reader, 0, lfit_score_optimistic_value);
        /* Note: only makes sense with lfit_score_optimistic_value=1.0
         * At least a rule will have to be added.
         * Supposing that the rule to be added score is as big as all the current, this rule will have a penalty >= than the biggest rule in current set, otherwise that rule will be parsed before in this loop as it would have a better optimistic score.
        */
        if (lfit_score_optimistic_value == 1.0) {
            optimistic_score *= 2; 
        }
        
        if ( optimistic_score > best_score) {
            possibly_valid_indexed_results.insert(ScoreAndIndexedRuleSet(optimistic_score, new_indexed_candidate));
            if (verbose>=4) {
                CLOG_IF(verbose>=2, INFO, "trivial") << "---------";
                CLOG_IF(verbose>=2, INFO, "trivial") << "Optimistic score with ";
                new_ruleset_candidate.get_score(transitions, config_reader, verbose, lfit_score_optimistic_value);
            }
        }
    }
    
    std::set<IndexedRuleSet> already_added_sets;
    
    uint number_iterations_without_change = 0;
    uint number_rules_analyzed = 0;
    uint number_iterations = 0;
    while ( !possibly_valid_indexed_results.empty() && 
//             (number_rules_analyzed < max_number_rules) && 
            (number_iterations_without_change < (max_number_iterations/4)) &&
            (number_iterations < max_number_iterations) ) {
//         ScoreAndIndexedRuleSet current_score_indexed_result = possibly_valid_indexed_results.top();
//         possibly_valid_indexed_results.pop();
        ScoreAndIndexedRuleSet current_score_indexed_result = *possibly_valid_indexed_results.rbegin(); // last element = biggest
        size_t number_erased = possibly_valid_indexed_results.erase(current_score_indexed_result);
        assert(number_erased==1);
        if ( (verbose >= 3) && (current_score_indexed_result.first > -0.4) ) {
            CLOG_IF(verbose>=3, INFO, "trivial") << "************************";
            CLOG_IF(verbose>=3, INFO, "trivial") << "Optimistic score was " << current_score_indexed_result.first;
            RuleWithCoverageSet new_ruleset_candidate = transform_indexed_ruleset_to_standard(current_score_indexed_result.second);
            new_ruleset_candidate.get_score(transitions, config_reader, verbose, lfit_score_optimistic_value);
        }
        
        if (current_score_indexed_result.first < best_score) {
            possibly_valid_indexed_results.clear();
        }
        
        int number_rules_iteration = 0;
        // for each new rule candidate
        for (std::set<ScoreAndIndexedRuleSet>::reverse_iterator set_it = possibly_valid_indexed_results.rbegin(); set_it != possibly_valid_indexed_results.rend(); ++ set_it) {
            if (set_it->first < best_score) {
                std::set<ScoreAndIndexedRuleSet>::reverse_iterator current = set_it++;
                possibly_valid_indexed_results.erase(*current);
            }
            else {
                bool conflicting = false;
                for (int rule_idx = 0; rule_idx < (int)ruleset.size(); ++rule_idx) {
                    if (current_score_indexed_result.second.count(rule_idx) && set_it->second.count(rule_idx)) {
                        conflicting = true;
                    }
                }
                for (int rule_idx = 0; rule_idx < (int)ruleset.size(); ++rule_idx) {
                    for (int rule_idx2 = 0; rule_idx < (int)ruleset.size(); ++rule_idx) {
                        if (current_score_indexed_result.second.count(rule_idx)) {
                            if (set_it->second.count(rule_idx)) {
                                if (conflict_array_ptr->is_there_a_conflict(rule_idx, rule_idx2)) {
                                    conflicting = true;
                                }
                            }
                        }
                    }
                }
                if (!conflicting) {
                    number_rules_analyzed++;
                    number_rules_iteration++;
                    IndexedRuleSet new_indexed_candidate(current_score_indexed_result.second);
                    new_indexed_candidate.insert(set_it->second.begin(), set_it->second.end());
                    if (already_added_sets.count(new_indexed_candidate) == 0) {
                        RuleWithCoverageSet new_ruleset_candidate = transform_indexed_ruleset_to_standard(new_indexed_candidate);
                        float new_score = new_ruleset_candidate.get_score(transitions, config_reader, 0);
                        if (new_score > best_score) {
                            best_ruleset = new_indexed_candidate;
                            best_score = new_score;
                            number_iterations_without_change = 0;
                            if (verbose>=2) {
                                CLOG_IF(verbose>=2, INFO, "trivial") << "---------";
                                CLOG_IF(verbose>=2, INFO, "trivial") << "New best score with";
                                new_ruleset_candidate.get_score(transitions, config_reader, verbose);
                            }
                        }
                        
                        float optimistic_score = new_ruleset_candidate.get_score(transitions, config_reader, 0, lfit_score_optimistic_value);
                        // TODO
                        
                        already_added_sets.insert(new_indexed_candidate);
                        if ( optimistic_score > best_score) {
                            possibly_valid_indexed_results.insert(ScoreAndIndexedRuleSet(optimistic_score, new_indexed_candidate));
                        }
                    }
                }
            }
            if (number_rules_iteration > (int)max_number_rules_per_iter) { // HACK Use only the top candidates
                break;
            }
        }
//         }
        number_iterations_without_change++;
        LOG_EVERY_N(99,INFO) << "Current size: " << possibly_valid_indexed_results.size();
        number_iterations++;
    }
    
    if (verbose>=1) {
        CLOG_IF(verbose>=1, INFO, "trivial") << "###############";
        CLOG_IF(verbose>=2, INFO, "trivial") << "Best score with";
        transform_indexed_ruleset_to_standard(best_ruleset).get_score(transitions, config_reader, 2);
        
        CLOG_IF(verbose>=1, INFO, "trivial") << "###############";
        if (possibly_valid_indexed_results.empty()) {
            CLOG_IF(verbose>=1, INFO, "trivial") << "Reason to finish: " << "Completed checking all candidates.";
        }
//         else if (number_rules_analyzed >= max_number_rules) {
//             CLOG_IF(verbose>=1, INFO, "trivial") << "Reason to finish: " << "Max number of rules analyzed reached (" << max_number_rules << ").";
//         }
        else if (number_iterations_without_change >= (max_number_iterations/4)) {
            CLOG_IF(verbose>=1, INFO, "trivial") << "Reason to finish: " << "Number of iterations without change >= max/4=" << (max_number_iterations/4) << ".";
        }
        else if (number_iterations >= max_number_iterations) {
            CLOG_IF(verbose>=1, INFO, "trivial") << "Reason to finish: " << "Number of iterations >= max=" << (max_number_iterations) << ".";
        }
        else {
            CLOG_IF(verbose>=1, WARNING, "trivial") << "Reason to finish: " << "WTF!";
        }
    }
    return transform_indexed_ruleset_to_standard(best_ruleset);
}

