#include "lfit_learner.h"



LFITLearner::LFITLearner(const ConfigReader& config_reader) :
    config_reader(config_reader)
{
    lfit_rules_path = config_reader.get_path("lfit_rules_path");
    transitions_learning_path = config_reader.get_path("transitions_tmp_path");
    learner_path = config_reader.get_binary_path("learner_path");
    max_action_variables = config_reader.get_variable<uint>("lfit_learner_max_action_variables");
    maximum_number_preconditions = config_reader.get_variable<uint>("lfit_learner_max_preconditions");
}


RuleSetPtr LFITLearner::learn_quiet(const TransitionGroup& transitions)
{
    CINFO("planning") << "Learning... " << NOENDL;

    std::streambuf* cout_sbuf = std::cout.rdbuf(); // save original sbuf
    std::ofstream   fout("/dev/null");
    std::cout.rdbuf(fout.rdbuf()); // redirect 'cout' to a 'fout'
    RuleSetPtr result = learn(transitions);
    std::cout.rdbuf(cout_sbuf); // restore the original stream buffer

    CINFO("planning") << "learned  " << result->size() << " rules.";

    return result;
}


RuleSetPtr LFITLearner::learn(const TransitionGroup& transitions)
{
    CINFO("planning") << "LFITLearner: " << "LFIT learner started";
    
    // normal domain (only needed to know the constants)
    // TODO get the constants in a more efficient way
    Domain domain(config_reader);
    domain.index_transitions(transitions);
    
    TransitionGroup symbolic_transitions = transitions.get_symbolic_transitions(max_action_variables);
    symbolic_transitions.name_grounding();
    
    Domain domain_lfit(config_reader);
    domain_lfit.index_transitions(symbolic_transitions);
    TransitionGroup symbolic_transitions_non_closed_world = domain_lfit.get_transitions_no_closed_world(symbolic_transitions);
    
    std::stringstream ss_domain_types;
    domain_lfit.write_LFIT_predicate_types(ss_domain_types);
    symbolic_transitions_non_closed_world.write_lfit_transitions_to_file(transitions_learning_path, 
                                                                         ss_domain_types.str(), 
                                                                         domain_lfit.get_actions_non_typed());
    
    learn_lfit();
    LOG(INFO) << "LFIT learner finished";
    
    // get learned rules
    ReaderLFIT reader_lfit(lfit_rules_path);
    RuleSetPtr new_rules_ptr;
    LOG(INFO) << "Selecting best subset of rules";
    reader_lfit.read_ruleset(new_rules_ptr, transitions, domain.get_constant_predicates(), config_reader);
    
    return new_rules_ptr;
}


bool LFITLearner::learn_lfit()
{
    bool lf1t = true; // true = lf1t  |  false = lfkt
    std::string lfit_command = learner_path + " " + transitions_learning_path;
    if (lf1t) { // LF1T
        lfit_command += " 0";
        lfit_command += " 1"; // specialization
//         lfit_command += " 1"; // delay
//         lfit_command += " " + boost::lexical_cast<std::string>(maximum_number_preconditions); // delay
    }
    else { // LFKT (LUST)
        lfit_command += " 1";
        lfit_command += " 3"; // probabilistic
    }
    // probabilistic
    lfit_command += " > " + lfit_rules_path;
//     if (verbose < 1) {
//         lfit_command += " &>/dev/null";
//     }
    
    LOG(INFO) << "Executing LFIT: " + lfit_command;
    int res = system(lfit_command.c_str());
    if ((res >> 8) == 127) {
        // /bin/sh not being able to find the program to execute.
        LOG(ERROR) << "Learner executable cannot be found!!!";
        throw std::ios_base::failure("Learner executable couldn't be found");
    }
    else if (res != 0) {
        LOG(WARNING) << "Learner exited with error " << res;
        throw std::runtime_error("LFIT learner failed");
    }
    
    return true;
}


void LFITLearner::remove_unused_rules(RuleSetPtr rules_ptr, const TransitionGroup& transitions)
{
    LOG(INFO) << "Removing unused rules";
    for (RuleSet::iterator rule_ptr_it = rules_ptr->begin(); rule_ptr_it != rules_ptr->end(); ) {
        if ((*rule_ptr_it)->is_covered_by_at_least_one_transition(transitions, true, true)) {
            ++rule_ptr_it;
        }
        else {
            rule_ptr_it = rules_ptr->erase(rule_ptr_it);
        }
    }
}

