/*
 * LFIT types
 * Copyright (C) 2015  David Martínez <dmartinez@iri.upc.edu>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "lfit_types.h"

#define PENALTY_PREDICATE 1
#define PENALTY_RULE 2

using namespace lfit;

namespace lfit {
bool operator<(const Coverage& l, const Coverage& r)
{
    return l.get_estimator() < r.get_estimator();
}

bool operator<(const RuleWithCoverage& l, const RuleWithCoverage& r)
{
    if (l.rule_ptr < r.rule_ptr)
        return true;
    else if (r.rule_ptr < l.rule_ptr)
        return false;
    else
        return l.coverage < r.coverage;
}

std::ostream& operator<<(std::ostream &out, const RuleWithCoverage& rule_cov)
{
    out << "[";
    for (OutcomeGroup::const_iterator out_it = rule_cov.rule_ptr->outcomes.begin(); out_it != rule_cov.rule_ptr->outcomes.end(); ++out_it) {
        out << *out_it;
        if ( boost::next(out_it, 1) != rule_cov.rule_ptr->outcomes.end()) {
            out << " || ";
        }
    }
    out << "]";
    out << " <- "
        << rule_cov.rule_ptr->name << " & " 
        << rule_cov.rule_ptr->get_preconditions()
        << " (" << rule_cov.coverage.n_covered << "/" << rule_cov.coverage.n_total << ")";
    return out;
}
}
            
const RuleWithCoverage::TransitionsChangesCoverage& RuleWithCoverage::get_covered_transitions_full(const TransitionGroup& transitions) const
{
    if (!(*covered_transitions_full_opt_ptr)) {
        (*covered_transitions_full_opt_ptr) = TransitionsChangesCoverage(transitions.size_set());
        TransitionSet::iterator transition_set_it = transitions.begin_set();
        for (size_t i = 0; i < transitions.size_set(); ++i) {
            (**covered_transitions_full_opt_ptr)[i] = rule_ptr->covers_transition_effects(*transition_set_it, transitions.ppddl_object_manager);
            ++transition_set_it;
        }
    }
    
    return (**covered_transitions_full_opt_ptr);
}

bool RuleWithCoverage::is_good_candidate(const ConfigReader& config_reader, const Domain& domain, const TransitionGroup& transitions) const
{
    uint MAX_PRECONDITIONS = 4; // TODO config reader
//     float MIN_PROBABLITY = 0.5;
    
    if ( // no use for a 0.0 probability outcome
         (!utils::is_close(rule_ptr->outcomes.begin()->probability, (float)0.0)) &&
         // no need for negated effects if variables will change automatically to false
         (config_reader.get_variable<bool>("variable_maintains_to_next_state") || rule_ptr->outcomes.begin()->begin()->positive) && 
         domain.check_rule_types(*rule_ptr) &&
         (rule_ptr->get_preconditions().size() <= MAX_PRECONDITIONS) &&
//          (rule_ptr->outcomes.begin()->probability > MIN_PROBABLITY) &&
         rule_ptr->is_covered_by_at_least_one_transition(transitions, true, true) 
//          !static_cast<const SymbolicRule>(*rule_ptr).check_if_rule_groundings_cover_transition_preconditions_more_than_once(transitions)
       ) 
    {
//         RuleWithCoverageSet rule_set;
//         rule_set.insert(*this);
//         if (rule_set.get_score(transitions, config_reader, 0, config_reader.get_variable<float>("lfit_learner_score_optimistic_value")) > -10) {
//             *covered_transitions_full_opt_ptr = TransitionsChangesCoverageOpt();
            return true;
//         }
//         else {
//             *covered_transitions_full_opt_ptr = TransitionsChangesCoverageOpt();
//             return false;
//         }
    }
    else {
        return false;
    }
}

RuleWithCoverage RuleWithCoverage::get_inverse_rule() const
{
    // rule_ptr
    RulePtr inverse_rule_ptr = rule_ptr->make_shared();
    PredicateGroup<TypedAtom> inverse_outcome_preds;
    inverse_outcome_preds.insert(inverse_rule_ptr->outcomes.begin()->begin()->get_negated());
    Outcome inverse_outcome(inverse_outcome_preds, 1.0 - inverse_rule_ptr->outcomes.begin()->probability);
    inverse_rule_ptr->outcomes.clear();
    inverse_rule_ptr->outcomes.insert(inverse_outcome);
    // coverage
    Coverage inverse_coverage(coverage.n_total - coverage.n_covered, coverage.n_total);
    // result
    return RuleWithCoverage(inverse_rule_ptr, inverse_coverage);
}

float RuleWithCoverage::get_confidence(const float confidence_interval) const
{
    // hoeffding inequality
    // http://en.wikipedia.org/wiki/Hoeffding%27s_inequality
    // \alpha <= 1 - (2 * e ^ (-2 n t^2))
    // n = number samples
    // t = confidence interval
    return utils::get_hoeffding_inequality_confidence(coverage.n_total, confidence_interval);
}

RuleSet RuleWithCoverageSet::get_rule_set() const
{
    RuleSet result;
    BOOST_FOREACH(const RuleWithCoverage& res_rule, *this) {
        result.push_back(res_rule.rule_ptr);
    }
    return result;
}

float RuleWithCoverageSet::get_confidence(const float confidence_interval) const
{
    if (empty()) {
        return 0.0;
    }
    // average??, sum??
    // TODO
    float rules_confidence = 0;
    BOOST_FOREACH(const RuleWithCoverage& rule, *this) {
        rules_confidence += rule.get_confidence(confidence_interval);
    }
    return rules_confidence / size();
}

float RuleWithCoverageSet::get_likelihood(const TransitionGroup& transitions, const float optimistic_value) const
{
    double total_likelihood = 0;
    
    uint transition_idx = 0;
    for (TransitionSet::const_iterator transition_it = transitions.begin_set(); transition_it != transitions.end_set(); ++transition_it) {
        Transitions::Coverage transition_likelihood(transition_it->get_changes());
        BOOST_FOREACH(const RuleWithCoverage& rule_cov, *this) {
            transition_likelihood += rule_cov.get_covered_transitions_full(transitions)[transition_idx];
        }
        total_likelihood += std::max(log(transition_likelihood.get_likelihood(true, optimistic_value)),-10000.0) 
                            * transition_it->get_number_repetitions();
        ++transition_idx;
    }
    return total_likelihood/transitions.size_historial();
}

float RuleWithCoverageSet::get_penalty() const
{
    if (empty()) {
        return 0.0;
    }
    uint number_predicates = 0;
//     uint number_rules = size();
    BOOST_FOREACH(const RuleWithCoverage& rule, *this) {
        number_predicates += rule.rule_ptr->get_preconditions().size();
    }
    
//     return (number_rules * PENALTY_RULE) + ((number_predicates * PENALTY_PREDICATE)/number_rules);
    return (number_predicates * PENALTY_PREDICATE);
}

float RuleWithCoverageSet::get_score(const TransitionGroup& transitions,
                                     const ConfigReader& config_reader,
                                     const int verbose,
                                     const float optimistic_value
                                    ) const
{
    float k_like = config_reader.get_variable<float>("lfit_learner_score_likelihood_scaling");
    float k_regu = config_reader.get_variable<float>("lfit_learner_score_regularization_scaling");
    
    CLOG_IF(verbose>=1, INFO, "trivial") << "Score rules with " << transitions.size_historial() << " transitions";
    BOOST_FOREACH(const RuleWithCoverage& current_rule, *this) {
        CLOG_IF(verbose>=1, INFO, "trivial") << current_rule;
    }
    
//     float confidence = get_confidence(config_reader.get_variable<float>("lfit_learner_score_confidence_interval"));
    float confidence = utils::get_hoeffding_inequality_confidence(transitions.size_historial(), config_reader.get_variable<float>("lfit_learner_score_confidence_interval"));
    float likelihood = get_likelihood(transitions, optimistic_value);
    float regularization = get_penalty();
    float score;
    if (config_reader.get_variable<bool>("lfit_learner_score_use_confidence")) {
        confidence = std::max(confidence, (float)0.0001); // hardcoded min
        score = (k_like*likelihood) - ((1/confidence)*k_regu*regularization);
        CLOG_IF(verbose>=1, INFO, "trivial") << "Conf: " << confidence << "*" << ", "
                                             << "Lik: " << likelihood << "*" << k_like << ", "
                                             << "Pen: " << (1/confidence) << "*" << regularization << "*" << k_regu << ", "
                                             << "All: " << score;
    }
    else {
        score = (k_like*likelihood) - (k_regu*regularization);
        CLOG_IF(verbose>=1, INFO, "trivial") << "Conf: " << confidence << "*" << ", "
                                             << "Lik: " << likelihood << "*" << k_like << ", "
                                             << "Pen: " << regularization << "*" << k_regu << ", "
                                            << "All: " << score;
    }
    
    return score;
}

RuleWithCoverageSet RuleWithCoverageSet::get_rules_covered_by(const RuleWithCoverage& rule) const
{
    RuleWithCoverageSet result;
    BOOST_FOREACH(const RuleWithCoverage& orig_rule, *this) {
        if (orig_rule.rule_ptr->get_preconditions().includes_predicates(rule.rule_ptr->get_preconditions())) {
            result.insert(orig_rule);
        }
    }
    return result;
}
