/*
 * LFIT probabilistic postprocessing
 * Copyright (C) 2015  David Martínez <dmartinez@iri.upc.edu>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

// only for debug purposes
// #define IGNORE_NEGATIVE_RULES
// #define IGNORE_POSITIVE_RULES

#include "probabilistic_lfit.h"
#include <boost/graph/graph_concepts.hpp>

using namespace lfit;


RuleWithCoverageSet ProbabilisticLFITAnalyzer::select_best_ruleset(const RuleWithCoverageSet& current_candidates,
                                                                   const TransitionGroup& head_transitions,
                                                                   const int verbose
                                                                  )
{
    RuleConflictSolver conflict_solver(current_candidates, head_transitions, config_reader);
    RuleWithCoverageSet non_conflicting_ruleset;
    return conflict_solver.get_best_subset_non_conflicting_heuristic(head_transitions, config_reader, verbose);
}

bool ProbabilisticLFITAnalyzer::remove_conflicting_leafs(const TransitionGroup& head_transitions, const RuleWithCoverageSet& rules_to_check, 
                                                         RuleSubsumptionGraph& subsumption_graph, 
                                                         RuleWithCoverageSet& best_operators, double& best_candidate_score)
{
    CLOG_IF(debug >=2, INFO, "trivial") << "Conflicting rules size is " << rules_to_check.size();
    RuleWithCoverageSet current_best_ruleset = select_best_ruleset(rules_to_check, head_transitions, debug);
    CLOG_IF(debug >=2, INFO, "trivial") << "Finished planning operator selection.";
    double new_score = current_best_ruleset.get_score(head_transitions, config_reader);
    CLOG_IF(debug >=3, INFO, "trivial") << "New score is: " << new_score;
    if (new_score > best_candidate_score) {
        best_candidate_score = new_score;
        best_operators = current_best_ruleset;
        if (debug >=1) {
            CLOG_IF(debug>=2, INFO, "trivial") << "---------";
            CLOG_IF(debug >=1, INFO, "trivial") << "New best candidate (1) with score: " << new_score;
            best_operators.get_score(head_transitions, config_reader, debug);
        }
    }
    return subsumption_graph.remove_leafs_not_in(current_best_ruleset);
}

RuleWithCoverageSet ProbabilisticLFITAnalyzer::select_rules_for_head(const RuleWithCoverageSet& head_rules,
                                                                     const TransitionGroup& head_transitions)
{
    CLOG_IF(debug >=0, INFO, "trivial") << "Selecting rules from " << head_rules.size() << " candidates.";
    RuleSet result;
    
    CLOG_IF(debug >=0, INFO, "trivial") << "Removing potentially bad candidates.";
    RuleWithCoverageSet promising_head_rules;
    BOOST_FOREACH (const RuleWithCoverage& head_rule, head_rules) {
        RuleWithCoverageSet candidate_head_rule_set;
        candidate_head_rule_set.insert(head_rule);
        /* min optimistic score is:
         * For each effect, the max error should be p=0.5 -> ln(0.5)
         * For each precondition, add the penalty
         * min_opt_score = ln(0.5) * MAX_NUM_CHANGES + MAX_PRECONDITIONS * PENALTY_PRECONDITION
         * min_opt_score = -0.69   *       2         +          10        *         0.05
         * Using min_opt_score = -2 as a safe value
        */
        float rule_optimistic_score = candidate_head_rule_set.get_score(head_transitions, config_reader, false, config_reader.get_variable<float>("lfit_learner_score_optimistic_value"));
        if (rule_optimistic_score > -1.0) { // TODO each effect, max 0.5 prob error -> ln(0.5) = -0.69
//             head_rule.coverage.optimistic_score = rule_optimistic_score;
            promising_head_rules.insert(head_rule);
        }
    }
    CLOG_IF(debug >=0, INFO, "trivial") << "Selecting rules from " << promising_head_rules.size() << " promising candidates.";
    
    RuleSubsumptionGraph subsumption_graph(promising_head_rules);
    CLOG_IF(debug >=4, INFO, "trivial") << "Starting graph " << subsumption_graph;
    
    RuleWithCoverageSet best_operators;
    double best_candidate_score = -1 * std::numeric_limits<float>::max();
    bool something_changed = true;
    while (something_changed) {
        // Solve conflicts with leafs
        something_changed = remove_conflicting_leafs(head_transitions, subsumption_graph.get_leafs(), subsumption_graph, best_operators, best_candidate_score);
        
        // Solve conflicts with leafs and their parents
        something_changed = remove_conflicting_leafs(head_transitions, subsumption_graph.get_leafs_and_first_parent(), subsumption_graph, best_operators, best_candidate_score) || 
                            something_changed;
    }
    
    return best_operators;
}

boost::optional<RuleSet> ProbabilisticLFITAnalyzer::check_inverted_rules(const MapHeadRules& learned_rule_sets, const Predicate<TypedAtom>& current_head, const TransitionGroup& head_transitions)
{
    LOG(INFO) << "Checking inverted rules...";
    // if already learned the negated_outcome, try negating the learned rules
    RuleWithCoverageSet inverted_rules;
    MapHeadRules::const_iterator learned_rule_sets_it = learned_rule_sets.find(current_head.get_negated());
    if (learned_rule_sets_it != learned_rule_sets.end()) {
        RuleWithCoverageSet learned_rule_set_for_head = learned_rule_sets_it->second;
        BOOST_FOREACH(const RuleWithCoverage& rule_to_invert, learned_rule_set_for_head) {
            RulePtr inverted_rule_ptr = rule_to_invert.rule_ptr->make_shared();
            OutcomeGroup inverted_outcome_group;
            Outcome inverted_outcome(inverted_rule_ptr->outcomes.begin()->probability);
            
            Predicate<TypedAtom>::Opt inverted_outcome_predicate_opt;
            BOOST_FOREACH(const Predicate<TypedAtom>& precondition, inverted_rule_ptr->get_preconditions()) {
                if (precondition.name == current_head.name) {
                    inverted_outcome_predicate_opt = precondition.get_negated();
                }
            }
            if (inverted_outcome_predicate_opt) {
                inverted_outcome.insert(*inverted_outcome_predicate_opt);
                inverted_outcome_group.insert(inverted_outcome);
                inverted_rule_ptr->outcomes = inverted_outcome_group;
                RuleWithCoverage inverted_rule_with_cov(inverted_rule_ptr, rule_to_invert.coverage);
                inverted_rules.insert(inverted_rule_with_cov);
            }
        }
    }
    if (inverted_rules.get_score(head_transitions, config_reader) > -1.0) {
        LOG(INFO) << "Inverted rules succeeded! Skipping candidate selection.";
        LOG(INFO) << "Inverted rules score is: ";
        inverted_rules.get_score(head_transitions, config_reader, 2);
        return inverted_rules.get_rule_set();
    }
    else {
        return boost::optional<RuleSet>();
    }
}

TransitionGroup ProbabilisticLFITAnalyzer::obtain_head_transitions(const TransitionGroup& transitions, const Predicate<TypedAtom>& head)
{
    LOG(INFO) << "Getting transitions for change: " << head;
    TransitionGroup head_transitions(transitions.ppddl_object_manager);
    for (TransitionGroup::TransitionHistorial::const_iterator transition_it_it = transitions.begin_historial();
            transition_it_it != transitions.end_historial(); 
            ++transition_it_it) 
    {
        PredicateGroup<TypedAtom> post_state;
        if (config_reader.get_variable<bool>("variable_maintains_to_next_state")) {
            post_state = (*transition_it_it)->get_prev_state(); // everything maintains as default
        }
        // add head predicates values
        BOOST_FOREACH(const Predicate<TypedAtom>& post_predicate, (*transition_it_it)->get_post_state()) {
            if ( (post_predicate.name == head.name) && // Update only predicates with same name
                    (post_predicate.positive == head.positive) ) { // We are learning predicate changes, sign has to be considered also
                post_state.update_predicate(post_predicate);
            }
        }
        
        Transition new_transition((*transition_it_it)->get_prev_state(), 
                                    (*transition_it_it)->get_action(), 
                                    post_state, 
                                    (*transition_it_it)->get_reward(), 
                                    config_reader.get_variable<bool>("variable_maintains_to_next_state"));
        head_transitions.add_transition(new_transition);
    }
    return head_transitions;
}

RuleSet ProbabilisticLFITAnalyzer::select_rules_for_each_head(const MapHeadRules& map_head_rules, const TransitionGroup& transitions)
{
    RuleSet result;
    MapHeadRules learned_rule_sets;
    BOOST_FOREACH(const MapHeadRules::value_type& pair_head_rules, map_head_rules) {
#ifdef IGNORE_NEGATIVE_RULES
        // for debug purposes only
        if (!pair_head_rules.first.positive) { 
            continue;
        }
#endif
#ifdef IGNORE_POSITIVE_RULES
        // for debug purposes only
        if (pair_head_rules.first.positive) { 
            continue;
        }
#endif
        
        TransitionGroup head_transitions = obtain_head_transitions(transitions, pair_head_rules.first);
        
        boost::optional<RuleSet> inverted_rules_opt = check_inverted_rules(learned_rule_sets, pair_head_rules.first, head_transitions);
        if (inverted_rules_opt) {
            result.add_rules(*inverted_rules_opt);
        }
        else {
            RuleWithCoverageSet new_rules = select_rules_for_head(pair_head_rules.second, head_transitions);
            result.add_rules(new_rules.get_rule_set());
            learned_rule_sets[pair_head_rules.first] = new_rules;
        }
    }
    return result;
}

RuleSetPtr ProbabilisticLFITAnalyzer::select_rules(const std::set<lfit::RuleWithCoverage>& lfit_rules)
{
    MapHeadRules map_head_rules;
    
    // group rules
    BOOST_FOREACH(const RuleWithCoverage& rule_coverage, lfit_rules) {
        map_head_rules[*rule_coverage.rule_ptr->outcomes.begin()->begin()].insert(rule_coverage);
    }
    
    return RuleSetPtr(new RuleSet(select_rules_for_each_head(map_head_rules, transitions)));
}

