/*
 * LFIT reader and parser
 * Copyright (C) 2015  David Martínez <dmartinez@iri.upc.edu>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#define DEBUG_VALUE 0

#include "reader_lfit.h"

using namespace lfit;

/**
 * \class ParserLFIT
 */

ParserLFIT::ParserLFIT():
    debug(DEBUG_VALUE)
{ }

void ParserLFIT::set_default_rule_variables(RulePtr lfit_rule_ptr)
{
    // Here we have
    // action(?Y) & prec1(?Y ?X) :- out1(?Y ?X)
    // Change to have action parameters ordered (?X, ?Y, ...)
    // action(?X) & prec1(?X ?Y) :- out1(?X ?Y)
    // change variables to aux numbers so no variable is overwritten
    for (size_t param_idx = 0; param_idx < lfit_rule_ptr->params.size(); ++param_idx) {
        if (lfit_rule_ptr->params[param_idx].name != DEFAULT_PARAM_NAMES[param_idx]) {
            std::string orig_var = lfit_rule_ptr->params[param_idx].name;
            std::string new_var = DEFAULT_PARAM_NAMES[param_idx];
            std::string param_idx_str = boost::lexical_cast<std::string>(param_idx);
            // copy with auxiliar var so no variable is overwritten
            lfit_rule_ptr->change_variable(orig_var, "?" + param_idx_str);
            lfit_rule_ptr->change_variable(new_var, orig_var);
            lfit_rule_ptr->change_variable("?" + param_idx_str, new_var);
        }
    }
    
    //**************************
    
    std::vector<PPDDL::Object> deictic_params = lfit_rule_ptr->get_deictic_params_v();
    
    // get head params and remove non-deictic
    std::vector<PPDDL::Object> lfit_head_deictic_params = lfit_rule_ptr->outcomes.begin()->begin()->params;
    BOOST_FOREACH(const PPDDL::Object& action_param, lfit_rule_ptr->params) {
        for (std::vector<PPDDL::Object>::iterator lfit_head_param_it = lfit_head_deictic_params.begin();
             lfit_head_param_it != lfit_head_deictic_params.end(); ) {
            if (action_param.name == lfit_head_param_it->name) {
                // not deictic param
                lfit_head_param_it = lfit_head_deictic_params.erase(lfit_head_param_it);
            }
            else {
                ++lfit_head_param_it;
            }
        }
    }
    
    // reorder deictic params: first head_deictic, then other deictic params
    // so remove other_deictic and add them at the end
    BOOST_FOREACH(const PPDDL::Object& lfit_head_deictic_param, lfit_head_deictic_params) {
        deictic_params.erase( std::remove( deictic_params.begin(), deictic_params.end(), lfit_head_deictic_param ),
                              deictic_params.end() );
    }
    BOOST_REVERSE_FOREACH(const PPDDL::Object& lfit_head_deictic_param, lfit_head_deictic_params) {
        deictic_params.insert(deictic_params.begin(), lfit_head_deictic_param);
    }
    
    // change variables to aux numbers so no variable is overwritten
    for (size_t i = 0; i < deictic_params.size(); ++i) {
        lfit_rule_ptr->change_variable(deictic_params[i].name, "?" + boost::lexical_cast<std::string>(i));
    }
    // change aux numbers to default variable names
    for (size_t i = 0; i < deictic_params.size(); ++i) {
        lfit_rule_ptr->change_variable("?" + boost::lexical_cast<std::string>(i), 
                                         DEFAULT_PARAM_NAMES[i + lfit_rule_ptr->params.size()]);
    }
    
    // Here we have
    // action(?X ?Y) & prec1(?X ?V) :- out1(?Z ?U)
    // Change to outcome vars first
    // action(?Z ?V) & prec1(?Z ?V) :- out1(?X ?Y)
    
    // change variables to aux numbers so no variable is overwritten
    std::vector<PPDDL::Object> head_params = lfit_rule_ptr->outcomes.begin()->begin()->params;
    for (size_t out_idx = 0; out_idx < head_params.size(); ++out_idx) {
        if (out_idx < lfit_rule_ptr->params.size()) { // else no more changes are needed
            std::string orig_var = lfit_rule_ptr->params[out_idx].name; // == DEFAULT_PARAM_NAMES[out_idx]
            std::string new_var = head_params[out_idx].name;
            std::string out_idx_str = boost::lexical_cast<std::string>(out_idx);
            // copy with auxiliar var so no variable is overwritten
            lfit_rule_ptr->change_variable(orig_var, "?" + out_idx_str);
            lfit_rule_ptr->change_variable(new_var, orig_var);
            lfit_rule_ptr->change_variable("?" + out_idx_str, new_var);
        }
    }
    
    lfit_rule_ptr->update_deictic_params();
}

NonTypedAtom ParserLFIT::parse_atom(const std::string& str) const
{
    CLOG_IF(debug > 6, INFO, "trivial") << "Atom string is " << str;
    return NonTypedAtom::get_atom_from_grounded_name_item(str);
}

Predicate<NonTypedAtom> ParserLFIT::parse_predicate(const std::string& str) const
{
    CLOG_IF(debug > 6, INFO, "trivial") << "Predicate string is " << str;
    bool positive = false;
    std::string positive_str = str.substr(str.find("(")+1, 1);
    if (positive_str == "1") {
        positive = true;
    }
    else if (positive_str == "0") {
        positive = false;
    }
    else {
        throw ReaderUnsupportedTypeException("RDDL", "Special param: " + positive_str + "");
    }
    
    std::string symbol_str = str.substr(0, str.find("("));
    Predicate<NonTypedAtom> predicate(parse_atom(symbol_str), positive);
    CLOG_IF(debug > 5, INFO, "trivial") << "Predicate is " << predicate;
    return predicate;
}

void ParserLFIT::parse_head(const std::string& str, Predicate<NonTypedAtom>::Opt& head, Coverage& rule_coverage) const
{
    if (str.find(",T,0,") == std::string::npos) { // ",0,0)" -> means no covering
        head = Predicate<NonTypedAtom>::Opt(parse_predicate(str));
        std::string n_covered_str = str.substr(str.find(",T")+3, str.find(")") - str.find(",T") - 3);
        rule_coverage.n_covered = boost::lexical_cast<uint>(n_covered_str.substr(0, n_covered_str.find(",")));
        rule_coverage.n_total = boost::lexical_cast<uint>(n_covered_str.substr(n_covered_str.find(",")+1, std::string::npos));
        assert(rule_coverage.n_covered != 0);
        assert(rule_coverage.n_total != 0);
    }
}

PredicateGroup<NonTypedAtom> ParserLFIT::parse_predicate_group(const std::string& str) const
{
    CLOG_IF(debug > 6, INFO, "trivial") << "PredGroup string is " << str;
    PredicateGroup<NonTypedAtom> result;
    std::stringstream ss_str(str);
    std::string item;
    char separator = ')';
    while (std::getline(ss_str, item, separator)) {
        // remove spaces and empty strings
        item.erase(remove_if(item.begin(), item.end(), isspace), item.end());
        if (item[0] == ',') {
            item = item.substr(1,std::string::npos);
        }
        // ignore action (can't be represented with predicate, doesn't take 0,1 values but name value).
        if ((item != "") && (item.find("action") == std::string::npos)) { 
            result.insert(parse_predicate(item));
        }
    }
    
    CLOG_IF(debug > 5, INFO, "trivial") << "PredGroup is " << result;
    return result;
}

RuleWithCoverage ParserLFIT::parse_symbolic_rule(const std::string& original_rule_string, const bool variable_maintains_to_next_state, const Domain& domain) const
{
    CLOG_IF(debug > 6, INFO, "trivial") << "Rule string is " << original_rule_string;
    
    Coverage rule_coverage(0,0);
    Predicate<NonTypedAtom>::Opt effect_predicate;
    
    // if starts with "action(" -> ignore
    if (original_rule_string.find("action(") == 0) {
        return RuleWithCoverage(SymbolicRulePtr(), Coverage(0,0));
    }
    if (isupper(original_rule_string[0])) { // constant!
        return RuleWithCoverage(SymbolicRulePtr(), Coverage(0,0));
    }
    
    // Remove tailing "." as it is not needed
    std::string rule_string = original_rule_string.substr(0, original_rule_string.size()-1);
    
    // effect
    std::string effect_string = rule_string.substr(0, rule_string.find(":"));
    parse_head(effect_string, effect_predicate, rule_coverage);
    if (rule_coverage.n_covered == 0) {
        LOG(WARNING) << "CHECK THIS!";
        return RuleWithCoverage(SymbolicRulePtr(), rule_coverage);
    }
    
    // preconditions
    TypedAtom action;
    PredicateGroup<NonTypedAtom> preconditions;
    size_t preconditions_pos;
    if ((preconditions_pos = rule_string.find(":-")) != std::string::npos) {
        std::string preconditions_string = rule_string.substr(preconditions_pos+2, std::string::npos);
        preconditions = parse_predicate_group(preconditions_string);
        TypedAtom::Opt action_opt = parse_action_in_preconditions(preconditions_string);
        if (action_opt) {
            // copy action and get types
            action = domain.transform_to_typed(NonTypedAtom(action_opt->name, action_opt->get_param_names()));
        }
        else {
            return RuleWithCoverage(SymbolicRulePtr(), rule_coverage); // no action...
        }
    }
    else {
        return RuleWithCoverage(SymbolicRulePtr(), rule_coverage); // no preconditions also means no action
    }
    
    SymbolicRulePtr result_rule_ptr(new SymbolicRule(action, domain.transform_to_typed(preconditions)));
    PredicateGroup<NonTypedAtom> effect_predicate_group;
    if (effect_predicate->positive) {
        effect_predicate_group.insert(*effect_predicate);
    }
    else {
        // Complement of the head and coverage
        effect_predicate_group.insert(effect_predicate->get_negated());
        rule_coverage = Coverage(rule_coverage.n_total - rule_coverage.n_covered, rule_coverage.n_total);
    }
    Outcome effect_outcome(domain.transform_to_typed(effect_predicate_group), 
                            (float)rule_coverage.n_covered / (float)rule_coverage.n_total);
    result_rule_ptr->outcomes.insert(effect_outcome);
    
    // if number any param in outcome is not in rule nor preconditions, ignore the rule (it is not deictic!, cannot be grounded!)
    std::vector<PPDDL::Object> out_params = result_rule_ptr->outcomes.begin()->begin()->params;
    std::vector<PPDDL::Object> in_params = result_rule_ptr->params;
    std::vector<PPDDL::Object> in_deictic_params = result_rule_ptr->get_deictic_params_v();
    std::set<PPDDL::Object> in_all_params_set(in_params.begin(), in_params.end());
    in_all_params_set.insert(in_deictic_params.begin(), in_deictic_params.end());
    BOOST_FOREACH(const PPDDL::Object& out_param, out_params) {
        if (in_all_params_set.count(out_param) == 0) {
            return RuleWithCoverage(SymbolicRulePtr(), rule_coverage); // not enough deictic params, the action cannot be grounded, not usable
        }
    }
    
    // if any precondition is negative and params are deictic and not in outcomes, ignore the rule
    std::set<PPDDL::Object> in_params_set(in_params.begin(), in_params.end());
    std::set<PPDDL::Object> out_params_set(out_params.begin(), out_params.end());
    BOOST_FOREACH(const Predicate<TypedAtom>& precondition, result_rule_ptr->get_preconditions()) {
        if (!precondition.positive) {
            BOOST_FOREACH(const PPDDL::Object& prec_param, precondition.params) {
                if ( (out_params_set.count(prec_param) == 0) && (in_params_set.count(prec_param) == 0) ) {
                    // Check if another precondition has the param
                    bool another_precondition_has_param = false;
                    BOOST_FOREACH(const Predicate<TypedAtom>& compared_precondition, result_rule_ptr->get_preconditions()) {
                        if (precondition != compared_precondition) {
                            if (compared_precondition.has_param(prec_param)) {
                                if (compared_precondition.positive || !compared_precondition.params.empty()) {
                                    another_precondition_has_param = true;
                                    // TODO check if this precondition actually is useful
                                }
                            }
                        }
                    }
                    if (!another_precondition_has_param) {
                        return RuleWithCoverage(SymbolicRulePtr(), rule_coverage); // deictic param can be grounded to everything as predicate is negated
                    }
                }
            }
        }
    }
    
    CLOG_IF(debug > 5, INFO, "trivial") << "Rule is " << *result_rule_ptr;
    return RuleWithCoverage(result_rule_ptr, rule_coverage);
}

TypedAtom::Opt ParserLFIT::parse_action_in_preconditions(const std::string& str) const
{
    CLOG_IF(debug > 6, INFO, "trivial") << "Action str is " << str;
    size_t action_str_pos = str.find("action(");
    if (action_str_pos == std::string::npos) {
        return TypedAtom::Opt(MAKE_TYPED_ATOM_NOACTION);
    }
    std::string action_string = str.substr(action_str_pos + 7, std::string::npos);
    action_string = action_string.substr(0, action_string.find(","));
    NonTypedAtom action_atom = NonTypedAtom::get_atom_from_grounded_name_item(action_string);
    TypedAtom action_typed_atom(action_atom.name);
    BOOST_FOREACH(const std::string& param_name, action_atom.get_param_names()) {
        action_typed_atom.params.push_back(PPDDL::Object(param_name, DEFAULT_OBJECT_TYPE));
    }
    
    CLOG_IF(debug > 5, INFO, "trivial") << "Recognized action is " << action_typed_atom;
    return TypedAtom::Opt(action_typed_atom);
}


/**
 * \class ReaderLFIT
 */

ReaderLFIT::ReaderLFIT(const std::string& file_path_): 
    FileReader(file_path_),
    debug(DEBUG_VALUE)
{}

void ReaderLFIT::read_ruleset(RuleSetPtr& read_ruleset, 
                              const TransitionGroup& transitions, 
                              const PredicateGroup<TypedAtom>& constant_predicates, 
                              const ConfigReader& config_reader)
{
    PredicateGroup<TypedAtom> constant_predicates_symbolic(constant_predicates);
    constant_predicates_symbolic.make_symbolic();
    
    ParserLFIT lfit_parser;
    MapActionToLfitRules map_actions_to_lfit_rules;
    
    Domain domain(config_reader);
    domain.index_transitions(transitions);
    CLOG_IF(debug > 0, INFO, "trivial") << "Reading rules";
    open_file();
    boost::optional<std::string> line_opt;
    std::set<RuleWithCoverage> lfit_rules;
    while ((line_opt = read_line())) {
        if (*line_opt != "") {
            RuleWithCoverage lfit_rule_cov = lfit_parser.parse_symbolic_rule(*line_opt, config_reader.get_variable<bool>("variable_maintains_to_next_state"), domain);
            if (lfit_rule_cov.rule_ptr) {
                assert(lfit_rule_cov.rule_ptr->outcomes.size() >= 1);
                assert(lfit_rule_cov.rule_ptr->outcomes.begin()->size() == 1);
                lfit_parser.set_default_rule_variables(lfit_rule_cov.rule_ptr);
                domain.set_rule_types(*lfit_rule_cov.rule_ptr);
                
                if ( lfit_rule_cov.is_good_candidate(config_reader, domain, transitions) ) {
                    lfit_rules.insert(lfit_rule_cov);
                }
                
                // Try inverse rule
                RuleWithCoverage inverse_lfit_rule_cov = lfit_rule_cov.get_inverse_rule();
                if ( inverse_lfit_rule_cov.is_good_candidate(config_reader, domain, transitions) ) {
                    lfit_rules.insert(inverse_lfit_rule_cov);
                }
            }
        }
    }
    
    CLOG_IF(debug > 0, INFO, "trivial") << "Rule selection with " << lfit_rules.size() << " candidates";
    ProbabilisticLFITAnalyzer probabilistic_analyzer(DEBUG_VALUE, transitions, config_reader);
    read_ruleset = probabilistic_analyzer.select_rules(lfit_rules);
    CLOG_IF(debug > 0, INFO, "trivial") << "Finished parsing LFIT rules";
}

