/*
 * Parser for LFIT files.
 * Copyright (C) 2014  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef IRI_RULE_LEARNER_READER_LFIT_H
#define IRI_RULE_LEARNER_READER_LFIT_H 1

#include <boost/optional.hpp>
#include <symbolic/reader_base.h>
#include <symbolic/domain.h>
#include <lfit/probabilistic_lfit.h>


/**
 * \class ParserLFIT
 * \brief Parses LFIT strings.
 * 
 */
class ParserLFIT : FileParser
{
private:
    uint debug; // 0 -> nothing, 1 -> basic algorithm, 2 -> data read, 3 -> data parsed, 4 -> details
    
public:
    /*!
     *  \brief Constructor
     */
    ParserLFIT();
    
    virtual NonTypedAtom  parse_atom                  (const std::string& str) const;
    
    virtual TypedAtom     parse_typed_atom            (const std::string& str) const
    {
        throw ReaderUnsupportedTypeException("RDDL", "typed_atom");
    };
    
    void parse_head(const std::string& str, Predicate<NonTypedAtom>::Opt& head, lfit::Coverage& rule_coverage) const;
    
    virtual Predicate<NonTypedAtom> parse_predicate             (const std::string& str) const;
    
    virtual PredicateGroup<NonTypedAtom> parse_predicate_group  (const std::string& str) const;
    
    virtual Outcome         parse_outcome                 (const std::string& str) const
    {
        throw ReaderUnsupportedTypeException("RDDL", "outcome");
    };
    
    virtual OutcomeGroup    parse_outcome_group           (const std::string& str) const
    {
        throw ReaderUnsupportedTypeException("RDDL", "outcome_group");
    };
    
//     virtual SymbolicRulePtr parse_symbolic_rule           (const std::string& str) const;
    
    lfit::RuleWithCoverage parse_symbolic_rule           (const std::string& str, const bool variable_maintains_to_next_state, const Domain& domain) const;
    
    virtual RuleSetPtr      parse_symbolic_ruleset        (const std::string& str) const
    {
        throw ReaderUnsupportedTypeException("RDDL", "symbolic_ruleset");
    };
    
    virtual State           parse_state                   (const std::string& str) const
    {
        throw ReaderUnsupportedTypeException("RDDL", "state");
    };
    
    virtual PPDDL::TypeVector  parse_pddl_types              (const std::string& str) const
    {
        throw ReaderUnsupportedTypeException("RDDL", "types");
    };
    
    virtual GoalGroup       parse_global_goals            (const std::string& str) const
    {
        throw ReaderUnsupportedTypeException("RDDL", "global_goals");
    };
    
    virtual PPDDLObjectManager parse_objects_from_simple_string(const std::string& str) const
    {
        throw ReaderUnsupportedTypeException("RDDL", "objects");
    };
    
    TypedAtom::Opt parse_action_in_preconditions(const std::string& str) const;
    
    void set_default_rule_variables(RulePtr lfit_rule_ptr);
    
};


/**
 * \class ReaderLFIT
 * \brief Reads LFIT files and extracts rules..
 * 
 */
class ReaderLFIT : FileReader
{
private:
    uint debug; // 0 -> nothing, 1 -> basic algorithm, 2 -> data read, 3 -> data parsed, 4 -> details
    
public:
    /**
     * \brief Constructor
     */
    ReaderLFIT(const std::string& file_path_);
    
    /**
     * \brief Read a rule set
     */
    virtual void read_ruleset(RuleSetPtr& ruleset) { throw ReaderUnsupportedTypeException("RDDL", "ruleset"); } ;
    
    // TODO clean
    void read_ruleset(RuleSetPtr& read_ruleset, const TransitionGroup& transitions, const PredicateGroup<TypedAtom>& constant_predicates, const ConfigReader& config_reader);
    
    /**
     * \brief Read the domain types
     */
    virtual void read_domain_types(PPDDL::TypeVector& types)
    {
        throw ReaderUnsupportedTypeException("RDDL", "object_types");
    };
    
    /**
     * \brief Read the state and goals
     */
    virtual void read_state_and_goals(State& state, GoalGroup& goals)
    {
        throw ReaderUnsupportedTypeException("RDDL", "state, goals");
    };
    
    /**
     * \brief Read a set of transitions
     */
    virtual void read_transitions(TransitionGroup& transitions)
    {
        throw ReaderUnsupportedTypeException("RDDL", "transitions");
    };
    
};

#endif