#include "subsumption_graph.h"

namespace lfit {

bool operator<(const Vertex& l, const Vertex& r)
{
    return (l.rulecov < r.rulecov);
}

std::ostream& operator<<(std::ostream &out, const RuleSubsumptionGraph& graph)
{
    BOOST_FOREACH(const Vertex::Ptr& leaf, graph.leafs) {
        out << "---- New leaf ----" << std::endl;
        std::list<Vertex::Ptr> list_of_vertex_to_print;
        list_of_vertex_to_print.push_back(leaf);
        while (!list_of_vertex_to_print.empty()) {
            Vertex::Ptr new_vertex_ptr = list_of_vertex_to_print.front();
            list_of_vertex_to_print.pop_front();
            out << new_vertex_ptr->rulecov.rule_ptr->name << " <- " << new_vertex_ptr->rulecov.rule_ptr->get_preconditions() << " [" << new_vertex_ptr->adjacency_list.size() << "]" << std::endl;
            BOOST_FOREACH(const Vertex::Ptr& parent_vertex, new_vertex_ptr->inverse_adjacency_list) {
                list_of_vertex_to_print.push_front(parent_vertex);
            }
        }
    }
    return out;
}

void RuleSubsumptionGraph::add_vertex(const RuleWithCoverage &rule)
{
    if(vertex_map.count(rule.rule_ptr) == 0)
    {
        Vertex::Ptr vertex_ptr(new Vertex(rule));
        vertex_map[rule.rule_ptr] = vertex_ptr;
        leafs.insert(vertex_ptr);
    }
}

void RuleSubsumptionGraph::add_edge(const RuleWithCoverage& from, const RuleWithCoverage& to)
{
    Vertex::Ptr from_vertex = vertex_map[from.rule_ptr];
    Vertex::Ptr to_vertex = vertex_map[to.rule_ptr];
    from_vertex->adjacency_list.push_back(to_vertex);
    to_vertex->inverse_adjacency_list.push_back(from_vertex);
    leafs.erase(from_vertex);
}

void RuleSubsumptionGraph::remove_vertex(const RuleWithCoverage& rule)
{
    assert(vertex_map.find(rule.rule_ptr) != vertex_map.end());
    Vertex::Ptr rule_vertex = vertex_map[rule.rule_ptr];
    BOOST_FOREACH(Vertex::Ptr &child, rule_vertex->adjacency_list) {
        child->inverse_adjacency_list.remove(rule_vertex);
    }
    BOOST_FOREACH(Vertex::Ptr &parent, rule_vertex->inverse_adjacency_list) {
        parent->adjacency_list.remove(rule_vertex);
        if (parent->adjacency_list.empty()) {
            leafs.insert(parent);
        }
    }
    leafs.erase(rule_vertex);
    vertex_map.erase(rule.rule_ptr);
}

RuleSubsumptionGraph::RuleSubsumptionGraph(const RuleWithCoverageSet& lfit_rules) {
        
    for (RuleWithCoverageSet::iterator rule_it1 = lfit_rules.begin(); rule_it1 != lfit_rules.end(); ++rule_it1) {
        add_vertex(*rule_it1);
        RuleWithCoverageSet::iterator rule_it2 = rule_it1;
        for (++rule_it2; rule_it2 != lfit_rules.end(); ++rule_it2) {
            add_vertex(*rule_it2);
            if (rule_it1->rule_ptr->name == rule_it2->rule_ptr->name) {
                if (abs(rule_it1->rule_ptr->get_preconditions().size() - rule_it2->rule_ptr->get_preconditions().size()) == 1) {
                    // only nodes when there is only one predicate difference
                    if (rule_it2->rule_ptr->get_preconditions().includes_predicates(rule_it1->rule_ptr->get_preconditions())) {
                        add_edge(*rule_it1, *rule_it2);
                    }
                    else if (rule_it1->rule_ptr->get_preconditions().includes_predicates(rule_it2->rule_ptr->get_preconditions())) {
                        add_edge(*rule_it2, *rule_it1);
                    }
                }
            }
        }
    }
}

RuleWithCoverageSet RuleSubsumptionGraph::get_leafs() const {
    RuleWithCoverageSet result;
    BOOST_FOREACH(const Vertex::Ptr &leaf_vertex, leafs) {
        result.insert(leaf_vertex->rulecov);
    }
    return result;
}

RuleWithCoverageSet RuleSubsumptionGraph::get_leafs_and_first_parent() const {
    RuleWithCoverageSet result;
    BOOST_FOREACH(const Vertex::Ptr &leaf_vertex, leafs) {
        result.insert(leaf_vertex->rulecov);
        RuleWithCoverageSet parents = get_rule_parents(leaf_vertex->rulecov);
        result.insert(parents.begin(),parents.end());
    }
    return result;
}

RuleWithCoverageSet RuleSubsumptionGraph::get_rule_parents(const RuleWithCoverage& rule) const {
    RuleWithCoverageSet result;
    // TODO When upgrading to C++11, update to ::at()
    VertexMap::const_iterator rule_vertex_it = vertex_map.find(rule.rule_ptr);
    assert(rule_vertex_it != vertex_map.end());
    BOOST_FOREACH(const Vertex::Ptr &parent_vertex, rule_vertex_it->second->inverse_adjacency_list) {
        result.insert(parent_vertex->rulecov);
    }
    return result;
}

void RuleSubsumptionGraph::remove_leaf(const RuleWithCoverage& rule) {
    remove_vertex(rule);
}

bool RuleSubsumptionGraph::remove_leafs_not_in(const RuleWithCoverageSet& ruleset) {
    RuleWithCoverageSet leafs = get_leafs();
    BOOST_FOREACH(const RuleWithCoverage& good_rule, ruleset) {
        leafs.erase(good_rule);
    }
    BOOST_FOREACH(const RuleWithCoverage& bad_leaf, leafs) {
        remove_leaf(bad_leaf);
    }
    return !leafs.empty(); // if nothing to remove -> false
}

}