/*
 * Graph representing subsumption relationships between rule preconditions.
 * Copyright (C) 2015  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef REXD_LFIT_SUBSUMPTION_GRAPH_H
#define REXD_LFIT_SUBSUMPTION_GRAPH_H

#include <symbolic/rules.h>
#include <symbolic/predicates.h>

#include <lfit/lfit_types.h>

namespace lfit {


struct Vertex{
    typedef boost::shared_ptr<Vertex> Ptr;
    std::list<Vertex::Ptr> adjacency_list;
    std::list<Vertex::Ptr> inverse_adjacency_list;
    RuleWithCoverage rulecov;
    
    Vertex(const Vertex& vertex):
        adjacency_list(vertex.adjacency_list),
        inverse_adjacency_list(vertex.inverse_adjacency_list),
        rulecov(vertex.rulecov)
    {}
    
    Vertex(const RuleWithCoverage& rule_with_cov):
        rulecov(rule_with_cov)
    { }
    
    Vertex::Ptr make_shared() const { return Vertex::Ptr(new Vertex(*this)); }
    
    friend bool operator<(const Vertex& l, const Vertex& r);
};

typedef std::map<RulePtr, Vertex::Ptr> VertexMap;


/**
 * \class RuleSubsumptionGraph
 * \brief Graph representing subsumption relationships between rule preconditions.
 * 
 * If the set of states covered by one rule contains the set of states contained by another rule, the first rule is subsumed of the second.
 */
class RuleSubsumptionGraph {
public:
    RuleSubsumptionGraph(const RuleWithCoverageSet& lfit_rules);
    
    RuleWithCoverageSet get_leafs() const;
    RuleWithCoverageSet get_leafs_and_first_parent() const;
    
    RuleWithCoverageSet get_rule_parents(const RuleWithCoverage& rule) const;
    
    void remove_leaf(const RuleWithCoverage& rule);
    bool remove_leafs_not_in(const RuleWithCoverageSet& ruleset);
    
    friend std::ostream& operator<<(std::ostream &out, const RuleSubsumptionGraph& graph);
    
private:
    VertexMap vertex_map;
    std::set<Vertex::Ptr> leafs;
    
    void add_vertex(const RuleWithCoverage& rule);
    void add_edge(const RuleWithCoverage& from, const RuleWithCoverage& to);
    void remove_vertex(const RuleWithCoverage& rule);
};

}

#endif
