/*
    <one line to give the program's name and a brief idea of what it does.>
    Copyright (C) 2013  dmartinez <email>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "explorer.h"
#include <boost/graph/graph_concepts.hpp>

#define DEBUG_EXPLORATION 0

Explorer::Explorer(const Scenario& scenario_, TransitionGroupConstPtr transitions_ptr_):
    scenario(scenario_),
    transitions_ptr(transitions_ptr_)
{ }


bool Explorer::is_rule_known(RuleConstPtr rule_ptr, const float visited_threshold) const
{
    int DEBUG = DEBUG_EXPLORATION;
    if (DEBUG > 0) {
        if (rule_ptr->dangerous)
            CINFO("explore") << "Not exploring rule " << rule_ptr->name << " as it is dangerous.";
    }
    return ((rule_ptr->dangerous) || 
            (rule_ptr->is_special()) || 
            (rule_ptr->count_covering_transitions_preconditions(*transitions_ptr) >= visited_threshold));
}


RulePtr Explorer::create_explore_rule(RuleConstPtr rule_ptr, const float rmax)
{
    RulePtr resulting_rule = rule_ptr->make_shared();
    Goal explore_goal(resulting_rule->get_preconditions(), rmax);
    resulting_rule->goals.push_back(explore_goal);
    return resulting_rule;
}


RuleSet Explorer::get_unknown_rules(const float visited_threshold) const
{
    RuleSet unknown_rules;
    BOOST_FOREACH(RuleConstPtr rule_ptr, scenario.get_rules()) {
        if (!is_rule_known(rule_ptr, visited_threshold)) {
            unknown_rules.push_back(rule_ptr->make_shared());
        }
    }
    return unknown_rules;
}


RuleSetPtr Explorer::get_rmax_explore_rules(const float visited_threshold, const float rmax) const
{
    int DEBUG = DEBUG_EXPLORATION;
    if (DEBUG > 0)
        CINFO("explore") << "\033[1;35m" << "Start" << "\033[0m";

    RuleSetPtr explore_rules = RuleSetPtr(new RuleSet());

    BOOST_FOREACH(const TypedAtom& action, scenario.get_action_list()) {
        RuleSet action_rules = scenario.get_rules().get_action_rules(action);
        if (!action_rules.empty()) {
            BOOST_FOREACH(RuleConstPtr action_rule_ptr, action_rules) {
                if (is_rule_known(action_rule_ptr, visited_threshold)) {
                    explore_rules->push_back(action_rule_ptr->make_shared());
                }
                else {
                    if (DEBUG > 0)
                        CINFO("explore") << "Creating explore rule";
                    explore_rules->push_back( create_explore_rule(action_rule_ptr, rmax) );
                }
            }
        }
        else {
            // create empty rule
            RulePtr action_rule_ptr = RulePtr(new SymbolicRule(action, PredicateGroup<TypedAtom>()));
            explore_rules->push_back( create_explore_rule(action_rule_ptr, rmax) );
        }
    }
    return explore_rules;
}


RulePtr Explorer::get_least_explored_rule(const TypedAtom& action, const int verbose, uint& min_rule_coverage) const
{
    assert(!action.is_special());
    bool avoid_dangerous_actions = scenario.config_reader.get_variable<bool>("confirm_dangerous_actions");
    RulePtr best_rule_ptr;
    
    RuleSet action_rules = scenario.get_rules().get_action_rules(action);
    RuleSet covering_rules = action_rules.get_covering_rules(scenario.get_state());
    if (covering_rules.size() >= 1) {
        CLOG_IF(verbose >=1, INFO, "explore") << "(There are many covering rules) " << NOENDL;
        for (RuleList::iterator rule_it = covering_rules.begin(); rule_it != covering_rules.end(); ++rule_it) {
            if ( (!avoid_dangerous_actions) || 
                 (!((*rule_it)->dangerous)) ) 
            {
                uint num_explorations = (*rule_it)->count_covering_transitions_preconditions(*transitions_ptr);
                if (num_explorations < min_rule_coverage) {
                    min_rule_coverage = num_explorations;
                    best_rule_ptr = boost::make_shared<SymbolicRule>(*static_cast<const SymbolicRule*>(&(**rule_it)));
                    CLOG_IF(verbose >=1, INFO, "explore") << "(Unexplored rule found for action: " << action << ", explored only " << num_explorations << " times) ";
                }
            }
        }
    }
    else if (covering_rules.size() == 0) {
        if (action_rules.empty()) { // Rules for this action have not been learned yet (e.g. noaction from LFIT-based learner)
            uint num_explorations = transitions_ptr->get_action_transitions(action).size_historial();
            if (num_explorations < min_rule_coverage) {
                CLOG_IF(verbose >=1, INFO, "explore") << "(Unexplored action without rules, not enough transitions for action " << action << ")";
                min_rule_coverage = num_explorations;
                best_rule_ptr = RulePtr(new SymbolicRule(action, PredicateGroup<TypedAtom>()));
            }
            else {
                CLOG_IF(verbose >=1, INFO, "explore") << "(Known rule, enough transitions for action " << action << ", but no rule yet)";
            }
        }
        else {
            // get action transitions  not covered by other action rules (action_rules not empty)
            TransitionGroup action_transitions = transitions_ptr->get_action_transitions(action);
            TransitionGroup not_covered_by_rule_transitions(transitions_ptr->ppddl_object_manager);
            for (TransitionGroup::TransitionHistorial::const_iterator action_transition_it = action_transitions.begin_historial();
                 action_transition_it != action_transitions.end_historial();
                 ++action_transition_it) {
                if (action_rules.covers_transition_preconditions(**action_transition_it, transitions_ptr->ppddl_object_manager) == 0.0) {
                    not_covered_by_rule_transitions.add_transition(**action_transition_it);
                }
            }
            if (not_covered_by_rule_transitions.size_historial() < min_rule_coverage) {
                CLOG_IF(verbose >=1, INFO, "explore") << "(Unexplored action without rule for current state, not enough transitions for action " << action << ")";
                min_rule_coverage = not_covered_by_rule_transitions.size_historial();
                best_rule_ptr = RulePtr(new SymbolicRule(action, PredicateGroup<TypedAtom>()));
            }
        }
    }
    return best_rule_ptr;
}


bool Explorer::is_state_known() const
{
    uint known_threshold = scenario.config_reader.get_variable<uint>("exploration_known_threshold");
    CINFO("explore") << "Checking if state is known... " << NOENDL;
    BOOST_FOREACH(const TypedAtom& action, scenario.get_action_list()) {
        if (!action.is_special()) {
            if (get_least_explored_rule(action, 1, known_threshold)) { // if rule is obtained, then it is unkown
                return false;
            }
        }
    }
    CINFO("explore") << " True.";
    return true;
}


TypedAtom::Opt Explorer::select_action_for_exploration(const bool forced, const uint visited_threshold) const
{
    int DEBUG = DEBUG_EXPLORATION;
    if (DEBUG > 0)
        CINFO("explore") << "\033[1;35m" << "Start" << "\033[0m";
    TypedAtom::Opt selected_action_opt;
    RulePtr best_rule_ptr;
    uint min_coverage = visited_threshold;
    if (forced) {
        min_coverage = std::numeric_limits< uint >::max(); // Any coverage is fine, doesn't have to be < visited_threshold
    }
    std::vector<TypedAtom> action_list_shuffle = scenario.get_action_list();
    std::random_shuffle ( action_list_shuffle.begin(), action_list_shuffle.end() );

    for (std::vector< TypedAtom >::const_iterator action_it = action_list_shuffle.begin(); action_it != action_list_shuffle.end(); ++action_it) {
        if (!action_it->is_special()) {
            RulePtr least_explored_rule = get_least_explored_rule(*action_it, 1, min_coverage);
            if (least_explored_rule) {
                best_rule_ptr = least_explored_rule;
            }
        }
    }

    TypedAtom::Opt action_to_explore;
    if (forced || (min_coverage < visited_threshold)) {
        if (best_rule_ptr) {
            CINFO("explore") << "Selecting rule " << best_rule_ptr->name << " with min coverage " << min_coverage << NOENDL;
            if (DEBUG > 2)
                CINFO("explore") << "\033[1;35m" << "Best rule is" << *best_rule_ptr << "\033[0m";
            RuleSet grounded_rules = (*static_cast<const SymbolicRule*>(&(*best_rule_ptr))).get_pddl_grounded_rules(scenario.get_state(), false);
            std::random_shuffle ( grounded_rules.begin(), grounded_rules.end() );
            assert(!grounded_rules.empty());
            action_to_explore = TypedAtom::Opt(*grounded_rules.front());
            CINFO("explore") << " (best grounding: " << *action_to_explore << ")";
        }
        else {
            CINFO("explore") << "\033[1;35m" << "There is no best rule" << "\033[0m";
            if (selected_action_opt) {
                action_to_explore = TypedAtom::Opt(*selected_action_opt);
                std::vector<PPDDL::Type> param_types = action_to_explore->get_param_types();
                PPDDLObjectManager::CombinationsPPDDLObjects combinations_of_params = *scenario.get_state().ppddl_object_manager.get_combinations_of_objecs_for_given_params(param_types);
                assert(!combinations_of_params.empty());
                std::random_shuffle ( combinations_of_params.begin(), combinations_of_params.end() );
                std::vector<PPDDL::Object> random_params;
                for (size_t param_idx = 0; param_idx < param_types.size(); ++param_idx) {
                    random_params.push_back(PPDDL::Object(combinations_of_params.front()[param_idx], param_types[param_idx]));
                }
                action_to_explore->ground(random_params);
                CINFO("explore") << " (random grounding for action " << *selected_action_opt << " is " << *action_to_explore << " )";
            }
            else { // no safe action to explore, explore a random grounded action
                RuleSet grounded_rules = scenario.get_rules().get_pddl_grounded_rules(scenario.get_state());
                grounded_rules.remove_special_rules();
                if (grounded_rules.size() >0) {
                    CINFO("explore") << " (random grounding)";
                    std::random_shuffle ( grounded_rules.begin(), grounded_rules.end() );
                    action_to_explore = TypedAtom::Opt(*grounded_rules.front());
                }
                else {
                    CINFO("explore") << " (no action available!!)";
                    action_to_explore = TypedAtom::Opt();
                }
            }
        }
    }
    return action_to_explore;
}

