/*
 * REX_D scenario class to work with a PPDDL representation
 * Copyright (C) 2014  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "ppddl_scenario.h"

PPDDLScenario::PPDDLScenario(const RexdScenario& rexd_scenario): 
    RexdScenario(rexd_scenario)
{ }

PPDDLScenario::PPDDLScenario(const Scenario& scenario): 
    RexdScenario(scenario)
{ }


void PPDDLScenario::prepare_IPPC_PPDDL_rules_and_state(RuleSetPtr& complete_rules_ptr, StatePtr& state_ppddl_ptr) const
{
    using namespace boost::lambda;
    
    complete_rules_ptr = boost::make_shared<RuleSet>(rules);
    state_ppddl_ptr = boost::make_shared<State>(state);
    
    // RMAX explore actions
    if (use_rmax) {
        Explorer explorer(*this, transitions_ptr);
        complete_rules_ptr = explorer.get_rmax_explore_rules(explore_threshold, rmax_reward);
    }
    
    complete_rules_ptr->remove_nop_rules(); // cannot do below, would remove min reward teacher action from rmax_vmin

    // V-MIN teacher action
    if (use_vmin) {
        // -started() required so teacher can just be called as the first rule
        complete_rules_ptr->add_predicate_to_outcomes(Predicate<TypedAtom>(MAKE_TYPED_ATOM_STARTED, true));
        state_ppddl_ptr->update_predicate(Predicate<TypedAtom>(MAKE_TYPED_ATOM_STARTED, false));
        // V-MIN Teacher action
        RulePtr rule_ppddl_teacher = create_rule_IPPC_PPDDL_teacher(vmin_value);
        complete_rules_ptr->push_back(rule_ppddl_teacher);
    }
    
    if (config_reader.get_variable<bool>("final_reward_action")) {
        // add current_reward to final_reward_action as it was already obtained
        RuleSet::iterator finish_action_rule_it = std::find_if(complete_rules_ptr->begin(), complete_rules_ptr->end(), *_1 == MAKE_TYPED_ATOM_FINISH_ACTION);
        assert(finish_action_rule_it != complete_rules_ptr->end()); // there should be always a finish_action
        assert(*finish_action_rule_it); // there should be always a finish_action
        assert(!(*finish_action_rule_it)->goals.empty()); // there should be always a finish_action with goals
        (*finish_action_rule_it)->goals.front().set_reward( (*finish_action_rule_it)->goals.front().get_reward() + current_reward);
    }
    
    complete_rules_ptr = boost::make_shared<RuleSet>(complete_rules_ptr->get_pddl_grounded_rules(state));
    complete_rules_ptr->PPDDL_grounding();
}


void PPDDLScenario::create_IPPC_PPDDL_scenario(RuleSetPtr& rules_ppddl_ptr, StatePtr& state_ppddl_ptr, GoalGroupPtr& goals_ppddl_ptr, DomainPtr& domain_ppddl_ptr) const
{
    prepare_IPPC_PPDDL_rules_and_state(rules_ppddl_ptr, state_ppddl_ptr);

    // get PPDDL grounded goal
    goals_ppddl_ptr = boost::make_shared<GoalGroup>(goals);
    goals_ppddl_ptr->PPDDL_grounding();

    // get PPDDL domain.
    domain_ppddl_ptr = DomainPtr(new Domain(config_reader));
    domain_ppddl_ptr->index_rules(*rules_ppddl_ptr);
    domain_ppddl_ptr->index_predicates(goals_ppddl_ptr->get_predicates());

    //optional: adding state. Info about the rules may be enough (exception if excuses available)
    state_ppddl_ptr = boost::make_shared<State>(state); // copy
    state_ppddl_ptr->PPDDL_grounding();
    domain_ppddl_ptr->index_state(*state_ppddl_ptr);

    // add special rules
    RulePtr rule_ppddl_noop = create_rule_IPPC_PPDDL_noop(domain_ppddl_ptr, goals_ppddl_ptr, noop_reward);
    rules_ppddl_ptr->push_back(rule_ppddl_noop);
}


void PPDDLScenario::write_IPPC_PPDDL(const std::string template_file, std::ostream &out) const
{
    RuleSetPtr rules_ppddl_ptr;
    StatePtr state_ppddl_ptr;
    GoalGroupPtr goal_ppddl_ptr;
    DomainPtr domain_ppddl_ptr;
    create_IPPC_PPDDL_scenario(rules_ppddl_ptr, state_ppddl_ptr, goal_ppddl_ptr, domain_ppddl_ptr);

    // write init and domain
    out << utils::get_file_contents(template_file, "<init>" , "<pre_domain>");
    domain_ppddl_ptr->write_PPDDL_grounded_predicates(out);

    // write rules
    std::map<std::string, RuleSet> rules_per_action = rules_ppddl_ptr->get_rules_per_action();
    write_actions_IPPC_PPDDL(out, rules_per_action, *goal_ppddl_ptr);

    // write problem vars and end
    out << utils::get_file_contents(template_file, "<post_rules>" , "<horizon>");
    out << planning_horizon;
    out << utils::get_file_contents(template_file, "<horizon>" , "<end>");
    out << std::endl;
}


void PPDDLScenario::write_actions_IPPC_PPDDL(std::ostream& out, std::map<std::string, RuleSet> &rules_per_action, const GoalGroup& goals_ppddl) const
{
    // TODO HACK IMPROVE!!!

    // Several actions
    for (std::map<std::string, RuleSet>::iterator action_it = rules_per_action.begin(); action_it != rules_per_action.end(); ++action_it) {
        GoalGroup rule_goals;
        out << "  (:action " << action_it->first << std::endl;
        out << "    :effect (and" << std::endl;

        // Action has several rules
        for (RuleSet::iterator rule_it = action_it->second.begin(); rule_it != action_it->second.end(); ++rule_it) {
            // Action has several outcomes
            if ((*rule_it)->has_useful_outcome()) {
                if (!(*rule_it)->get_preconditions().empty()) {
                    out << "      (when ";
                    (*rule_it)->get_preconditions().write_PPDDL(out, (*rule_it)->params);
                }
                (*rule_it)->outcomes.write_PPDDL(out, (*rule_it)->params);
                if (!(*rule_it)->get_preconditions().empty()) {
                    out << ")" << std::endl;
                }
            }

            if ((*rule_it)->goals.size() > 0) { // use rule goals
                rule_goals.insert(rule_goals.end(), (*rule_it)->goals.begin(), (*rule_it)->goals.end());
            }
        }
//         utils::remove_duplicates(rule_goals);
        rule_goals.write_PPDDL(out, std::vector<PPDDL::Object>());
        if ((action_it->first != ACTION_NAME_FINISH) && (action_it->first != ACTION_NAME_NOOP))
            goals_ppddl.write_PPDDL(out, std::vector<PPDDL::Object>());
        out << "    )" << std::endl;
        out << "  )" << std::endl;
    }
}


RulePtr PPDDLScenario::create_rule_IPPC_PPDDL_noop(const DomainPtr domain_ptr, const GoalGroupPtr goals_ppddl_ptr, const float noop_reward)
{
    RulePtr rule_noop_ptr = RulePtr(new SymbolicRule(MAKE_TYPED_ATOM_NOOP_ACTION, PredicateGroup<TypedAtom>()));

    // obtain domain positive and negative predicates
    PredicateGroup<TypedAtom> positive_known_predicates, negated_known_predicates;
    BOOST_FOREACH(const Predicate<TypedAtom>& domain_pred, domain_ptr->get_predicates()) {
        Predicate<TypedAtom> predicate(domain_pred);
        predicate.positive = true;
        positive_known_predicates.insert(predicate);
        predicate.positive = false;
        negated_known_predicates.insert(predicate);
    }
    PredicateGroup<TypedAtom> rule_noop_preconditions = rule_noop_ptr->get_preconditions();
    rule_noop_preconditions.add_predicates(positive_known_predicates);
    rule_noop_preconditions.add_predicates(negated_known_predicates);
    rule_noop_ptr->set_preconditions(rule_noop_preconditions);
    Outcome outcome_pos(positive_known_predicates, 0.5);
    Outcome outcome_neg(negated_known_predicates, 0.5);
    rule_noop_ptr->add_outcome(outcome_pos);
    rule_noop_ptr->add_outcome(outcome_neg);

    rule_noop_ptr->goals = *goals_ppddl_ptr;
    Goal noop_goal(noop_reward); // noop reward is small
//     rule_noop_ptr->goals.push_back(goal_ppddl_ptr);
    rule_noop_ptr->goals.push_back(noop_goal);

    return rule_noop_ptr;
}



RulePtr PPDDLScenario::create_rule_IPPC_PPDDL_teacher(const float min_reward)
{
    PredicateGroup<TypedAtom> rule_teacher_preconditions;
    rule_teacher_preconditions.insert(Predicate<TypedAtom>(MAKE_TYPED_ATOM_STARTED, false)); // not started()
    rule_teacher_preconditions.insert(Predicate<TypedAtom>(MAKE_TYPED_ATOM_FINISHED, false)); // not finished()
    Outcome rule_teacher_effect(1.0);
    rule_teacher_effect.insert(Predicate<TypedAtom>(MAKE_TYPED_ATOM_STARTED, true)); // started()
    rule_teacher_effect.insert(Predicate<TypedAtom>(MAKE_TYPED_ATOM_FINISHED, true)); // finished()
    
    RulePtr rule_teacher_ptr = RulePtr(new SymbolicRule(MAKE_TYPED_ATOM_TEACHER, rule_teacher_preconditions));
    rule_teacher_ptr->outcomes.insert(rule_teacher_effect);
    Goal teacher_goal(rule_teacher_preconditions, min_reward);
    rule_teacher_ptr->goals.push_back(teacher_goal);
    return rule_teacher_ptr;
}

void PPDDLScenario::write_IPPC_PPDDL_to_file(const std::string template_file, const std::string file_path) const
{
    int DEBUG = 0;
    std::ofstream rules_file(file_path.c_str());
    if (DEBUG > 0)
        std::cout << "RuleSet::write_rules_to_file - " << "Writting rules to file: '" << file_path << "'" << std::endl;
    if (rules_file.is_open()) {
        write_IPPC_PPDDL(template_file, rules_file);
    } else
        std::cerr << "Couldn't write rules to " << file_path << std::endl;
    rules_file.close();
}

