/*
 * REX_D scenario class to work with a RDDL representation
 * Copyright (C) 2015  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "rddl_scenario.h"

using namespace RDDL;

RDDLScenario::RDDLScenario(const RexdScenario& rexd_scenario): 
    RexdScenario(rexd_scenario)
{ }

RDDLScenario::RDDLScenario(const Scenario& scenario): 
    RexdScenario(scenario)
{ }


void RDDLScenario::prepare_IPPC_RDDL_rules_and_state(RuleSetPtr& complete_rules_ptr, StatePtr& state_ppddl_ptr) const
{
    // TODO move to RexdScenario
    
    using namespace boost::lambda;
    
    complete_rules_ptr = boost::make_shared<RuleSet>(rules);
    state_ppddl_ptr = boost::make_shared<State>(state);
    
    // RMAX explore actions
    if (use_rmax) {
        Explorer explorer(*this, transitions_ptr);
        complete_rules_ptr = explorer.get_rmax_explore_rules(explore_threshold, rmax_reward);
    }
    
    complete_rules_ptr->remove_nop_rules(); // cannot do below, would remove min reward teacher action from rmax_vmin

    // V-MIN teacher action
    if (use_vmin) {
        // -started() required so teacher can just be called as the first rule
        complete_rules_ptr->add_predicate_to_outcomes(Predicate<TypedAtom>(MAKE_TYPED_ATOM_STARTED, true));
        state_ppddl_ptr->update_predicate(Predicate<TypedAtom>(MAKE_TYPED_ATOM_STARTED, false));
        // V-MIN Teacher action
        RulePtr rule_ppddl_teacher = create_rule_IPPC_RDDL_teacher(vmin_value);
        complete_rules_ptr->push_back(rule_ppddl_teacher);
    }
    
    if (config_reader.get_variable<bool>("final_reward_action")) {
        // add current_reward to final_reward_action as it was already obtained
        RuleSet::iterator finish_action_rule_it = std::find_if(complete_rules_ptr->begin(), complete_rules_ptr->end(), *_1 == MAKE_TYPED_ATOM_FINISH_ACTION);
        assert(finish_action_rule_it != complete_rules_ptr->end()); // there should be always a finish_action
        assert(*finish_action_rule_it); // there should be always a finish_action
        assert(!(*finish_action_rule_it)->goals.empty()); // there should be always a finish_action with goals
        (*finish_action_rule_it)->goals.front().set_reward( (*finish_action_rule_it)->goals.front().get_reward() + current_reward);
    }
}


void RDDLScenario::create_IPPC_RDDL_scenario(RuleSetPtr& rules_ppddl_ptr, StatePtr& state_ppddl_ptr, GoalGroupPtr& goals_ppddl_ptr, DomainPtr& domain_ppddl_ptr) const
{
    prepare_IPPC_RDDL_rules_and_state(rules_ppddl_ptr, state_ppddl_ptr);

    // get RDDL grounded goal
    goals_ppddl_ptr = boost::make_shared<GoalGroup>(goals);

    // get RDDL domain.
    domain_ppddl_ptr = DomainPtr(new Domain(config_reader));
    domain_ppddl_ptr->index_rules(*rules_ppddl_ptr);
    domain_ppddl_ptr->index_predicates(goals_ppddl_ptr->get_predicates());

    //optional: adding state. Info about the rules may be enough (exception if excuses available)
    state_ppddl_ptr = boost::make_shared<State>(state); // copy
    domain_ppddl_ptr->index_state(*state_ppddl_ptr);

    // add special rules
    RulePtr rule_ppddl_noop = create_rule_IPPC_RDDL_noop(domain_ppddl_ptr, goals_ppddl_ptr, noop_reward);
    rules_ppddl_ptr->push_back(rule_ppddl_noop);
}


void RDDLScenario::obtain_rddl_symbolic_transition(Predicate<TypedAtom>& out_predicate,
                                                   PredicateGroup<TypedAtom>& preconditions,
                                                   TypedAtom& action) const
{
    std::vector<PPDDL::Object> head_params_v = out_predicate.params;
    std::vector<PPDDL::Object> body_params_V = preconditions.get_params();
    std::set<PPDDL::Object> head_params_s;
    std::set<PPDDL::Object> body_params_s;
    std::set<PPDDL::Object> deictic_params_s;
    head_params_s.insert(head_params_v.begin(), head_params_v.end());
    body_params_s.insert(body_params_V.begin(), body_params_V.end());
    // deictic_params_s = body_params_s - head_params_s;
    std::set_difference(body_params_s.begin(), body_params_s.end(), 
                        head_params_s.begin(), head_params_s.end(), 
                        std::inserter(deictic_params_s, deictic_params_s.end()));
    
    std::vector<PPDDL::Object> original_params;
    original_params.insert(original_params.end(), head_params_s.begin(), head_params_s.end());
    original_params.insert(original_params.end(), deictic_params_s.begin(), deictic_params_s.end());
    std::vector<PPDDL::Object> new_params(original_params);
    symbolic::make_vector_params_symbolic(new_params);
    
    // two step renaming so there are no overwrites
    for (size_t i = 0; i < new_params.size(); ++i) {
        out_predicate.change_param_name(original_params[i].name, new_params[i].name + "aux");
        preconditions.change_param_name(original_params[i].name, new_params[i].name + "aux");
        action.change_param_name(original_params[i].name, new_params[i].name + "aux");
    }
    for (size_t i = 0; i < new_params.size(); ++i) {
        out_predicate.change_param_name(new_params[i].name + "aux", new_params[i].name);
        preconditions.change_param_name(new_params[i].name + "aux", new_params[i].name);
        action.change_param_name(new_params[i].name + "aux", new_params[i].name);
    }
}


MapHeadBody RDDLScenario::get_rddl_transitions_from_rules(const RuleSet& rules) const
{
    MapHeadBody result;
    BOOST_FOREACH(const RulePtr& rule_ptr, rules) {
        BOOST_FOREACH(const Outcome& outcome, rule_ptr->outcomes) {
            BOOST_FOREACH(const Predicate<TypedAtom>& out_predicate, outcome) {
                Predicate<TypedAtom> new_out_predicate(out_predicate);
                PredicateGroup<TypedAtom> new_preconditions(rule_ptr->get_preconditions());
                TypedAtom new_action(*rule_ptr);
                obtain_rddl_symbolic_transition(new_out_predicate, new_preconditions, new_action);
                
                PairPreconditionsAction new_pair_preconditions_action(Outcome(new_preconditions, outcome.probability),
                                                                      new_action);
                if (new_out_predicate.positive) {
                    result[new_out_predicate].first.push_back(new_pair_preconditions_action);
                }
                else {
                    result[new_out_predicate].second.push_back(new_pair_preconditions_action);
                }
            }
        }
    }
    return result;
}


void RDDLScenario::write_IPPC_RDDL(const std::string template_file, std::ostream &out) const
{
    RuleSetPtr rules_ppddl_ptr;
    StatePtr state_ppddl_ptr;
    GoalGroupPtr goal_ppddl_ptr;
    DomainPtr domain_ppddl_ptr;
    create_IPPC_RDDL_scenario(rules_ppddl_ptr, state_ppddl_ptr, goal_ppddl_ptr, domain_ppddl_ptr);

    // write init and domain
    out << utils::get_file_contents(template_file, "<init>" , "<pre_rules>");
//     domain_ppddl_ptr->write_RDDL_objects(out, state_ppddl_ptr->pddl_object_manager);
//     domain_ppddl_ptr->write_RDDL_symbolic_predicates(out, state_ppddl_ptr->pddl_object_manager);

    // write rules
    MapHeadBody map_head_bodies = get_rddl_transitions_from_rules(*rules_ppddl_ptr);
    // get also symbols that do not change
    std::set<TypedAtom> non_changing_heads;
    non_changing_heads.insert(domain_ppddl_ptr->get_symbolic_atoms().begin(), domain_ppddl_ptr->get_symbolic_atoms().end());
    BOOST_FOREACH(const MapHeadBody::value_type& pair_head_body, map_head_bodies) {
        non_changing_heads.erase(pair_head_body.first);
    }
    // remove constant predicates
    BOOST_FOREACH(Predicate<TypedAtom> constant_predicate, state.get_constants()) {
        constant_predicate.make_symbolic();
        non_changing_heads.erase(constant_predicate);
    }
    write_actions_IPPC_RDDL(out, map_head_bodies, non_changing_heads);
    
    write_reward_IPPC_RDDL(out, *goal_ppddl_ptr);

    out << utils::get_file_contents(template_file, "<post_rules>" , "<end>");
    out << std::endl;
}

void RDDLScenario::write_RDDL_transition(std::ostream& out, 
                                         const PairPreconditionsAction& pair_prec_action, 
                                         const std::vector<PPDDL::Object>& head_params,
                                         const bool positive_outcome) const
{
    std::vector<PPDDL::Object> body_params_v(pair_prec_action.first.get_params());
    std::set<PPDDL::Object> head_params_s;
    std::set<PPDDL::Object> body_params_s;
    std::set<PPDDL::Object> deictic_params_s;
    head_params_s.insert(head_params.begin(), head_params.end());
    body_params_s.insert(body_params_v.begin(), body_params_v.end());
    // deictic_params_s = body_params_s - head_params_s;
    std::set_difference(body_params_s.begin(), body_params_s.end(), 
                        head_params_s.begin(), head_params_s.end(), 
                        std::inserter(deictic_params_s, deictic_params_s.end()));
    
    out << "if (^ ";
    if (pair_prec_action.second != MAKE_TYPED_ATOM_NOACTION) {
        pair_prec_action.second.write_RDDL(out);
    }
    if (!deictic_params_s.empty()) {
        out << " (exists ( ";
        BOOST_FOREACH(const PPDDL::Object& deictic_param, deictic_params_s) {
            out << "( " << deictic_param.name << " : " << deictic_param.type << ") ";
        }
        out << " ) (^ ";
    }
    BOOST_FOREACH(const Predicate<TypedAtom>& body_predicate, pair_prec_action.first) {
        out << " ";
        body_predicate.write_RDDL(out);
    }
    if (!deictic_params_s.empty()) {
        out << "))";
    }
    out << " )";
    if (positive_outcome) {
        if (pair_prec_action.first.probability >= 1.0) {
            out << " then (KronDelta true) ";
        }
        else {
            out << " then (Bernoulli (" << pair_prec_action.first.probability << ")) ";
        }
    }
    else {
        if (pair_prec_action.first.probability == 1.0) {
            out << " then (KronDelta false) ";
        }
        else {
            out << " then (Bernoulli (" << (1.0 - pair_prec_action.first.probability) << ")) ";
        }
    }
}

void RDDLScenario::write_actions_IPPC_RDDL(std::ostream& out, const MapHeadBody &map_head_bodies, const std::set<TypedAtom>& other_heads) const
{
    out << "cpfs {" << std::endl;
    BOOST_FOREACH(const MapHeadBody::value_type& pair_head_bodies, map_head_bodies) {
        out << "    ";
        pair_head_bodies.first.write_RDDL(out, true);
        if ( (!pair_head_bodies.second.first.empty()) || (!pair_head_bodies.second.second.empty()) ) {
            out << " = (";
            bool first = true;
            uint number_else = 0;
            BOOST_FOREACH(const PairPreconditionsAction& pair_prec_action, pair_head_bodies.second.first) {
                // positive preconditions
                if (!first) {
                    out << " else (";
                }
                write_RDDL_transition(out, pair_prec_action, pair_head_bodies.first.params, true);
                if (!first) {
                    number_else++;
                }   
                first = false;
            }
            BOOST_FOREACH(const PairPreconditionsAction& pair_prec_action, pair_head_bodies.second.second) {
                // negative preconditions
                if (!first) {
                    out << " else (";
                }
                write_RDDL_transition(out, pair_prec_action, pair_head_bodies.first.params, false);
                if (!first) {
                    number_else++;
                }
                first = false;
            }
            out << " else (KronDelta ";
            if (config_reader.get_variable<bool>("variable_maintains_to_next_state")) {
                pair_head_bodies.first.write_RDDL(out, false);
            }
            else {
                out << "false";
            }
            out << ")";
            for (size_t i = 0; i < number_else; ++i) {
                out << ")";
            }
            out << ");" << std::endl;
        }
        else {
            out << " (KronDelta ";
            pair_head_bodies.first.write_RDDL(out, false);
            out << ");" << std::endl;
        }
    }
    BOOST_FOREACH(const TypedAtom& other_head, other_heads) {
        out << "    ";
        other_head.write_RDDL(out, true);
        out << " = (KronDelta ";
        other_head.write_RDDL(out, false);
        out << ");" << std::endl;
    }
    out << "  };" << std::endl;
}

void RDDLScenario::write_reward_IPPC_RDDL(std::ostream& out, const GoalGroup& goals_ppddl) const
{
    LOG(WARNING) << "TODO!!! WRITE REWARD IN RDDL";
}


RulePtr RDDLScenario::create_rule_IPPC_RDDL_noop(const DomainPtr domain_ptr, const GoalGroupPtr goals_ppddl_ptr, const float noop_reward)
{
    RulePtr rule_noop_ptr = RulePtr(new SymbolicRule(MAKE_TYPED_ATOM_NOOP_ACTION, PredicateGroup<TypedAtom>()));

    return rule_noop_ptr;
}



RulePtr RDDLScenario::create_rule_IPPC_RDDL_teacher(const float min_reward)
{
    PredicateGroup<TypedAtom> rule_teacher_preconditions;
    rule_teacher_preconditions.insert(Predicate<TypedAtom>(MAKE_TYPED_ATOM_STARTED, false)); // not started()
    rule_teacher_preconditions.insert(Predicate<TypedAtom>(MAKE_TYPED_ATOM_FINISHED, false)); // not finished()
    Outcome rule_teacher_effect(1.0);
    rule_teacher_effect.insert(Predicate<TypedAtom>(MAKE_TYPED_ATOM_STARTED, true)); // started()
    rule_teacher_effect.insert(Predicate<TypedAtom>(MAKE_TYPED_ATOM_FINISHED, true)); // finished()
    
    RulePtr rule_teacher_ptr = RulePtr(new SymbolicRule(MAKE_TYPED_ATOM_TEACHER, rule_teacher_preconditions));
    rule_teacher_ptr->outcomes.insert(rule_teacher_effect);
    Goal teacher_goal(rule_teacher_preconditions, min_reward);
    rule_teacher_ptr->goals.push_back(teacher_goal);
    return rule_teacher_ptr;
}

void RDDLScenario::write_IPPC_RDDL_to_file(const std::string& domain_template_file, const std::string& domain_file_path, const std::string& file_path2) const
{
    int DEBUG = 0;
    std::ofstream rules_file(domain_file_path.c_str());
    if (DEBUG > 0)
        CINFO("planning") << "RuleSet::write_rules_to_file - " << "Writting rules to file: '" << domain_file_path << "'";
    if (rules_file.is_open()) {
        write_IPPC_RDDL(domain_template_file, rules_file);
    } else
        std::cerr << "Couldn't write rules to " << domain_file_path << std::endl;
    rules_file.close();
}

