#include "rl_main.h"
#include "rexd/pasula_learner.h"
#include "lfit/lfit_learner.h"

_INITIALIZE_EASYLOGGINGPP

ReinforcementLearner::ReinforcementLearner():
    transitions(PPDDLObjectManager()),
    transition_obtained(false),
    accumulated_reward(0),
    current_problem(1),
    current_episode(1),
    current_episode_iteration(0),
    total_number_iterations(0)
{
}

bool ReinforcementLearner::init(const std::string& config_path, const std::string& files_path)
{
    bool init_ok = config_reader.init(config_path, files_path);
    utils::init_loggers(files_path + "logger.conf");
    LOG(INFO) << "Starting new execution.\n\n";

    // generated random seed
    utils::initialize_random_seed();
    
    scenario_ptr = ScenarioPtr(new Scenario(config_reader.get_path("ppddl_domain_file"), 
                                            config_reader.get_path("simulator_ppddl_current_problem_file"),
                                            config_reader));
    
    // load action list if needed
    std::vector<TypedAtom> action_list;
    if (!config_reader.get_variable<bool>("teacher")) { 
        // if no teacher, the list of actions have to be provided to learn
        // TODO
        LOG(ERROR) << "TODO: Implement and check reading an action list with typed actions.";
        assert(0);
        // read_rules
        // get action_list from rules
        LOG(INFO) << "Action list is " << action_list;
    }
    
    transitions = read_transitions_from_file(config_reader.get_path("transitions_backup_path"), 
                                             scenario_ptr->get_state().ppddl_object_manager, 
                                             scenario_ptr->get_state().get_constants(),
                                             config_reader.get_variable<bool>("variable_maintains_to_next_state")
                                            );
    scenario_ptr->domain.index_transitions(transitions);
    
    // TODO read saved dead end excuses from previous execution
//     dead_end_excuses = PredicateGroup<Atom>("-broken(71) -broken(72) -broken(73) -broken(74) -broken(75) -broken(76) broken(71) broken(72) broken(73) broken(74) broken(75) broken(76) ");
//     rules_ptr->set_dangerous_rules(dead_end_excuses);

    // load planner and learner
    high_level_planner = boost::make_shared<HighLevelPlanner>(HighLevelPlanner(config_reader, &logging_file));
    if (config_reader.get_variable<std::string>("learner") == LEARNER_PASULA) {
        learner = boost::make_shared<PasulaLearner>(PasulaLearner(config_reader));
    }
    else if (config_reader.get_variable<std::string>("learner") == LEARNER_LFIT) {
        learner = boost::make_shared<LFITLearner>(LFITLearner(config_reader));
    }
    else {
        throw std::invalid_argument("Initializing learner: Unknown learner " + config_reader.get_variable<std::string>("learner"));
    }

    // load auto teacher if needed
    if (config_reader.get_variable<bool>("auto_teacher")) {
        auto_teacher_ptr = AutoTeacherPtr(new AutoTeacher(config_reader));
    }

    visualizer.init(config_reader);

    return init_ok;
}

bool ReinforcementLearner::main_iteration()
{
    if (config_reader.get_variable<bool>("monitoring")) {
        return main_monitoring_iteration();
    }
    else {
        return main_learning_iteration();
    }
}


bool ReinforcementLearner::reset_rexd()
{
    bool result;
    // total_number_iterations--; // special iteration, doesn't count as normal iteration -> anyway for visualization is fine to have a small pause
    logging_file << "Reward: " << accumulated_reward << std::endl;

    result = reset_scenario();
    scenario_ptr->reset();

    visualizer.write_episode_log(config_reader.get_variable<std::string>("domain"), current_problem, current_episode, accumulated_reward, total_number_iterations);
    accumulated_reward = 0;
    current_episode_iteration = 0;
    high_level_planner->reset();
    return result;
}


bool ReinforcementLearner::main_learning_iteration()
{
    LOG(INFO) << "\n" << "--------------------------------------------------------#----------------------------------------------------"; // 2 newlines
    // new iteration
    total_number_iterations++;
    current_episode_iteration++;

    // get state
    show_info_request_state();
    previous_state = scenario_ptr->get_state();
    scenario_ptr->set_state(read_state());

    // get reward
    last_reward = read_reward();
    accumulated_reward += last_reward;

    // check infinite loops
    if (current_episode_iteration > 100) { // If this episode > 100 iterations, suppose things have gone wrong
        transition_obtained = false;
        logging_file << "Dead." << std::endl;
        return reset_rexd();
    }

    // check goal
    if (scenario_ptr->is_in_goal()) {
        if (transition_obtained) { // learn last action when arriving to the goal
            learning();
        }
        LOG(INFO) << START_COLOR_GREEN << "IS GOAL! Reward: " << accumulated_reward << END_COLOR;
        transition_obtained = false;
        show_info_is_goal();
        return reset_rexd();
    }
    else { // normal behaviour (learn and plan)
        learning();
        return planning(); // true if OK, false if dead-end
    }
}


// WARNING This method is a infinite loop
bool ReinforcementLearner::main_monitoring_iteration()
{
    LOG(INFO) << "\n" << "********************************************************"; // 2 newlines

    // get state
    show_info_request_state();
    previous_state = scenario_ptr->get_state();
    scenario_ptr->set_state(read_state());
    read_reward();

    // check goal
    if (scenario_ptr->is_in_goal()) {
        LOG(INFO) << START_COLOR_GREEN << "IS GOAL!" << END_COLOR;
        transition_obtained = false;
        show_info_is_goal();

        // wait for action recognition
        read_action();

        return true;
    }
    else { // normal behaviour (learn and plan)
        monitoring();

        return true;
    }
}


void ReinforcementLearner::update_transitions()
{
    if (transition_obtained) {
        State current_state = scenario_ptr->get_state();
        current_state.remove_special_predicates();
        previous_state.remove_special_predicates();
        
        TypedAtom typed_previous_action = scenario_ptr->get_state().ppddl_object_manager.get_typed_atom(*previous_action);
        Transition transition_full(previous_state, 
                                   typed_previous_action, 
                                   current_state, 
                                   last_reward, 
                                   config_reader.get_variable<bool>("variable_maintains_to_next_state"));
        transitions.add_transition(transition_full);
        transitions.write_pasula_transitions_to_file(config_reader.get_path("transitions_backup_path"));
    }
}

void ReinforcementLearner::learning()
{
    update_transitions();

    struct timeval timer_start = utils::start_timer();
    if ((config_reader.get_variable<bool>("learning")) && (transition_obtained) && (transitions.size_set() > 0) && (previous_action->name.size()>1) ) {
        CINFO("planning") << START_COLOR_BLUE << "Learning" << END_COLOR;
        show_info_start_learning();
        RuleSetPtr new_rules_ptr, removed_rules_ptr, added_rules_ptr;
        TypedAtom typed_previous_action = scenario_ptr->get_state().ppddl_object_manager.get_typed_atom(*previous_action);
        if (config_reader.get_variable<std::string>("learner") == LEARNER_PASULA) {
            new_rules_ptr = learner->learn_action_rules(scenario_ptr->get_rules(),
                                                        transitions, 
                                                        typed_previous_action, 
                                                        removed_rules_ptr, added_rules_ptr);
        }
        else if (config_reader.get_variable<std::string>("learner") == LEARNER_LFIT) {
            new_rules_ptr = learner->learn_quiet(transitions);
        }
        else {
            throw std::invalid_argument("Learning rules: Unknown learner " + config_reader.get_variable<std::string>("learner"));
        }
        new_rules_ptr->set_dangerous_rules(dead_end_excuses);
        scenario_ptr->set_rules(*new_rules_ptr);
        scenario_ptr->write_PDDL_domain(config_reader.get_path("ppddl_domain_file"));

        // TODO re-enable this part (if using LFIT removed and added rules have to be found manually)
//         show_info_end_learning(*removed_rules_ptr, *added_rules_ptr, high_level_planner->transition_context);
//         visualizer.write_learning_log(*removed_rules_ptr, *added_rules_ptr, high_level_planner->transition_context, total_number_iterations);
    }
    else {
        CINFO("planning") << "Skipping learning (no transitions or no previous action)";
    }
    logging_file << "Time learning: " << utils::get_elapsed_time(timer_start) << std::endl;
}


bool ReinforcementLearner::planning()
{
    show_info_start_planning();
    struct timeval timer_start = utils::start_timer();
    transition_obtained = true; // default
    expected_reward = 0; // default
    PlanningFailureOpt excuse_opt;
    std::vector< std::string > actions_requiring_excuse;
    CINFO("planning") << START_COLOR_BLUE << "Planning" << END_COLOR;

    // explore and plan
    bool is_dangerous;
    TypedAtom::Opt planning_res = high_level_planner->exploit_or_explore(*scenario_ptr, transitions, excuse_opt, actions_requiring_excuse, is_dangerous);
    if (is_dangerous) {
        planning_res = ask_confirmation_dangerous_action(scenario_ptr->get_state(), planning_res); // transition_context set inside if teacher changes action
    }
    logging_file << "Time planning: " << utils::get_elapsed_time(timer_start) << std::endl;
    show_info_end_planning(planning_res, failure_analyzer_utils::planning_failure_to_string(excuse_opt));

    // If solution found execute it. Else teacher / explore
    if (planning_res) {
        execute_action(*planning_res);
    }
    else {
        // All trials to obtain a plan have failed, so it will request teacher demonstration 
        // Subgoals approach the goal as much as possible before requesting a teacher demonstration
        // After first teacher demonstration is done, subgoals no longer make sense, so the are disabled
        // This avoid loops with subgoals and auto_teacher
        // The same applies to random exploration
        high_level_planner->disable_subgoals();
        
        if (config_reader.get_variable<bool>("teacher")) {
            bool continue_ok = teacher(excuse_opt, actions_requiring_excuse);
            if (!continue_ok) {
                return false;
            }
        }
        else {
            planning_res = high_level_planner->explore(*scenario_ptr, transitions, true);
            execute_action(*planning_res);
        }
    }
    if ( planning_res && planning_res->is_special() ) {
        transition_obtained = false;
    }

    visualizer.write_status_log(accumulated_reward, expected_reward, total_number_iterations);
    return true;
}

bool ReinforcementLearner::teacher(const PlanningFailureOpt& excuse_opt, std::vector< std::string > actions_requiring_excuse)
{
    LOG(INFO) << START_COLOR_RED << "Teacher needed." << END_COLOR;
    LOG_IF(excuse_opt, INFO) << "Excuse is \"" << *excuse_opt << "\" requiring actions " << actions_requiring_excuse;
    LOG_IF(!excuse_opt, INFO) << "No excuse found.";
    show_info_teacher_request(failure_analyzer_utils::planning_failure_to_string(excuse_opt), actions_requiring_excuse);
    high_level_planner->transition_context = "Teacher";

    // obtain action from auto teacher or human teacher
    TypedAtom::Opt action_opt;
    if (config_reader.get_variable<bool>("auto_teacher")) {
        action_opt = auto_teacher_ptr->get_action(scenario_ptr->get_state(), scenario_ptr->get_action_list(), actions_requiring_excuse);
        if ((!action_opt) || (*action_opt == MAKE_TYPED_ATOM_HELP)) {
            action_opt = read_action();
        }
    }
    else {
        action_opt = read_action();
    }

    // Special cases (retry, dead...)
    if ((!action_opt) || (*action_opt == MAKE_TYPED_ATOM_RETRY) || (*action_opt == MAKE_TYPED_ATOM_UNKNOWN)) {
        transition_obtained = false;
    }
    // if autoteacher request finish action (and has not been requested by rexd planner)
    //   then finish action reward < Vmin -> dead_end!
    else if ((*action_opt == MAKE_TYPED_ATOM_DEAD) || (*action_opt == MAKE_TYPED_ATOM_FINISH_ACTION)) {
        transition_obtained = false;
        LOG(INFO) << START_COLOR_RED << "Dead end! Aborting iteration..." << END_COLOR;
        if (excuse_opt) {
            PredicateGroup<TypedAtom> generic_excuse(*excuse_opt);
            generic_excuse.make_symbolic();
            LOG(INFO) << "Adding " << generic_excuse << " as dangerous problem.";
            logging_file << "Dangerous: " << generic_excuse << std::endl;
            dead_end_excuses.add_predicates(generic_excuse);
            scenario_ptr->get_modificable_rules().set_dangerous_rules(dead_end_excuses);
        }
        logging_file << "Dead end." << std::endl;
        if (config_reader.get_variable<bool>("final_reward_action")) {
            if (config_reader.get_variable<bool>("simulator")) {
                execute_action(MAKE_TYPED_ATOM_FINISH_ACTION, false); // execute as if it isn't teacher
            }
            else {
                execute_action(MAKE_TYPED_ATOM_DEAD, false); // execute as if it isn't teacher
            }
        }
        else {
            // returning false just in case of dead end
            return reset_rexd();
        }
    }
    else { // Execution
        scenario_ptr->domain.index_action(*action_opt);
        previous_action = *action_opt;

        if (config_reader.get_variable<bool>("simulator") || config_reader.get_variable<bool>("robot_executes_teacher_actions")) {
            execute_action(*action_opt, true);
        }
        logging_file << "Teacher: " << *action_opt << std::endl;
    }
    return true;
}

void ReinforcementLearner::monitoring()
{
    LOG(INFO) << "Planning";
    show_info_start_planning();

    TransitionGroup transitions(scenario_ptr->get_state().ppddl_object_manager);
    PlanningFailureOpt excuse_opt;
    std::vector<std::string> actions_requiring_excuse;
    TypedAtom::Opt planning_res = high_level_planner->exploit_or_explore(*scenario_ptr, transitions, excuse_opt, actions_requiring_excuse);
    std::string excuse_str = failure_analyzer_utils::planning_failure_to_string(excuse_opt);

    if (!planning_res) {
        show_info_end_planning(TypedAtom::Opt(), excuse_str);
        LOG(INFO) << "No action";
    }
    else {
        int actions_until_finish = high_level_planner->get_planner_ptr()->estimate_distance_to_finish(*scenario_ptr);
        LOG(INFO) << "Recomended action " << *planning_res << " (actions to finish : " << actions_until_finish << ")";
        show_info_end_planning(planning_res, excuse_str, actions_until_finish);
    }

    // Wait for manipulation recognition
    read_action();
}

