#include "domain.h"


/**
 * OPERATORS
 *
 */

std::ostream& operator<<(std::ostream &out, const Domain& dom)
{
    out << "Domain::" << "<<" << " - " << "Domain is :\n";
    for (size_t i = 0; i < dom.known_predicates.size(); i++)
        out << "Domain::" << "<<" << " - " << "Predicate<TypedAtom> " << i << ": " << dom.known_predicates[i] << "\n";

    return out;
}

/**
 * Domain
 *
 */

Domain::Domain(const ConfigReader& config_reader):
    variable_maintains_to_next_state(config_reader.get_variable<bool>("variable_maintains_to_next_state")),
    generalize_first_action_param(config_reader.get_variable<bool>("ppddl_rules_generalize_first_param")),
    generalize_second_action_param(config_reader.get_variable<bool>("ppddl_rules_generalize_second_param"))
{ }

template<class T>
int Domain::index_item(const T& item, std::vector<T>& known_items)
{
    int idx;
    bool found = false;
    for (size_t i = 0; i < known_items.size(); i++) {
        if (item == known_items[i]) {
            idx = i;
            found = true;
        }
    }
    if (!found) {
        idx = known_items.size();
        known_items.push_back(item);
    }
    return idx;
}

template<class T>
int Domain::get_item_index(const T& item, const std::vector<T>& known_items) const
{
    for (size_t i = 0; i < known_items.size(); i++) {
        if (item == known_items[i]) {
            return i;
        }
    }
    return -1;
}

void Domain::index_state(const State& state)
{
    index_predicates(state);
    index_objects(state.ppddl_object_manager.get_ppddl_objects_per_type());
}

void Domain::index_objects(const PPDDL::TypeObjectsMap& ppddl_objects_per_type)
{
    BOOST_FOREACH(const PPDDL::TypeObjectsMap::value_type& type_objects, ppddl_objects_per_type) {
        BOOST_FOREACH(const std::string& object_name, type_objects.second) {
            index_item<PPDDL::Object>(PPDDL::Object(object_name, type_objects.first), known_objects);
        }
    }
}

void Domain::index_predicates(const PredicateGroup<TypedAtom>& predicates)
{
    for (PredicateGroup<TypedAtom>::const_iterator predicate = predicates.begin(); predicate != predicates.end(); ++predicate) {
        index_predicate(*predicate);
    }
}

void Domain::index_goals(const GoalGroup& goals)
{
    for (GoalGroup::const_iterator goal_it = goals.begin(); goal_it != goals.end(); ++goal_it) {
        index_predicates(goal_it->get_predicates());
    }
}

void Domain::index_transitions(const TransitionGroup& transitions)
{
    for (TransitionSet::const_iterator trans_it = transitions.begin_set(); trans_it != transitions.end_set(); ++trans_it) {
        index_predicates(trans_it->get_prev_state());
        index_predicates(trans_it->get_post_state());
        index_action(trans_it->get_action());
    }
}

int Domain::index_type(const PPDDL::Type& ppddl_type)
{
    return index_item<PPDDL::Type>(ppddl_type, known_types);
}

int Domain::index_atom(const TypedAtom& symbol)
{
    int id = index_item<TypedAtom>(symbol, known_atoms);
    TypedAtom symbolic_atom(symbol);
    symbolic_atom.make_symbolic();
    index_item<TypedAtom>(symbolic_atom, known_symbolic_atoms);

    // add objects if GroundedPredicate
    if (!symbol.is_symbolic()) {
        BOOST_FOREACH(const PPDDL::Object& param, symbol.params) {
            index_item<PPDDL::Object>(param, known_objects);
            index_item<PPDDL::Type>(param.type, known_types);
        }
    }
    return id;
}

int Domain::index_predicate(const Predicate<TypedAtom>& predicate)
{
    Predicate<TypedAtom> sym_predicate(predicate);
    sym_predicate.make_symbolic();
    sym_predicate.positive = true;
    int id = index_item<Predicate<TypedAtom> >(predicate, known_predicates);
    // also index the symbol
    index_atom(predicate);

    // add objects if GroundedPredicate
    if (!predicate.is_symbolic()) {
        BOOST_FOREACH(const PPDDL::Object& param, predicate.params) {
            index_item<PPDDL::Object>(param, known_objects);
        }
    }
    return id;
}

int Domain::index_action(const TypedAtom& action)
{
    TypedAtom generic_action(action);
    generic_action.make_symbolic();
    int res_index =  index_item<TypedAtom>(generic_action, known_actions);
    BOOST_FOREACH(const PPDDL::Object& param, action.params) {
        index_type(param.type);
    }
    
    if (generalize_first_action_param && (generic_action.params.size() >= 1) ) {
        BOOST_FOREACH(PPDDL::Type& pddl_type, known_types) {
            generic_action.params[0].type = pddl_type;
            index_item<TypedAtom>(generic_action, known_actions);
        }
    }
    
    if (generalize_second_action_param && (generic_action.params.size() >= 2) ) {
        BOOST_FOREACH(PPDDL::Type& pddl_type, known_types) {
            generic_action.params[1].type = pddl_type;
            index_item<TypedAtom>(generic_action, known_actions);
        }
    }
    
    return res_index;
}

void Domain::index_rules(const RuleSet& rules)
{
    for (RuleList::const_iterator rule_it = rules.begin(); rule_it != rules.end(); ++rule_it) {
        index_rule(**rule_it);
    }
}

void Domain::index_rule(const Rule& rule)
{
    index_action(rule);
    index_predicates(rule.get_preconditions());
    BOOST_FOREACH(const Outcome& out_it, rule.outcomes) {
        index_predicates(out_it);
    }
}


TransitionGroup Domain::get_transitions_with_indexed_objects(const TransitionGroup& transitions)
{
    TransitionGroup indexed_transitions(transitions.ppddl_object_manager);
    // TODO change ppddl_object_manager
    for ( TransitionGroup::TransitionHistorial::const_iterator transition_it_it = transitions.begin_historial(); 
          transition_it_it != transitions.end_historial();
          ++transition_it_it) 
          {
        Transition indexed_transition = get_transition_with_indexed_objects(**transition_it_it);
        indexed_transitions.add_transition(indexed_transition);
    }
    return indexed_transitions;
}

Transition Domain::get_transition_with_indexed_objects(const Transition& transition)
{
    return Transition(get_predicates_with_indexed_objects(transition.get_prev_state()),
                      get_typed_atom_with_indexed_objects(transition.get_action()),
                      get_predicates_with_indexed_objects(transition.get_post_state()),
                      transition.get_reward(),
                      variable_maintains_to_next_state
                     );
}

PredicateGroup<TypedAtom> Domain::transform_to_typed(const PredicateGroup<NonTypedAtom>& non_typed_preds) const
{
    PredicateGroup<TypedAtom> result;
    BOOST_FOREACH(const Predicate<NonTypedAtom>& non_typed_pred, non_typed_preds) {
        Predicate<TypedAtom> typed_pred(transform_to_typed(static_cast<const NonTypedAtom>(non_typed_pred)), non_typed_pred.positive);
        result.insert(typed_pred);
    }
    return result;
}

TypedAtom Domain::transform_to_typed(const NonTypedAtom& non_typed_atom) const
{
    TypedAtom typed_atom(TypedAtom(non_typed_atom.name, std::vector<PPDDL::Object>()));
    bool done = false;
    // First check known symbols
    BOOST_FOREACH(const TypedAtom& known_atom, known_symbolic_atoms) {
        if (non_typed_atom.name == known_atom.name) {
            assert(non_typed_atom.params.size() == known_atom.params.size());
            for (size_t i = 0; i < non_typed_atom.params.size(); ++i) {
                typed_atom.params.push_back(PPDDL::Object(non_typed_atom.params[i].name, known_atom.params[i].type));
            }
            done = true;
            break;
        }
    }
    if (!done) {
        // Else check known actions
        BOOST_FOREACH(const TypedAtom& known_atom, known_actions) {
            if (non_typed_atom.name == known_atom.name) {
                assert(non_typed_atom.params.size() == known_atom.params.size());
                for (size_t i = 0; i < non_typed_atom.params.size(); ++i) {
                    typed_atom.params.push_back(PPDDL::Object(non_typed_atom.params[i].name, known_atom.params[i].type));
                }
                done = true;
                break;
            }
        }
    }
    if (!done) {
        // Else set default type
        for (size_t i = 0; i < non_typed_atom.params.size(); ++i) {
            typed_atom.params.push_back(PPDDL::Object(non_typed_atom.params[i].name, DEFAULT_OBJECT_TYPE));
        }
    }
    return typed_atom;
}

PredicateGroup<TypedAtom> Domain::get_predicates_with_indexed_objects(const PredicateGroup<TypedAtom>& predicates)
{
    PredicateGroup<TypedAtom> indexed_predicates;
    BOOST_FOREACH(const Predicate<TypedAtom>& predicate, predicates) {
        indexed_predicates.insert(get_predicate_with_indexed_objects(predicate));
    }
    return indexed_predicates;
}

Predicate<TypedAtom> Domain::get_predicate_with_indexed_objects(const Predicate<TypedAtom>& predicate)
{
    return Predicate<TypedAtom>(get_typed_atom_with_indexed_objects(predicate), predicate.positive);
}

TypedAtom Domain::get_typed_atom_with_indexed_objects(const TypedAtom& symbol)
{
    TypedAtom indexed_atom(symbol);
    BOOST_FOREACH(PPDDL::Object& param, indexed_atom.params) {
        // TODO UGLY HACK !!!
        param.name = boost::lexical_cast<std::string>(get_object_index(param) + 60);
    }
    return indexed_atom;
}

bool Domain::check_rule_types(const Rule& rule) const
{
    std::set< Predicate<TypedAtom> > predicates_to_check;
    predicates_to_check.insert(rule.get_preconditions().begin(), rule.get_preconditions().end());
    BOOST_FOREACH(const Outcome& outcome, rule.outcomes) {
        predicates_to_check.insert(outcome.begin(), outcome.end());
    }
    BOOST_FOREACH(const Predicate<TypedAtom>& predicate_to_check, predicates_to_check) {
        bool found = false;
        std::vector<Predicate<TypedAtom> >::const_iterator pred_it = known_predicates.begin();
        while ((!found) && (pred_it != known_predicates.end())) {
            if (predicate_to_check.name == pred_it->name) {
                if(predicate_to_check.params.size() != pred_it->params.size()) {
                    return false;
                }
                for (size_t i = 0; i < pred_it->params.size(); ++i) {
                    if (predicate_to_check.params[i].type != pred_it->params[i].type) {
                        return false;
                    }
                }
                found = true;
            }
            ++pred_it;
        }
        if (!found) {
            return false;
        }
    }
    return true;
}

void Domain::set_rule_types(Rule& rule)
{
    bool found = false;
    std::vector<TypedAtom>::const_iterator action_it = known_actions.begin();
    while ((!found) && (action_it != known_actions.end())) {
        if (action_it->name == rule.name) {
            assert(action_it->params.size() == rule.params.size());
            rule.change_param_types(rule.get_param_names(), action_it->get_param_types());
            found = true;
        }
        ++action_it;
    }
    std::vector<PPDDL::Object> deictic_params = rule.get_deictic_params_v();
    BOOST_FOREACH(const PPDDL::Object& deictic_param, deictic_params) {
        PredicateGroup<TypedAtom> deictic_preconditions = rule.get_preconditions().get_predicates_with_var(deictic_param);
        // can be empty, when updating rule param types, we update for all params in predicate,
        // so we may update deictic params that have not been parsed yet in the loop, and when
        // they are parsed the deictic_preconditions will be empty
        if (!deictic_preconditions.empty()) { 
            Predicate<TypedAtom> deictic_prec = *deictic_preconditions.begin();
            
            found = false;
            std::vector<TypedAtom>::const_iterator atom_it = known_symbolic_atoms.begin();
            while ((!found) && (atom_it != known_symbolic_atoms.end())) {
                if (atom_it->name == deictic_prec.name) {
                    assert(atom_it->params.size() == deictic_prec.params.size());
                    rule.change_param_types(deictic_prec.get_param_names(), atom_it->get_param_types());
                    found = true;
                }
                ++atom_it;
            }
        }
    }
}

size_t Domain::get_atom_index(const TypedAtom& symbol)
{
    int id = get_item_index(symbol, known_atoms);
    if (id < 0) { // not found
        id = index_atom(symbol);
    }
    return id;
}

size_t Domain::get_predicate_index(const Predicate<TypedAtom>& predicate)
{
    Predicate<TypedAtom> sym_predicate(predicate);
    sym_predicate.make_symbolic();
    int id = get_item_index(predicate, known_predicates);
    if (id < 0) { // not found
        id = index_predicate(predicate);
    }
    return id;
}

size_t Domain::get_predicate_index_const(const Predicate<TypedAtom>& predicate) const
{
    Predicate<TypedAtom> sym_predicate(predicate);
    sym_predicate.make_symbolic();
    int id = get_item_index(predicate, known_predicates);
    return id;
}

std::vector<size_t> Domain::get_predicate_group_indexes_const(const PredicateGroup<TypedAtom>& predicates) const
{
    std::vector<size_t> result;
    BOOST_FOREACH(const Predicate<TypedAtom>& predicate, predicates) {
        result.push_back(get_predicate_index_const(predicate));
    }
    return result;
}

IndexedAtom Domain::get_indexed_atom(const TypedAtom& symbol)
{
    return IndexedAtom(get_atom_index(symbol));
}

Predicate<IndexedAtom> Domain::get_indexed_predicate(const Predicate<TypedAtom>& predicate)
{
    return Predicate<IndexedAtom>(get_indexed_atom(predicate), predicate.positive);
}

PredicateGroup<IndexedAtom> Domain::get_indexed_predicate_group(const PredicateGroup<TypedAtom>& pred_group)
{
    PredicateGroup<IndexedAtom> indexed_predicates;
    BOOST_FOREACH(const Predicate<TypedAtom>& predicate, pred_group) {
        indexed_predicates.insert(get_indexed_predicate(predicate));
    }
    return indexed_predicates;
}

PredicateGroup<TypedAtom> Domain::get_predicate_group(const PredicateGroup<IndexedAtom>& indexed_predicates) {
    PredicateGroup<TypedAtom> result;
    BOOST_FOREACH(const Predicate<IndexedAtom>& indexed_predicate, indexed_predicates) {
        result.insert(get_predicate(indexed_predicate));
    }
    return result;
}

size_t Domain::get_action_index(const TypedAtom& action)
{
    int id = get_item_index(action, known_actions);
    if (id < 0) { // not found
        id = index_action(action);
    }
    return id;
}

size_t Domain::get_object_index(const PPDDL::Object& object)
{
    return get_item_index(object, known_objects);
}

PredicateGroup<TypedAtom> Domain::get_constant_predicates() const
{
    PredicateGroup<TypedAtom> result;
    BOOST_FOREACH(const Predicate<TypedAtom>& known_predicate, known_predicates) {
        if (known_predicate.is_constant()) {
            result.insert(known_predicate);
        }
    }
    return result;
}

PredicateGroup<TypedAtom> Domain::get_predicate_group_no_closed_world(const PredicateGroup<TypedAtom>& predicates, const bool symbolic) const
{
    PredicateGroup<TypedAtom> result(predicates);
    
    std::set< Predicate<TypedAtom> > predicates_to_add;
    if (symbolic) {
        PPDDLObjectManager ppddl_object_manager;
        ppddl_object_manager.add_objects(known_objects.begin(), known_objects.end());
        std::set< Predicate<TypedAtom> > symbolic_predicates_to_add;
        BOOST_FOREACH(Predicate<TypedAtom> copy_predicate, known_predicates) {
            copy_predicate.make_symbolic();
            symbolic_predicates_to_add.insert(copy_predicate);
        }
        BOOST_FOREACH(Predicate<TypedAtom> symbolic_predicate, symbolic_predicates_to_add) {
            PPDDLObjectManager::CombinationsPPDDLObjectsPtr combinations_of_params = ppddl_object_manager.get_combinations_of_objecs_for_given_params(symbolic_predicate.get_param_types());
            // get all groundings
            
            BOOST_FOREACH(const std::vector< std::string >& objects, *combinations_of_params) {
                std::vector<PPDDL::Object> new_params(symbolic_predicate.params);
                for (size_t i = 0; i < new_params.size(); ++i) {
                    new_params[i].name = objects[i];
                }
                Predicate<TypedAtom> new_predicate(symbolic_predicate);
                new_predicate.params = new_params;
                predicates_to_add.insert(new_predicate);
            }
        }
    }
    else {
        predicates_to_add.insert(known_predicates.begin(), known_predicates.end());
    }
    BOOST_FOREACH(const Predicate<TypedAtom>& predicate, predicates) {
        std::set< Predicate<TypedAtom> >::iterator pred_found_iterator;
        pred_found_iterator = predicates_to_add.find(predicate);
        if (pred_found_iterator != predicates_to_add.end()) {
            predicates_to_add.erase(pred_found_iterator);
        }
        pred_found_iterator = predicates_to_add.find(predicate.get_negated());
        if (pred_found_iterator != predicates_to_add.end()) {
            predicates_to_add.erase(pred_found_iterator);
        }
    }
    
    BOOST_FOREACH(Predicate<TypedAtom> predicate_to_add, predicates_to_add) {
        predicate_to_add.positive = false;
        result.insert(predicate_to_add);
    }
    
    return result;
}

TransitionGroup Domain::get_transitions_no_closed_world(const TransitionGroup& symbolic_transitions) const
{
    TransitionGroup result(symbolic_transitions.ppddl_object_manager);
    
    for ( TransitionGroup::TransitionHistorial::const_iterator transition_it_it = symbolic_transitions.begin_historial(); 
          transition_it_it != symbolic_transitions.end_historial();
          ++transition_it_it) 
          {
        Transition new_transition(get_predicate_group_no_closed_world((*transition_it_it)->get_prev_state()),
                                  (*transition_it_it)->get_action(),
                                  get_predicate_group_no_closed_world((*transition_it_it)->get_post_state()),
                                  (*transition_it_it)->get_reward(),
                                  variable_maintains_to_next_state
                                 );
        result.add_transition(new_transition);
    }
    return result;
}

void Domain::write_LFIT_predicate_types(std::ostream& out) const
{
    std::set<TypedAtom> predicate_atom_set; // to write domain we don't need sign(+/-). Predicate<TypedAtom> is just a symbol with a sign(+/-)
    BOOST_FOREACH(const Predicate<TypedAtom> & predicate, known_predicates) {
        predicate_atom_set.insert(predicate);
    }

    BOOST_FOREACH(TypedAtom symbol, predicate_atom_set) {
        if (isupper(symbol.name[0])) { // constant
            out << "VAR* ";
        }
        else {
            out << "VAR ";
        }
        out << symbol.name;
        out << " 0 1";
        out << std::endl;
    }
    
    out << "VAR* action ";
    BOOST_FOREACH(TypedAtom symbol, known_actions) {
        out << symbol.name << " ";
    }
    if (known_actions.size() == 1) {
        out << MAKE_NONTYPED_ATOM_NOACTION.name;
    }
//     out << MAKE_NONTYPED_ATOM_NOACTION.name;
    out << std::endl;
}

void Domain::write_PPDDL_grounded_predicates(std::ostream& out) const
{
    std::set<TypedAtom> predicate_atom_set; // to write domain we don't need sign(+/-). Predicate<TypedAtom> is just a symbol with a sign(+/-)
    BOOST_FOREACH(const Predicate<TypedAtom> & predicate, known_predicates) {
        predicate_atom_set.insert(predicate);
    }

    out << "  (:predicates " << std::endl;
    BOOST_FOREACH(const TypedAtom & symbol, predicate_atom_set) {
        out << "    ";
        symbol.write_PPDDL(out);
        out << std::endl;
    }
    out << "  )" << std::endl;
}

void Domain::write_PPDDL_symbolic_predicates(std::ostream& out, const PPDDLObjectManager& pddl_object_manager) const
{
    // TODO Load actual predicates
    std::set<TypedAtom> typed_atom_set; // to write domain we don't need sign(+/-). Predicate<TypedAtom> is just a symbol with a sign(+/-)
    BOOST_FOREACH(const Predicate<TypedAtom> & predicate, known_predicates) {
        TypedAtom typed_atom(predicate);
        typed_atom.make_symbolic();
        if ( (typed_atom.params.empty()) || (typed_atom.params.front().type != DEFAULT_OBJECT_TYPE) ) {
            // duplicates predicates because we don't know type for ?X or ?Y in some symbolic predicates
            typed_atom_set.insert(typed_atom);
        }
    }

    out << "  (:predicates ";
    BOOST_FOREACH(const TypedAtom & typed_atom, typed_atom_set) {
        out << "               ";
        typed_atom.write_PPDDL(out);
        out << std::endl;
    }
    out << "  )" << std::endl;
}

void Domain::write_RDDL_objects(std::ostream& out, const PPDDLObjectManager& pddl_object_manager) const
{
    out << "  types {";
//     BOOST_FOREACH(const PDDLObjectTypeMap::value_type & pair_object_type, pddl_object_manager.get_ppddl_objects_types()) {
//         out << "    " << pair_object_type << std::endl;
//     }
    BOOST_FOREACH(const PPDDL::TypeObjectsMap::value_type & pair_type_objects, pddl_object_manager.get_ppddl_objects_per_type()) {
        out << "    " << pair_type_objects.first << " : object" << std::endl;
    }
    out << "  };" << std::endl;
}

void Domain::write_RDDL_symbolic_predicates(std::ostream& out, const PPDDLObjectManager& pddl_object_manager) const
{
    out << "  pvariables {";
    BOOST_FOREACH(const TypedAtom & symbol, known_atoms) {
        out << "    " << NonTypedAtom(symbol.name, symbol.get_param_names()) << " : {state-fluent, bool, default = false}" << std::endl;
    }
    out << "  };" << std::endl;
}

void Domain::write_NID(std::ostream& out) const
{
    std::set<TypedAtom> predicate_atom_set, action_atom_set; // to write domain we don't need sign(+/-). Predicate<TypedAtom> is just a symbol with a sign(+/-)
    BOOST_FOREACH(const Predicate<TypedAtom> & predicate, known_predicates) {
        predicate_atom_set.insert(predicate);
    }

    BOOST_FOREACH(const TypedAtom & action, known_actions) {
        action_atom_set.insert(action);
    }

    out << "# predicates" << std::endl;
    BOOST_FOREACH(const TypedAtom & primitive, predicate_atom_set) {
        out << primitive.name << " " << primitive.params.size() << " primitive binary" << std::endl;
    }

    out << "# actions" << std::endl;
    BOOST_FOREACH(const TypedAtom & action, action_atom_set) {
        out << action.name << " " << action.params.size() << " action binary" << std::endl;
    }
}





