/*
 * Class that represents the predicates and rules domain. It stores the different existing objects and symbols.
 * Copyright (C) 2014  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef IRI_RULE_LEARNER_DOMAIN_H
#define IRI_RULE_LEARNER_DOMAIN_H 1

#include <vector>
#include <fstream>
#include <cmath>

#include <boost/shared_ptr.hpp>
#include <boost/make_shared.hpp>

#include <symbolic/symbols.h>
#include <symbolic/predicates.h>
#include <symbolic/transitions.h>
#include <symbolic/state.h>
#include <symbolic/rules.h>
#include <rexd/config_reader.h>

class Rule;
class RuleSet;

/*!
\class Domain
\brief Class indexing all domain symbols, actions and objects.

At this moment it just indexes symbols, actions and objects. This indexing is just used to generate excuses, but it would be nice to use it in other parts also. If it was used for planning and learning, symbol arguments wouldn't have to be just one character or a number to be compatible with libprada implementation.
*/
class Domain
{
public:

    Domain(const ConfigReader& config_reader);

    void clear() {
        known_predicates.clear();
        known_actions.clear();
        known_objects.clear();
    }

    int index_type(const PPDDL::Type& ppddl_type);
    int index_atom(const TypedAtom& symbol);
    int index_predicate(const Predicate<TypedAtom>& predicate);
    int index_action(const TypedAtom& action);
    void index_state(const State& state);
    void index_predicates(const PredicateGroup<TypedAtom>& predicates);
    void index_goals(const GoalGroup& goals);
    void index_transitions(const TransitionGroup& transitions);
    void index_rule(const Rule& rule);
    void index_rules(const RuleSet& rules);
    
    /**
     * \brief Index a set of PDDL object-type
     * 
     * \param ppddl_objects_per_type Vector of objects per type
     */
    void index_objects(const PPDDL::TypeObjectsMap& ppddl_objects_per_type);
    
    /**
     * \brief Get transitions with object indexes
     * 
     * Substitutes objects by the indexes
     */
    TransitionGroup get_transitions_with_indexed_objects(const TransitionGroup& transitions);

    size_t get_object_index(const PPDDL::Object& object);
    size_t get_action_index(const TypedAtom& action);
    size_t get_atom_index(const TypedAtom& symbol);
    size_t get_predicate_index(const Predicate<TypedAtom>& predicate);
    size_t get_predicate_index_const(const Predicate<TypedAtom>& predicate) const;
    std::vector<size_t> get_predicate_group_indexes_const(const PredicateGroup<TypedAtom>& predicates) const;
    
    /**
     * \brief Transforms to typed based on knwon symbols
     */
    PredicateGroup<TypedAtom> transform_to_typed(const PredicateGroup<NonTypedAtom>& non_typed_preds) const;
    TypedAtom transform_to_typed(const NonTypedAtom& non_typed_atom) const;
    
    IndexedAtom get_indexed_atom(const TypedAtom& symbol);
    Predicate<IndexedAtom> get_indexed_predicate(const Predicate<TypedAtom>& predicate);
    PredicateGroup<IndexedAtom> get_indexed_predicate_group(const PredicateGroup<TypedAtom>& pred_group);

    size_t get_predicates_size() {
        return known_predicates.size();
    }

    const std::vector< Predicate<TypedAtom> >& get_predicates() const {
        return known_predicates;
    }
    PredicateGroup< TypedAtom > get_constant_predicates() const;
    const std::vector< TypedAtom >& get_actions() const {
        return known_actions;
    }
    const std::vector< TypedAtom >& get_atoms() const {
        return known_atoms;
    }
    const std::vector< TypedAtom >& get_symbolic_atoms() const {
        return known_symbolic_atoms;
    }
    const std::vector< PPDDL::Object >& get_objects() const {
        return known_objects;
    }
    const std::vector< PPDDL::Type >& get_ppddl_types() const {
        return known_types;
    }
    std::vector< NonTypedAtom > get_actions_non_typed() const {
        std::vector<NonTypedAtom> non_typed_known_actions;
        BOOST_FOREACH(const TypedAtom& typed_action, known_actions) {
            non_typed_known_actions.push_back(NonTypedAtom(typed_action.name, typed_action.get_param_names()));
        }
        return non_typed_known_actions;
    }
    Predicate<TypedAtom> get_predicate(const size_t index) {
        return known_predicates[index];
    }
    Predicate<TypedAtom> get_predicate(Predicate<IndexedAtom> indexed_predicate) {
        return Predicate<TypedAtom>(known_atoms[indexed_predicate.index], indexed_predicate.positive);
    }
    PredicateGroup<TypedAtom> get_predicate_group(const PredicateGroup<IndexedAtom>& indexed_predicates);
    TypedAtom get_action(size_t index) {
        return known_actions[index];
    }
    PPDDL::Object get_object(size_t index) {
        return known_objects[index];
    }
    
    PredicateGroup<TypedAtom> get_predicate_group_no_closed_world(const PredicateGroup<TypedAtom>& predicates, const bool symbolic = false) const;
    TransitionGroup get_transitions_no_closed_world(const TransitionGroup& transitions) const;
    
    bool check_rule_types(const Rule& rule) const;
    void set_rule_types(Rule& rule);

    void write_LFIT_predicate_types(std::ostream& out) const;
    void write_PPDDL_grounded_predicates(std::ostream& out) const;
    void write_PPDDL_symbolic_predicates(std::ostream& out, const PPDDLObjectManager& pddl_object_manager) const;
    void write_RDDL_objects(std::ostream& out, const PPDDLObjectManager& pddl_object_manager) const;
    void write_RDDL_symbolic_predicates(std::ostream& out, const PPDDLObjectManager& pddl_object_manager) const;
    void write_NID(std::ostream& out) const;

    friend std::ostream& operator<<(std::ostream &out, const Domain& dom);

private:
    std::vector< Predicate<TypedAtom> > known_predicates;
    std::vector< TypedAtom > known_atoms;
    std::vector< TypedAtom > known_symbolic_atoms;
    std::vector< TypedAtom > known_actions;
    std::vector< PPDDL::Object > known_objects;
    std::vector< PPDDL::Type> known_types;
    bool variable_maintains_to_next_state;
    bool generalize_first_action_param;
    bool generalize_second_action_param;
    
    template<class T>
    int index_item(const T& item, std::vector<T>& known_items);
    template<class T>
    int get_item_index(const T& item, const std::vector<T>& known_items) const;
    
    // Other
    Transition get_transition_with_indexed_objects(const Transition& transition);
    TypedAtom get_typed_atom_with_indexed_objects(const TypedAtom& symbol);
    PredicateGroup<TypedAtom> get_predicates_with_indexed_objects(const PredicateGroup<TypedAtom>& predicates);
    Predicate<TypedAtom> get_predicate_with_indexed_objects(const Predicate<TypedAtom>& predicate);
};

typedef boost::shared_ptr<Domain> DomainPtr;


#endif
