#include "goals.h"


std::ostream& operator<<(std::ostream &out, const Goal& g)
{
    out << g.reward << " ";
    out << g.goal_predicates;
    out << std::flush;
    return out;
}

std::ostream& operator<<(std::ostream &out, const GoalGroup& g)
{
    for (GoalGroup::const_iterator goal_it = g.begin(); goal_it != g.end(); ++goal_it) {
        out << *goal_it << std::endl;
    }
    return out;
}

void Goal::write_PPDDL(std::ostream &out, const std::vector<PPDDL::Object>& rule_params) const
{
    if (!goal_predicates.empty()) {
        out << "                (when  ";
        goal_predicates.write_PPDDL(out, rule_params);
    }
    out << " (";
    if (reward >= 0)
        out << "increase ";
    else
        out << "decrease ";
    out << "(reward) ";
    out << std::abs(reward);
    out << ")";
    if ( !goal_predicates.empty() ) {
        out << ")";
    }
    out << std::endl;
}

void GoalGroup::write_PPDDL(std::ostream &out, const std::vector<PPDDL::Object>& rule_params) const
{
//     out << std::endl << "    ; Reward" << std::endl;
    for (GoalGroup::const_iterator goal_it = begin(); goal_it != end(); ++goal_it) {
        goal_it->write_PPDDL(out, rule_params);
    }
}


// Goal
Goal::Goal(const std::string& goal_string):
    reward(0)
{
    std::string ss_item;
    std::stringstream goal_ss(goal_string);
    std::getline(goal_ss, ss_item, ' ');
    try {
        reward = boost::lexical_cast<float>(ss_item);
        std::getline(goal_ss, ss_item);
        if (goal_ss.str() == ss_item) { // if no predicates getline does not obtain anything new
            ss_item = "";
        }
        add_predicates(PredicateGroup<TypedAtom>(ss_item));
    } catch (boost::bad_lexical_cast &) {
        LOG(ERROR) << "Error parsing goal " << goal_string;
        LOG(WARNING) << "Compatibility mode, setting reward to 1.0 and goal predicates to " << goal_string;
        reward = 1.0;
        add_predicates(PredicateGroup<TypedAtom>(goal_string));
    }
}

void Goal::change_param(const std::string& orig_var, const std::string& new_var)
{
    this->goal_predicates.change_param_name(orig_var, new_var);
}

float Goal::obtain_reward(const PredicateGroup<TypedAtom>& predicates) const
{
    float obtained_reward = 0;
    if (goal_predicates.is_satisfied_by(predicates)) {
        obtained_reward += this->reward;
    }
    return obtained_reward;
}

GoalGroup Goal::get_pddl_grounded_goals(const std::vector<std::string>& symbolic_params, 
                                        const std::vector<std::string>& grounded_params, 
                                        const PPDDLObjectManager& object_manager) const
{
    GoalGroup result_goals;
    
    Goal copy_goal(*this);
    copy_goal.goal_predicates.change_param_names(symbolic_params, grounded_params);
    
    // get symbolic params with their corresponding type
    PPDDL::ObjectTypeMap symbolic_object_type_params;
    BOOST_FOREACH(const Predicate<TypedAtom>& goal_pred, copy_goal.goal_predicates) {
        BOOST_FOREACH(const PPDDL::Object& param, goal_pred.params) {
            if ( symbolic::is_param_symbolic(param.name) ) {
                symbolic_object_type_params[param.name] = param.type;
            }
        }
    }
    result_goals.push_back(copy_goal);
    
    // create all groundings for each param based on its types
    BOOST_FOREACH(const PPDDL::ObjectTypeMap::value_type& object_type_param, symbolic_object_type_params) {
        GoalGroup new_goals;
        BOOST_FOREACH(const Goal& result_goal, result_goals) {
            BOOST_FOREACH(const std::string& object, object_manager.get_type_objects(object_type_param.second)) {
                Goal new_goal(result_goal);
                new_goal.goal_predicates.change_param_name(object_type_param.first, object);
                new_goals.push_back(new_goal);
            }
        }
        result_goals = new_goals;
    }
    
//     LOG(INFO) << "Grounded goals are " << result_goals;
    return result_goals;
}


// GoalGroup
GoalGroup::GoalGroup(std::string file_path)
{
    std::ifstream goals_file(file_path.c_str());
    if (goals_file.is_open()) {
        while (goals_file.good()) {
            std::string goal_string;
            std::getline(goals_file, goal_string);
            goal_string = utils::trim(goal_string);
            if (goal_string != "") {
                push_back(Goal(goal_string));
            }
        }
        goals_file.close();
    } else LOG(ERROR) << "Unable to open file: "  << file_path;
}

bool GoalGroup::is_goal(const PredicateGroup<TypedAtom>& state) const
{
    if (empty()) { // no goal
        return false;
    }
    bool result = true;
    if (!state.has_predicate(Predicate<TypedAtom>(MAKE_TYPED_ATOM_FINISHED, true))) { // if not finished
        for (GoalGroup::const_iterator goal_it = begin(); goal_it != end(); ++goal_it) { // check if all goals are satisfied
            if (!goal_it->is_goal(state)) {
                result = false;
            }
        }
    }
    return result;
}


PredicateGroup<TypedAtom> GoalGroup::get_predicates() const
{
    PredicateGroup<TypedAtom> predicates;
    for (GoalGroup::const_iterator goal_it = begin(); goal_it != end(); ++goal_it) {
        predicates.add_predicates(goal_it->get_predicates());
    }
    return predicates;
}

float GoalGroup::obtain_reward(const PredicateGroup<TypedAtom>& predicates) const
{
    float obtained_reward = 0;
    for (GoalGroup::const_iterator goal_it = begin(); goal_it != end(); ++goal_it) {
        obtained_reward += goal_it->obtain_reward(predicates);
    }
    return obtained_reward;
}

Goal GoalGroup::adapt_finish_rule_goal_to_simple_predicate_goal()
{
    Goal simple_goal(1.0);

    for (GoalGroup::const_iterator goal_it = begin(); goal_it != end(); ++goal_it) {
        PredicateGroup<TypedAtom> goal_predicates = goal_it->get_predicates();
        goal_predicates.erase(Predicate<TypedAtom>(MAKE_TYPED_ATOM_FINISHED, false));
        if (goal_it->get_reward() < 0.0) {
            goal_predicates = goal_predicates.get_negated();
        }
        simple_goal.add_predicates(goal_predicates);
    }

    return simple_goal;
}

GoalGroup GoalGroup::get_pddl_grounded_goals(const std::vector<std::string>& symbolic_params, const std::vector<std::string>& grounded_params, const PPDDLObjectManager& pddl_object_manager) const
{
    GoalGroup result;
    for (GoalGroup::const_iterator goal_it = begin(); goal_it != end(); ++goal_it) {
        std::vector<Goal> new_grounded_goals = goal_it->get_pddl_grounded_goals(symbolic_params, grounded_params, pddl_object_manager);
        result.insert(result.end(), new_grounded_goals.begin(), new_grounded_goals.end());
    }
    return result;
}

void GoalGroup::PPDDL_grounding()
{
    for (GoalGroup::iterator goal_it = begin(); goal_it != end(); ++goal_it) {
        goal_it->PPDDL_grounding();
    }
}

void GoalGroup::transform_symbolic_NID_to_symbolic_PDDL()
{
    BOOST_FOREACH(Goal& goal, *this) {
        goal.transform_symbolic_NID_to_symbolic_PDDL();
    }
}

void GoalGroup::change_param_types(const std::vector<std::string>& param_names, const std::vector<std::string>& param_types)
{
    BOOST_FOREACH(Goal& goal, *this) {
        goal.change_param_types(param_names, param_types);
    }
}

