#include "predicates.h"
#include "symbolic/utils.h"

using namespace std;


bool operator==(const Outcome &o1, const Outcome &o2)
{
    return ( (*static_cast<const PredicateGroup<TypedAtom>*>(&o1) == o2) && (utils::is_close(o1.probability, o2.probability)) );
}

std::ostream& operator<<(std::ostream &out, const Outcome& o)
{
    out << o.probability << " ";
    if (o.noisy) {
        out << "<noise>";
    } else {
        out << Predicates::transform_to_nontyped(o); // *static_cast<const PredicateGroup<TypedAtom>*>(&o);
    }
    return out;
}

std::ostream& operator<<(std::ostream &out, const OutcomeGroup& outcome_group)
{
    BOOST_FOREACH(const Outcome& outcome, outcome_group) {
        out << outcome << std::endl;
    }
    return out;
}

bool operator<(const Outcome& l, const Outcome& r)
{
    if (l.noisy && !r.noisy) // noisy is last!
        return false;
    else if (!l.noisy && r.noisy) // noisy is last!
        return true;
    else if (l.empty() && !r.empty()) // empty is last!
        return false;
    else if (!l.empty() && r.empty()) // empty is last!
        return true;
    else if (utils::is_significantly_less(l.probability, r.probability)) // larger prob is first!
        return false;
    else if (utils::is_significantly_less(r.probability, l.probability)) // larger prob is first!
        return true;
    else if (*static_cast<const PredicateGroup<TypedAtom>*>(&l) < *static_cast<const PredicateGroup<TypedAtom>*>(&r))
        return true;
    else if (*static_cast<const PredicateGroup<TypedAtom>*>(&r) < *static_cast<const PredicateGroup<TypedAtom>*>(&l))
        return false;
    else
        return false;
}

bool operator<(const OutcomeGroup& l, const OutcomeGroup& r)
{
    return (*static_cast<const std::set<Outcome>*>(&l)) < r;
}

void Outcome::write_PPDDL(std::ostream &out, const std::vector<PPDDL::Object>& rule_params) const
{
    out << " " << probability << " ";
    Predicates::transform_to_nontyped(*this).write_PPDDL(out, rule_params);
}

void OutcomeGroup::write_PPDDL(std::ostream &out, const std::vector<PPDDL::Object>& rule_params) const
{
    out << "(probabilistic ";
    BOOST_FOREACH(Outcome outcome, *this) {
        if (outcome.is_useful()) {
            outcome.write_PPDDL(out, rule_params);
        }
    }
    out << ")";
}


/**
 * Outcome
 */
double Outcome::probability_obtain_predicate(const Predicate<TypedAtom> predicate) const
{
    if (has_predicate(predicate))
        return probability;
    else
        return 0.0;
}




Outcome::Outcome(std::string line):
    is_dangerous(false)
{
    std::string item;
    noisy = false;

    // remove leading spaces
    line = line.substr(line.find_first_not_of(" \t"));
    std::stringstream ss(line);

    if (std::getline(ss, item, ' ')) {
        //std::cout << "Probability: " << item << std::endl;
    }
    probability = atof(item.c_str());

    if (ss.eof())
        return;

    // get rest of the line
    std::getline(ss, item);
    if (item.find("<noise>") != std::string::npos) {
        noisy = true;
    } else if (item.find("<no-change>") != std::string::npos) {

    } else {
        replace_predicates(PredicateGroup(item));
    }
}

bool Outcome::changes_predicates() const
{
    if ((size() > 0) && (!noisy)) {
        return true;
    } else {
        return false;
    }
}

bool Outcome::is_useful() const
{
    if (changes_predicates() && (probability > 0.0)) {
        return true;
    } else {
        return false;
    }
}

bool Outcome::apply(PredicateGroup& input_predicates) const
{
    if (changes_predicates()) {
        input_predicates.update_predicates(*this);
        return true;
    } else
        return false;
}

bool OutcomeGroup::changes_predicates() const
{
    bool result = false;
    for (OutcomeGroup::const_iterator outcome_it = this->begin(); outcome_it != this->end(); outcome_it++) {
        if (outcome_it->changes_predicates()) {
            result = true;
        }
    }
    return result;
}

void OutcomeGroup::transform_symbolic_NID_to_symbolic_PDDL()
{
    std::set<Outcome> old_set;
    swap(old_set);
    BOOST_FOREACH(Outcome outcome, old_set) {
        outcome.transform_symbolic_NID_to_symbolic_PDDL();
        insert(outcome);
    }
}

void OutcomeGroup::add_predicate_to_outcomes(const Predicate<TypedAtom>& new_predicate)
{
    std::set<Outcome> old_set;
    swap(old_set);
    BOOST_FOREACH(Outcome outcome, old_set) {
        outcome.update_predicate(new_predicate);
        insert(outcome);
    }
}

void OutcomeGroup::change_param_name(const std::string& current_name, const std::string& new_name)
{
    std::set<Outcome> old_set;
    swap(old_set);
    BOOST_FOREACH(Outcome outcome, old_set) {
        outcome.change_param_name(current_name, new_name);
        insert(outcome);
    }
}

void OutcomeGroup::PPDDL_grounding()
{
    std::set<Outcome> old_set;
    swap(old_set);
    BOOST_FOREACH(Outcome outcome, old_set) {
        outcome.PPDDL_grounding();
        insert(outcome);
    }
}

void OutcomeGroup::add_new_outcomes(OutcomeGroup& new_outcomes)
{
    insert( new_outcomes.begin(), new_outcomes.end() );
}

void OutcomeGroup::add_empty_outcome_to_fill_probability()
{
    float total_probability = 0.0;
    BOOST_FOREACH(const Outcome& outcome, *this) {
        total_probability += outcome.probability;
    }
    if (total_probability < 1.0) {
        insert(Outcome(1.0 - total_probability));
    }
}

void OutcomeGroup::change_param_types(const std::vector<std::string>& param_names, const std::vector<std::string>& param_types)
{
    std::set<Outcome> old_set;
    swap(old_set);
    BOOST_FOREACH(Outcome outcome, old_set) {
        outcome.change_param_types(param_names, param_types);
        insert(outcome);
    }
}


// NON-CLASS FUNCTIONS

namespace Predicates {
PredicateGroup<TypedAtom> read_predicates_from_file(const std::string& file_path)
{
    LOG(WARNING) << "Deprecated";
    std::string predicates_string;

    std::ifstream predicates_file(file_path.c_str());
    if (predicates_file.is_open()) {
        if (predicates_file.good()) {
            std::getline(predicates_file, predicates_string);
        }
        predicates_file.close();
    } else LOG(ERROR) << "Unable to open file: " << file_path;

    return PredicateGroup<TypedAtom>(predicates_string);
}

PredicateGroup<TypedAtom> transform_to_typed(const PredicateGroup<NonTypedAtom>& non_typed_preds, const PPDDLObjectManager& object_manager)
{
    PredicateGroup<TypedAtom> result;
    BOOST_FOREACH(const Predicate<NonTypedAtom>& non_typed_pred, non_typed_preds) {
        assert(!non_typed_pred.is_symbolic());
        Predicate<TypedAtom> typed_pred(TypedAtom(non_typed_pred.name, std::vector<std::string>(), std::vector<std::string>()), non_typed_pred.positive);
        BOOST_FOREACH(const std::string& param_name, non_typed_pred.get_param_names()) {
            typed_pred.params.push_back(PPDDL::Object(param_name, object_manager.get_object_type(param_name)));
        }
        result.insert(typed_pred);
    }
    return result;
}

PredicateGroup<NonTypedAtom> transform_to_nontyped(const PredicateGroup<TypedAtom>& typed_preds)
{
    PredicateGroup<NonTypedAtom> result;
    BOOST_FOREACH(const Predicate<TypedAtom>& typed_pred, typed_preds) {
        result.insert(Predicate<NonTypedAtom>(NonTypedAtom(typed_pred.name, typed_pred.get_param_names()), typed_pred.positive));
    }
    return result;
}

void update_object_types(PredicateGroup<TypedAtom>& preds_to_update, PPDDLObjectManager& object_manager)
{
    PredicateGroup<TypedAtom> result;
    BOOST_FOREACH(Predicate<TypedAtom> new_pred_updated, preds_to_update) {
        BOOST_FOREACH(PPDDL::Object& param, new_pred_updated.params) {
            param.type = object_manager.get_object_type(param.name);
        }
        result.insert(new_pred_updated);
    }
    result.swap(preds_to_update);
}

// Predicate<Atom> get_predicate_from_string(std::string item)
// {
//     std::string aux;
//     bool positive;
// 
//     item = utils::trim(item);
// //     std::cout << "Generating predicate for: " << item << std::endl;
// 
//     aux = item.substr(0, 1);
//     if (aux.compare("-") == 0) {
//         item = item.substr(1);
//         positive = false;
//     } else
//         positive = true;
// 
//     return Predicate<Atom>(Atom(item), positive);
// }

}


