#include "rules.h"

#define PASULA
#define ENABLE_DEICTIC_RULES


/**
 * OPERATORS
 *
 */

std::ostream& operator<<(std::ostream &out, const Rule& r)
{
    out << "ACTION: # " << r.identifier << "\n";
    out << *static_cast<const TypedAtom*>(&r);
    out << "\n";
    out << "CONTEXT:\n";
    out << r.preconditions;
    out << "\n";
    out << "OUTCOMES:\n";
    for (OutcomeGroup::const_iterator out_it = r.outcomes.begin() ; out_it != r.outcomes.end(); ++out_it) {
        out << *out_it << std::endl;
    }
    if (!r.goals.empty()) {
        out << "REWARD:\n";
        out << r.goals;
    }
    out << std::endl;

    return out;
}


std::ostream& operator<<(std::ostream &out, const RuleSet& r)
{
    for (RuleList::const_iterator rule_it = r.begin() ; rule_it != r.end(); ++rule_it) {
        out << **rule_it;
    }

    // if no rules, write default rule
    if (r.empty()) {
        out << "ACTION:" << std::endl << "  default()" << std::endl
            << "CONTEXT:" << std::endl << "  --" << std::endl
            << "OUTCOMES:" << std::endl << "  0.5 <no-change>" << std::endl << "  0.5 <noise>" << std::endl;
    }

    return out;
}

bool operator==(const Rule &r1, const Rule &r2)
{
    bool result;
    result = (*static_cast<const TypedAtom*>(&r1) == r2);
    result = (result && (r1.deictic_params == r2.deictic_params));
    result = (result && (r1.preconditions == r2.preconditions));
    result = (result && (r1.outcomes == r2.outcomes));
    return result;
}

bool operator<(const Rule& l, const Rule& r)
{
    if (l.is_special() && (!r.is_special())) { // special actions go last
        return false;
    }
    else if ((!l.is_special()) && r.is_special()) {
        return true;
    }
    else if (*static_cast<const TypedAtom*>(&l) < r)
        return true;
    else if (*static_cast<const TypedAtom*>(&r) < l)
        return false;
    else if (l.deictic_params < r.deictic_params)
        return true;
    else if (r.deictic_params < l.deictic_params)
        return false;
    else if (l.outcomes < r.outcomes)
        return true;
    else if (r.outcomes < l.outcomes)
        return false;
    else if (l.preconditions < r.preconditions)
        return true;
    else if (r.preconditions < l.preconditions)
        return false;
    else
        return false;
}

bool operator>(const Rule& l, const Rule& r)
{
    return r < l;
}

bool operator<=(const Rule& l, const Rule& r)
{
    return ((l < r) || (!(l > r)));
}

bool operator>=(const Rule& l, const Rule& r)
{
    return ((l > r) || (!(l < r)));
}

bool operator==(const RuleSet &r1, const RuleSet &r2)
{
    if (r1.size() != r2.size()) {
        return false;
    }
    // make copies to sort them
    RuleSet r1_copy(r1), r2_copy(r2);
    std::sort(r1_copy.begin(), r1_copy.end());
    std::sort(r2_copy.begin(), r2_copy.end());
    RuleSet::iterator r1_it, r2_it;
    r2_it = r2_copy.begin();
    for (r1_it = r1_copy.begin(); r1_it != r1_copy.end(); ++r1_it) {
        if (**r1_it != **r2_it) {
            return false;
        }
        r2_it++;
    }
    return true;
}

bool operator!=(const RuleSet &r1, const RuleSet &r2)
{
    return !(r1 == r2);
}

RuleSet& RuleSet::operator=(const RuleSet& rhs) {
    clear();
    add_rules(rhs);
    return *this;
}

#if __cplusplus > 199711L // C++11
RuleSet& RuleSet::operator=(RuleSet&& rhs) {
    RuleList::operator =(rhs);
    return *this;
}
#endif

// bool operator<(const RuleSet& l, const RuleSet& r );
// bool operator>(const RuleSet& l, const RuleSet& r );
// bool operator<=(const RuleSet& l, const RuleSet& r );
// bool operator>=(const RuleSet& l, const RuleSet& r );

void Rule::write_PPDDL(std::ostream& out) const
{
    out << "  (:action " << name << std::endl;
    // parameters
    out << "   :parameters (";
    write_typed_parameters(out);
    out << ")" << std::endl;
    // preconditions
    out << "   :precondition ";
    if (!deictic_params.empty()) {
        out << "(exists ";
        out << "(";
        BOOST_FOREACH(const PPDDL::Object& deictic_param, deictic_params) {
            out << deictic_param.name << " - " << deictic_param.type << " ";
        }
        out << ") ";
    }
    Predicates::transform_to_nontyped(preconditions).write_PPDDL(out, params);
    if (!deictic_params.empty()) {
        out << " )";
    }
    out << std::endl;
    // effects
    out << "   :effect (and ";
    outcomes.write_PPDDL(out, params);
    out << std::endl;
    goals.write_PPDDL(out, params);
    // TODO REWARDS
    // TODO GLOBAL REWARDS
    out << "           )" << std::endl;
    out << "  )" << std::endl;
}

void RuleSet::write_PPDDL(std::ostream& out) const
{
    BOOST_FOREACH(const RulePtr& rule_ptr, *this) {
        rule_ptr->write_PPDDL(out);
    }
}

/**
 * Rule
 *
 */

Rule::Rule(const Rule& rule) :
    TypedAtom(rule),
    preconditions(rule.preconditions),
    deictic_params(rule.deictic_params),
    outcomes(rule.outcomes),
    goals(rule.goals),
    dangerous(rule.dangerous),
    identifier(rule.identifier)
{
}

Rule::Rule(const TypedAtom &symbol_, const PredicateGroup<TypedAtom> &preconditions_, const OutcomeGroup& outcomes_, const GoalGroup& goals_) :
    TypedAtom(symbol_),
    preconditions(preconditions_),
    outcomes(outcomes_),
    goals(goals_),
    dangerous(false)
{
    update_deictic_params();
    static ulong current_identifier = 0;
    identifier = current_identifier++;
}

void Rule::add_outcome(const PredicateGroup<TypedAtom> predicates, const float probability)
{
    outcomes.insert(Outcome(predicates, probability));
}

void Rule::add_outcome(const Outcome outcome)
{
    outcomes.insert(outcome);
}

double Rule::probability_obtain_predicate(const Predicate<TypedAtom> predicate) const
{
    double result = 0.0;
    for (OutcomeGroup::const_iterator outcome = this->outcomes.begin(); outcome != this->outcomes.end(); outcome++) {
        result += outcome->probability_obtain_predicate(predicate);
    }
    return result;
}

void Rule::sanitize_probabities()
{
    // Adds or reduces probability of the empty outcome to get a sum of probabilities = 1.0
    float sum_probs = 0.0;
    OutcomeGroup::iterator empty_outcome_it = this->outcomes.end();
    for (OutcomeGroup::iterator it = this->outcomes.begin(); it != this->outcomes.end(); it++) {
        sum_probs += it->probability;
        if (!it->is_useful()) {
            empty_outcome_it = it;
        }
    }
    if (sum_probs < 1.0) { // add missing probability to empty outcome
        if (empty_outcome_it == this->outcomes.end()) {
            outcomes.insert(Outcome(1.0 - sum_probs));
        } else {
            Outcome empty_outcome(*empty_outcome_it);
            outcomes.erase(empty_outcome);
            empty_outcome.probability += 1.0 - sum_probs;
            outcomes.insert(empty_outcome);
        }
    } else { // remove exceding probability from first outcome
        Outcome first_outcome(*outcomes.begin());
        outcomes.erase(first_outcome);
        first_outcome.probability -= (sum_probs - 1.0);
        outcomes.insert(first_outcome);
    }
}

bool Rule::has_useful_outcome() const
{
    bool result = false;
    for (OutcomeGroup::const_iterator outcome = this->outcomes.begin(); outcome != this->outcomes.end(); outcome++) {
        if (outcome->changes_predicates()) {
            result = true;
        }
    }
    return result;
}

std::vector<std::string> Rule::get_param_and_deictic_names() const
{
    std::vector<std::string> symbolic_params_with_deictic = get_param_names();
    BOOST_FOREACH(const PPDDL::Object& deictic_param, get_deictic_params()) {
        symbolic_params_with_deictic.push_back(deictic_param.name);
    }
    return symbolic_params_with_deictic;
}


uint Rule::count_covering_transitions_preconditions(const TransitionGroup& transitions) const
{
    int num = 0;
    for (TransitionSet::const_iterator transition_it = transitions.begin_set() ; transition_it != transitions.end_set(); ++transition_it) {
        if (covers_transition_preconditions(*transition_it, transitions.ppddl_object_manager) > 0.0) {
            num += transition_it->get_number_repetitions();
        }
    }
    return num;
}


void Rule::transform_nid_params_to_pddl_params() 
{
    static_cast<TypedAtom*>(&(*this))->transform_nid_params_to_pddl_params();
    preconditions.transform_symbolic_NID_to_symbolic_PDDL();
    outcomes.transform_symbolic_NID_to_symbolic_PDDL();
    goals.transform_symbolic_NID_to_symbolic_PDDL();
    update_deictic_params();
}

void Rule::change_param_types(const std::vector<std::string>& param_names, const std::vector<std::string>& param_types)
{
    static_cast<TypedAtom*>(&(*this))->change_param_types(param_names, param_types);
    preconditions.change_param_types(param_names, param_types);
    outcomes.change_param_types(param_names, param_types);
    goals.change_param_types(param_names, param_types);
    update_deictic_params();
}

bool Rule::is_covered_by_at_least_one_transition(const TransitionGroup& transitions, const bool check_effects, const bool check_changes) const
{
    RuleList grounded_rules = get_pddl_grounded_rules(PredicateGroup<TypedAtom>(), transitions.ppddl_object_manager, true);
    for (TransitionSet::const_iterator transition_it = transitions.begin_set() ; transition_it != transitions.end_set(); ++transition_it) {
        BOOST_FOREACH(const RulePtr& rule_ptr, grounded_rules) {
            if (check_effects) {
                if ( check_changes) {
                    if (!transition_it->get_changes().empty()) { // if changes.empty() -> no changes, no need to check
                        if (rule_ptr->covers_transition_effects(*transition_it, transitions.ppddl_object_manager).get_likelihood(false) > std::numeric_limits<float>::epsilon()) {
                            return true;
                        }
                    }
                }
                else {
                    if (rule_ptr->covers_transition_effects(*transition_it, transitions.ppddl_object_manager).get_likelihood(false) > std::numeric_limits<float>::epsilon()) {
                        return true;
                    }
                }
            }
            else {
                if (rule_ptr->covers_transition_preconditions(*transition_it, transitions.ppddl_object_manager) > 0.0) {
                    return true;
                }
            }
        }
    }
    return false;
}

int Rule::get_number_literals() const
{
    int result = 0;
    result += preconditions.size();
    BOOST_FOREACH(const Outcome& outcome, outcomes) {
        result += outcome.size();
    }
    return result;
}

void Rule::add_negated_outcomes_as_preconditions()
{
    for (OutcomeGroup::const_iterator outcome_it = this->outcomes.begin(); outcome_it != this->outcomes.end(); outcome_it++) {
        for (PredicateGroup<TypedAtom>::const_iterator predicate = outcome_it->begin(); predicate != outcome_it->end(); ++predicate) {
            add_predicate_to_preconditions(predicate->get_negated());
        }
    }
}

void Rule::set_preconditions(const PredicateGroup<TypedAtom>& new_preconditions) 
{ 
    preconditions = new_preconditions; 
    update_deictic_params();
}

void Rule::add_predicate_to_preconditions(const Predicate<TypedAtom>& new_predicate)
{
    BOOST_FOREACH(const PPDDL::Object& param, new_predicate.params) {
        if (symbolic::is_param_symbolic(param.name)) {
            if (std::find(params.begin(), params.end(), param) == params.end() ) {
                deictic_params.insert(PPDDL::Object(param.name, param.type));
            }
        }
    }
    preconditions.update_predicate(new_predicate);
    update_deictic_params();
}

void Rule::update_deictic_params()
{
    deictic_params.clear();
    if (!preconditions.empty()) {
        std::vector<PPDDL::Object> all_vars_v = preconditions.get_params();
        std::set<PPDDL::Object> action_vars_s(params.begin(), params.end());
        std::set<PPDDL::Object> all_vars_s(all_vars_v.begin(), all_vars_v.end());
        
        std::set_difference(all_vars_s.begin(), all_vars_s.end(), 
                            action_vars_s.begin(), action_vars_s.end(),
                            std::inserter(deictic_params, deictic_params.end()));
    }
}

void Rule::add_predicate_to_outcomes(const Predicate<TypedAtom>& new_predicate)
{
    outcomes.add_predicate_to_outcomes(new_predicate);
}

std::vector<PPDDL::Object> Rule::get_deictic_params_v() const
{
    return std::vector<PPDDL::Object>(deictic_params.begin(), deictic_params.end());
}

void Rule::change_variable(const std::string& current_name, const std::string& new_name)
{
    (static_cast<TypedAtom*>(this))->change_param_name(current_name, new_name);
    preconditions.change_param_name(current_name, new_name);
    outcomes.change_param_name(current_name, new_name);
    BOOST_FOREACH(Goal& goal, goals) {
        goal.change_param(current_name, new_name);
    }
}


/**
 * SymbolicRule
 *
 */

SymbolicRule::SymbolicRule(const Rule& rule):
    Rule(rule)
{
}

SymbolicRule::SymbolicRule(const SymbolicRule& rule):
    Rule(rule)
{
}

SymbolicRule::SymbolicRule(const TypedAtom &symbol_, const PredicateGroup<TypedAtom> &preconditions_, const OutcomeGroup& outcomes_, const GoalGroup& goals_):
    Rule(symbol_, preconditions_, outcomes_, goals_)
{
}

float SymbolicRule::covers_transition_preconditions(const Transition& transition, const PPDDLObjectManager& object_manager) const
{
    float covering = 0;
    if ( (transition.get_action().name == name) || (*this == MAKE_TYPED_ATOM_NOACTION) ) {
        RuleSet grounded_deictic_rules = get_pddl_grounded_rules(transition.get_prev_state(), object_manager, false);
        covering = grounded_deictic_rules.covers_transition_preconditions(transition, object_manager);
    }
    return covering;
}

bool SymbolicRule::check_if_rule_groundings_cover_transition_preconditions_more_than_once(const TransitionGroup& transitions) const
{
    for(TransitionSet::const_iterator transition_it = transitions.begin_set(); transition_it != transitions.end_set(); ++transition_it) {
        if ( (transition_it->get_action().name == name) ) { // MAKE_TYPED_ATOM_NOACTION are ignored
            RuleSet grounded_deictic_rules = get_pddl_grounded_rules(transition_it->get_prev_state(), transitions.ppddl_object_manager, false);
            uint number_groundings = 0;
            BOOST_FOREACH(const RulePtr& grounded_deictic_rule_ptr, grounded_deictic_rules) {
                if (transition_it->get_action() == *grounded_deictic_rule_ptr) {
                    ++number_groundings;
                }
            }
            if (number_groundings > 1) {
                return true;
            }
        }
    }
    return false;
}

Transitions::Coverage SymbolicRule::covers_transition_effects(const Transition& transition, const PPDDLObjectManager& object_manager, const bool check_preconditions) const
{
    if ( (transition.get_action().name == name) || (name == ACTION_NAME_NOACTION) ) {
        RuleSet grounded_deictic_rules(get_pddl_grounded_rules(transition.get_prev_state(), object_manager, false));
        // if action (excluding noaction) groundings cover transition preconditions more than once, then rule is bad and coverage error is set as 1
        uint number_groundings = 0;
        BOOST_FOREACH(const RulePtr& grounded_deictic_rule_ptr, grounded_deictic_rules) {
            if (transition.get_action() == *grounded_deictic_rule_ptr) {
                ++number_groundings;
            }
        }
        if (number_groundings > 1) {
            Transitions::Coverage coverage(transition.get_changes());
            coverage.add_error(1.0);
            return coverage;
        }
        else { // everything is ok
            return grounded_deictic_rules.covers_transition_effects(transition, object_manager, false);
        }
    }
    return Transitions::Coverage(transition.get_changes());
}

std::vector<int> SymbolicRule::experiences_per_outcome(const Transition transition, const PPDDLObjectManager& object_manager) const
{
    if ( (transition.get_action().name == name) || (name == ACTION_NAME_NOACTION) ) {
        RuleSet grounded_deictic_rules(get_pddl_grounded_rules(transition.get_prev_state(), object_manager, false));
        // if action (excluding noaction) groundings cover transition preconditions more than once, then rule is bad and coverage error is set as 1
        uint number_groundings = 0;
        BOOST_FOREACH(const RulePtr& grounded_deictic_rule_ptr, grounded_deictic_rules) {
            if (transition.get_action() == *grounded_deictic_rule_ptr) {
                ++number_groundings;
            }
        }
        if (number_groundings == 1) {
            return (*grounded_deictic_rules.begin())->experiences_per_outcome(transition, object_manager);
        }
    }
    return std::vector<int>(outcomes.size(), 0);
}

bool SymbolicRule::satisfiesPreconditions(const PredicateGroup<TypedAtom>& predicate_group, const PPDDLObjectManager& pddl_object_manager) const
{
    RuleList grounded_rules = get_pddl_grounded_rules(predicate_group, pddl_object_manager, false);
    return !grounded_rules.empty();
}

RuleList SymbolicRule::get_pddl_grounded_rules(const State& state, const bool full) const
{
    return get_pddl_grounded_rules(state, state.ppddl_object_manager, full);
}

RuleList SymbolicRule::get_pddl_grounded_rules(const PredicateGroup<TypedAtom>& state, const PPDDLObjectManager& pddl_object_manager, const bool full) const
{
    int DEBUG = 0;
    RuleList grounded_rules;
    
    std::vector<std::string> list_param_names = get_param_names();
    std::vector<std::string> list_param_types = get_param_types();
    std::vector<PPDDL::Object> deictic_params = get_deictic_params_v();
    // Add deictic param names
    BOOST_FOREACH(const PPDDL::Object& deictic_param, deictic_params) {
        list_param_names.push_back(deictic_param.name);
    }
    // Add deictic param types
    BOOST_FOREACH(const PPDDL::Object& deictic_param, deictic_params) {
        list_param_types.push_back(deictic_param.type);
    }
    CLOG_IF(DEBUG >=2, INFO, "trivial") << "Action param types are " << get_param_types();
    CLOG_IF(DEBUG >=2, INFO, "trivial") << "Param types are " << list_param_types;
    PPDDLObjectManager::CombinationsPPDDLObjectsPtr combinations_of_params(pddl_object_manager.get_combinations_of_objecs_for_given_params(list_param_types));
    
    CLOG_IF(DEBUG >=2, INFO, "trivial") << "Combinations of params are " << *combinations_of_params;

    // this vector is used to save copies and speed up
    std::vector<Predicate<TypedAtom> > list_preconditions(get_preconditions().begin(), get_preconditions().end());
    BOOST_FOREACH ( const std::vector<std::string>& a_combination_of_params, *combinations_of_params) {
        if (full) {
            GroundedRulePtr new_rule = boost::make_shared<GroundedRule>(*this, a_combination_of_params, pddl_object_manager);
            grounded_rules.push_back(new_rule);
            CLOG_IF(DEBUG >=3, INFO, "trivial") << "Grounded rules is: " << *new_rule;
        } else {
            // before creating grounded rule, check that the grounded preconditions are satisfied by the state
            bool preconditions_satisfied = true;
            size_t list_preconditions_idx = 0;
            PredicateGroup<TypedAtom>::const_iterator preconditions_it = get_preconditions().begin();
            while (preconditions_satisfied && (preconditions_it != get_preconditions().end())) {
//                 Predicate<TypedAtom> new_precondition = *preconditions_it;
                list_preconditions[list_preconditions_idx].set_param_names(preconditions_it->params);
                list_preconditions[list_preconditions_idx].change_param_names(list_param_names, a_combination_of_params);
                if (!state.satisfies(list_preconditions[list_preconditions_idx])) {
                    preconditions_satisfied = false;
                }
                ++preconditions_it;
                ++list_preconditions_idx;
            }
            
            if (preconditions_satisfied) {
                GroundedRulePtr new_rule = boost::make_shared<GroundedRule>(*this, list_param_names, a_combination_of_params, pddl_object_manager);
                grounded_rules.push_back(new_rule);
                CLOG_IF(DEBUG >=3, INFO, "trivial") << "Grounded rules is: " << *new_rule;
            }
        }
    }

    CLOG_IF(DEBUG >=1, INFO, "trivial") << "Number of grounded rules is " << grounded_rules.size();
    
    return grounded_rules;
}



/**
 * GroundedRule
 *
 */

GroundedRule::GroundedRule(const GroundedRule& grounded_rule):
    Rule(grounded_rule)
{
}

GroundedRule::GroundedRule(const Rule& rule, const std::vector< std::string >& symbolic_param_names, const std::vector< std::string >& grounded_param_names, const PPDDLObjectManager& pddl_object_manager):
    Rule(rule, PredicateGroup<TypedAtom>()) // don't initialize everything for performance reasons
{
    assert(grounded_param_names.size() == symbolic_param_names.size());
    dangerous = rule.dangerous;
    save_ground_preconditions(rule.get_preconditions(), symbolic_param_names, grounded_param_names);
    save_ground_outcomes(rule.outcomes, symbolic_param_names, grounded_param_names);
    save_ground_goals(rule.goals, symbolic_param_names, grounded_param_names, pddl_object_manager);
    
    // update params
    for(size_t i = 0; i < params.size(); ++i) { 
        params[i].name = grounded_param_names[i]; 
    }
    // update_deictic_params(); not needed, it is faster to do it manually
    assert(deictic_params.empty());
    size_t grounded_param_idx = params.size();
    BOOST_FOREACH(const PPDDL::Object& symbolic_deictic_param, rule.get_deictic_params()) {
        deictic_params.insert(PPDDL::Object(grounded_param_names[grounded_param_idx], symbolic_deictic_param.type));
        ++grounded_param_idx;
    }
}

#if __cplusplus > 199711L
GroundedRule::GroundedRule(const Rule& rule, const std::vector< std::string >& grounded_param_names, const PPDDLObjectManager& pddl_object_manager):
    GroundedRule(rule, rule.get_param_and_deictic_names(), grounded_param_names, pddl_object_manager)
{ }
#else
GroundedRule::GroundedRule(const Rule& rule, const std::vector< std::string >& grounded_param_names, const PPDDLObjectManager& pddl_object_manager):
    Rule(rule, PredicateGroup<TypedAtom>()) // don't initialize everything for performance reasons
{
    vector< string > symbolic_param_names = rule.get_param_names();
    std::vector<PPDDL::Object> symbolic_deictic_params = rule.get_deictic_params_v();
    BOOST_FOREACH(const PPDDL::Object& deictic_param, symbolic_deictic_params) {
        symbolic_param_names.push_back(deictic_param.name);
    }
    
    assert(grounded_param_names.size() == symbolic_param_names.size());
    dangerous = rule.dangerous;
    save_ground_preconditions(rule.get_preconditions(), symbolic_param_names, grounded_param_names);
    save_ground_outcomes(rule.outcomes, symbolic_param_names, grounded_param_names);
    save_ground_goals(rule.goals, symbolic_param_names, grounded_param_names, pddl_object_manager);
    
    // update params
    for(size_t i = 0; i < params.size(); ++i) { 
        params[i].name = grounded_param_names[i]; 
    }
    // update_deictic_params(); not needed, it is faster to do it manually
    assert(deictic_params.empty());
    size_t grounded_param_idx = params.size();
    BOOST_FOREACH(const PPDDL::Object& symbolic_deictic_param, rule.get_deictic_params()) {
        deictic_params.insert(PPDDL::Object(grounded_param_names[grounded_param_idx], symbolic_deictic_param.type));
        ++grounded_param_idx;
    }
}
#endif

RuleList GroundedRule::get_pddl_grounded_rules(const State& state, const bool full) const
{
    return get_pddl_grounded_rules(state, state.ppddl_object_manager, full);
}

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-parameter"
RuleList GroundedRule::get_pddl_grounded_rules(const PredicateGroup<TypedAtom>& pred_group, const PPDDLObjectManager& pddl_object_manager, const bool full) const
{
    RuleList result;
    result.push_back(boost::make_shared<GroundedRule>(*this));
    return result;
}
#pragma GCC diagnostic pop


void GroundedRule::save_ground_preconditions(const PredicateGroup<TypedAtom>& symbolic_preconditions, const std::vector<std::string>& used_symbolic_params, const std::vector<std::string>& used_grounded_params)
{
    BOOST_FOREACH(Predicate<TypedAtom> new_precondition, symbolic_preconditions) {
        new_precondition.change_param_names(used_symbolic_params, used_grounded_params);
        preconditions.insert(new_precondition);
    }
}

void GroundedRule::save_ground_outcomes(const OutcomeGroup& symbolic_outcomes, const std::vector<std::string>& used_symbolic_params, const std::vector<std::string>& used_grounded_params)
{
    BOOST_FOREACH(const Outcome& symbolic_outcome, symbolic_outcomes) {
        Outcome new_outcome(symbolic_outcome.probability, symbolic_outcome.noisy);
        BOOST_FOREACH(Predicate<TypedAtom> new_outcome_pred, symbolic_outcome) {
            new_outcome_pred.change_param_names(used_symbolic_params, used_grounded_params);
            new_outcome.insert(new_outcome_pred);
        }
        outcomes.insert(new_outcome);
    }
}

void GroundedRule::save_ground_goals(const GoalGroup& symbolic_goals, const std::vector<std::string>& used_symbolic_params, const std::vector<std::string>& used_grounded_params, const PPDDLObjectManager& pddl_object_manager)
{
    goals = symbolic_goals.get_pddl_grounded_goals(used_symbolic_params, used_grounded_params, pddl_object_manager);
}

float GroundedRule::covers_transition_preconditions(const Transition& transition, const PPDDLObjectManager& object_manager) const
{
    if ( ((transition.get_action() == *this) || (*this == MAKE_TYPED_ATOM_NOACTION)) && satisfiesPreconditions(transition.get_prev_state())) {
        return 1;
    }
    else {
        return 0;
    }
}

Transitions::Coverage GroundedRule::covers_transition_effects(const Transition& transition, const PPDDLObjectManager& object_manager, const bool check_preconditions) const
{
    Transitions::Coverage result(transition.get_changes());
    if ( ( (name == ACTION_NAME_NOACTION) || (transition.get_action() == *this) ) && 
         ( (!check_preconditions) || 
           preconditions.is_satisfied_by(transition.get_prev_state()) // equivalent to (but faster): satisfiesPreconditions(transition.get_prev_state())
         ) ) 
    {
        float possible_error_probability = 0.0; // only to be considered when rule does not cover any change
        float real_error_probability = 0.0;
        std::vector<float> vector_sum_probabilities(transition.get_changes().size());
        BOOST_FOREACH(const Outcome& outcome, outcomes) {
            float error_probability = 0;
            bool outcome_covers_a_change = false;
            int change_idx = 0;
            BOOST_FOREACH(const Predicate<TypedAtom>& change, transition.get_changes()) {
                if (outcome.has_grounded_predicate(change)) {
                    vector_sum_probabilities[change_idx] += outcome.probability;
                    outcome_covers_a_change = true;
                }
                ++change_idx;
            }
            if (!outcome.is_satisfied_by(transition.get_post_state())) {
                // something is wrong in outcome_it!!!
                error_probability += outcome.probability;
            }
            if (outcome_covers_a_change) {
                real_error_probability += error_probability;
            }
            else {
                possible_error_probability += error_probability;
            }
        }
        
        bool rule_adds_coverage = false;
        for(size_t change_idx = 0; change_idx < transition.get_changes().size(); ++change_idx) {
            if (vector_sum_probabilities[change_idx] > std::numeric_limits<float>::epsilon()) {
                result.add_change_coverage(change_idx, vector_sum_probabilities[change_idx]);
                rule_adds_coverage = true;
            }
        }
        if (rule_adds_coverage) { // only add error if rule didn't cover effects (otherwise this error is implicit!)
            result.add_error(real_error_probability);
        }
        else {
            result.add_error(possible_error_probability);
        }
        return result;
    }
    else {
        return result;
    }
}

float RuleSet::covers_transition_preconditions(const Transition& transition, const PPDDLObjectManager& object_manager) const
{
    float covering = 0.0;
    int n_covered_rules = 0;
    int n_action_covered_rules = 0;
    BOOST_FOREACH(RulePtr rule_ptr, *this) {
        if ( (*rule_ptr == transition.get_action()) || (*rule_ptr == MAKE_TYPED_ATOM_NOACTION) ) {
            covering += rule_ptr->covers_transition_preconditions(transition, object_manager);
            ++n_covered_rules;
            if (*rule_ptr != MAKE_TYPED_ATOM_NOACTION) {
                ++n_action_covered_rules;
            }
        }
    }
    if (n_covered_rules > 1) {
        covering = covering / (float)n_covered_rules; // if more than 2 rules cover it, it may be bigger than 1.0
//         if (n_action_covered_rules > 1) { // Now it happens with LFIT-based learner rules (there is one rule per effect)
//             LOG(WARNING) << "Bad defined rules, more than 1 cover the same transition";
//             LOG(WARNING) << *this;
//             covering = 0.0;
//         }
    }
    return covering;
}

Transitions::Coverage RuleSet::covers_transition_effects(const Transition& transition, const PPDDLObjectManager& object_manager, const bool check_preconditions) const
{
    Transitions::Coverage covering(transition.get_changes());
    BOOST_FOREACH(RulePtr rule_ptr, *this) {
        if ( (rule_ptr->is_symbolic() && (rule_ptr->symbolic_compare(transition.get_action()))) || 
             (*rule_ptr == transition.get_action()) || 
             (*rule_ptr == MAKE_TYPED_ATOM_NOACTION) )
        {
            covering = covering + rule_ptr->covers_transition_effects(transition, object_manager, check_preconditions);
        }
    }
    return covering;
}

std::vector<int> GroundedRule::experiences_per_outcome(const Transition transition, const PPDDLObjectManager& object_manager) const
{
    // supposes that preconditions are satisfied
    std::vector<int> result;

    for (OutcomeGroup::const_iterator outcome_it = this->outcomes.begin(); outcome_it != this->outcomes.end(); outcome_it++) {
        if (*outcome_it == transition.get_changes()) {
            result.push_back(1);
        } else {
            result.push_back(0);
        }
    }

    return result;
}

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-parameter"
bool GroundedRule::satisfiesPreconditions(const PredicateGroup<TypedAtom>& predicates, const PPDDLObjectManager& pddl_object_manager) const
{
    if (preconditions.is_satisfied_by(predicates))
        return true;
    else
        return false;
}
#pragma GCC diagnostic pop

bool GroundedRule::satisfiesPreconditions(const PredicateGroup<TypedAtom>& predicates) const
{
    return satisfiesPreconditions(predicates, PPDDLObjectManager());
}



/**
 * RuleSet
 *
 */

RuleSet::RuleSet()
{ }

#if __cplusplus > 199711L
RuleSet::RuleSet(RuleSet&& rules_):
    RuleList(rules_) 
    {};
    
RuleSet::RuleSet(RuleList&& rules_):
    RuleList(rules_) 
    {};
#endif

RuleSet::RuleSet(const RuleSet& rules_)
{
    add_rules(rules_);
}

RuleSet::RuleSet(const RuleList& rules_)
{
    add_rules(rules_);
}

// RuleSet::RuleSet(std::string file_path)
// {
//     read_rules_from_file(file_path);
// }

void RuleSet::add_rules(const RuleList& new_rules)
{
    BOOST_FOREACH(const RuleConstPtr& rule_ptr, new_rules) {
        push_back(rule_ptr->make_shared());
    }
}

void RuleSet::read_NID_rules(std::string file_path, const Domain& domain)
{
    // DEPRECATED
    std::string line;
    std::ifstream rules_file(file_path.c_str());

    // Clear previous rules
    clear();

    if (rules_file.is_open()) {
        std::string line_content;
        std::string line_content2;
        std::string line_title;

        getline(rules_file, line_title);
        while (rules_file.good()) {
            // ACTION:
            // read until ACTION: found
            while (line_title.find("ACTION:") == std::string::npos) {
                getline(rules_file, line_title);
            }
            
            boost::optional<uint> rule_id;
            // Get action id from ACTION: #X title
            if (line_title.find("#") != std::string::npos) { // if has id
                std::string ss_item;
                std::stringstream line_title_ss(line_title);
                std::getline(line_title_ss, ss_item, '#');
                std::getline(line_title_ss, ss_item, '#');
                try {
                    ss_item = utils::trim(ss_item);
                    rule_id = boost::lexical_cast<uint>(ss_item);
                } catch (boost::bad_lexical_cast &) {
                    // ignore
                    LOG(WARNING) << "Error when loading rule. Is there a # in a \"Action:\" line?";
                    LOG(WARNING) << "Rule line is " << line_title;
                }
            }

            // Action name
            getline(rules_file, line_content);

            // CONTEXT:
            getline(rules_file, line_title);

            // Preconditions
            getline(rules_file, line_content2);

            NonTypedAtom action = NonTypedAtom(utils::trim(line_content));
            PredicateGroup<NonTypedAtom> preconditions(line_content2);
            SymbolicRule rule = SymbolicRule(domain.transform_to_typed(action),
                                             domain.transform_to_typed(preconditions));
            if (rule_id) {
                rule.identifier = *rule_id;
            }

            // OUTCOMES:
            getline(rules_file, line_title);

            // Outcome i
            getline(rules_file, line_content);
            while ((line_content != "") && (line_content != "REWARD:")) {
                Outcome outcome_non_typed(line_content);
                Outcome outcome(outcome_non_typed.probability, outcome_non_typed.noisy);
                outcome.add_predicates(domain.transform_to_typed(Predicates::transform_to_nontyped(outcome_non_typed)));
                rule.add_outcome(outcome);
                getline(rules_file, line_content);
                line_content = utils::trim(line_content);
            }

            // check if rule has rewards
            line_title = line_content;
            if (line_title == "REWARD:") {
                getline(rules_file, line_content);
                line_content = utils::trim(line_content);
                while (line_content != "") {
                    // Add rule preconditions to rule goals
                    Goal goal(line_content);
                    goal.add_predicates(rule.get_preconditions());
                    rule.goals.push_back(goal);
                    // Get next line
                    getline(rules_file, line_content);
                    line_content = utils::trim(line_content);
                }
            }

            // Create rule
            if (rule.name != DEFAULT_ACTION_NAME) {
                push_back(boost::make_shared<SymbolicRule>(rule));
            }

            // read lines until not empty line
            getline(rules_file, line_title);
            line_title = utils::remove_spaces(line_title);
            while ((not rules_file.eof()) && rules_file.good() && (line_title.compare("") == 0)) {
                getline(rules_file, line_title);
            }
        }
    }
    else {
        LOG(ERROR) << "Couldn't open file to read rules: \"" << file_path << "\"";
    }

    sanitize_probabities();
    transform_symbolic_pasula_to_symbolic_PDDL();

//         std::cout << "Read rules are " << *this << std::endl;
    rules_file.close();
}

void RuleSet::write_rules_to_file(std::string file_path) const
{
    int DEBUG = 0;
    std::ofstream rules_file(file_path.c_str());
    if (DEBUG > 0)
        std::cout << "RuleSet::write_rules_to_file - " << "Writting rules to file: '" << file_path << "'" << std::endl;
    if (rules_file.is_open()) {
        rules_file << *this;
    } else {
        std::cerr << "Couldn't write rules to " << file_path << std::endl;
    }
    rules_file.close();
}


void RuleSet::sanitize_probabities()
{
    for (RuleList::iterator rule = begin() ; rule != end(); ++rule) {
        (*rule)->sanitize_probabities();
    }
}


double RuleSet::can_obtain_predicate(const Predicate<TypedAtom> predicate) const
{
    double result = 0.0;
    for (RuleList::const_iterator rule = begin() ; rule != end(); ++rule) {
        result += (*rule)->probability_obtain_predicate(predicate);
    }
    return result;
}


bool RuleSet::remove_rule(const RulePtr rule_to_remove)
{
    bool result = false;
    for (RuleList::iterator rule_it = begin() ; rule_it != end();) {
        if ((*rule_to_remove) == (**rule_it)) {
            result = true;
            rule_it = erase(rule_it);
        } else {
            ++rule_it;
        }
    }
    return result;
}


RuleSet RuleSet::get_action_rules(const TypedAtom& action) const
{
    RuleSet action_rules;

    BOOST_FOREACH(const RulePtr& rule_it, *this) {
        if ((*rule_it) == action) {
            action_rules.push_back(rule_it);
        }
    }
    return action_rules;
}


RuleSet RuleSet::get_grounded_action_rules(const TypedAtom& grounded_action, const State& state) const
{
    RuleSet grounded_action_rules;
    for (RuleSet::const_iterator rule_it = begin() ; rule_it != end(); ++rule_it) {
        GroundedRulePtr grounded_rule_ptr;
        if ((!(*rule_it)->is_symbolic()) && ((**rule_it) == grounded_action)) {  // if grounded and actions are equal
            grounded_rule_ptr = boost::make_shared<GroundedRule>(*static_cast<const GroundedRule*>(&(**rule_it)));
            if (grounded_rule_ptr) {
                grounded_action_rules.push_back(grounded_rule_ptr);
            }
        } else if (((*rule_it)->is_symbolic()) && ((*rule_it)->name == grounded_action.name)) { // if symbolic and actions names are equal (not arguments)
#ifdef ENABLE_DEICTIC_RULES
            // slow, consider deictic
            RuleList rule_it_grounded_rules = (*rule_it)->get_pddl_grounded_rules(state, false);
            BOOST_FOREACH(RulePtr grounded_rule_ptr, rule_it_grounded_rules) {
                if (*grounded_rule_ptr == grounded_action) {
                    grounded_action_rules.push_back(grounded_rule_ptr);
                }
            }
#else
            // fast -> no deictic
            grounded_rule_ptr = boost::make_shared<GroundedRule>(GroundedRule(**rule_it, grounded_action.param_names, state.pddl_object_manager)); 
            // TODO fast deictic param grounding
            if (grounded_rule_ptr) {
                // std::vector<std::string> deictic_params = grounded_rule_ptr->get_deictic_params();
                // do something with deitic params
                grounded_action_rules.push_back(grounded_rule_ptr);
            }
#endif
        }
    }
    return grounded_action_rules;
}


RuleSet RuleSet::remove_action_rules(TypedAtom action)
{
    RuleSet removed_rules;

    for (RuleList::iterator rule = begin() ; rule != end();) {
        if ((*rule)->symbolic_compare(action)) {
            removed_rules.push_back(*rule);
            rule = erase(rule);
        } else {
            ++rule;
        }
    }
    return removed_rules;
}

void RuleSet::remove_nop_rules()
{
    for (RuleList::iterator rule = begin() ; rule != end();) {
        if ( (*rule)->has_useful_outcome() || (!(*rule)->goals.empty()) ) {
            ++rule;
        } else {
            rule = erase(rule);
        }
    }
}

void RuleSet::remove_special_rules()
{
    for (RuleList::iterator rule = begin() ; rule != end();) {
        if ( (*rule)->is_special()) {
            rule = erase(rule);
        } else {
            ++rule;
        }
    }
}

void Rule::analyze_differences(const Rule& compared_rule, int& preconditions_diff_n, int& outcomes_diff_n,  double& probabilities_diff) const
{
    // preconditions
    PredicateGroup<TypedAtom> preconditions_diff = preconditions.get_differences(compared_rule.preconditions);
    preconditions_diff_n = preconditions_diff.size();

    // outcomes
    outcomes_diff_n = outcomes.size() - compared_rule.outcomes.size();

    // probabilities
    probabilities_diff = outcomes.begin()->probability - compared_rule.outcomes.begin()->probability;
}


void RuleSet::analyze_differences(RuleSet compared_ruleset, int& rule_number_diff, int& preconditions_diff, int& outcomes_num_diff, double& first_outcome_prob_diff)
{
    preconditions_diff = 0;
    outcomes_num_diff = 0;
    first_outcome_prob_diff = 0;
    rule_number_diff = - size();
    for (RuleList::iterator rule_it = begin() ; rule_it != end(); ++rule_it) {
        preconditions_diff -= (*rule_it)->get_preconditions().size();
        outcomes_num_diff -= (*rule_it)->outcomes.size();
        first_outcome_prob_diff -= (*rule_it)->outcomes.begin()->probability;
    }

    rule_number_diff += compared_ruleset.size();
    for (RuleList::iterator rule_it = compared_ruleset.begin() ; rule_it != compared_ruleset.end(); ++rule_it) {
        preconditions_diff += (*rule_it)->get_preconditions().size();
        outcomes_num_diff += (*rule_it)->outcomes.size();
        first_outcome_prob_diff += (*rule_it)->outcomes.begin()->probability;
    }
}


void RuleSet::analyze_differences(RuleSet compared_ruleset, RuleSet& added_rules, RuleSet& removed_rules,
                                  RuleSet& original_changed_rules, RuleSet& compared_changed_rules,
                                  std::vector<int>& preconditions_diff, std::vector<int>& outcomes_diff,  std::vector<double>& probabilities_diff
                                 ) const
{
    RuleSet original_rules(*this);
    std::sort(original_rules.rbegin(), original_rules.rend());  // reverse iterators! -> sorting bigger -> smaller
    std::sort(compared_ruleset.rbegin(), compared_ruleset.rend());  // reverse iterators! -> sorting bigger -> smaller

    RuleSet::iterator rule_it = original_rules.begin();
    RuleSet::iterator comp_rule_it = compared_ruleset.begin();
    while ((rule_it != original_rules.end()) && (comp_rule_it != compared_ruleset.end())) {
        if (**rule_it == **comp_rule_it) {
            preconditions_diff.push_back(0);
            outcomes_diff.push_back(0);
            probabilities_diff.push_back(0.0);
            original_changed_rules.push_back(*rule_it);
            compared_changed_rules.push_back(*comp_rule_it);
            ++rule_it;
            ++comp_rule_it;
        } else if (**rule_it > **comp_rule_it) {
            if (((rule_it + 1) != original_rules.end()) && (**(rule_it + 1) >= **comp_rule_it)) {   // exists rule_it+1 && rule_it+1 >= comp_rule_it
                removed_rules.push_back(*rule_it);
                ++rule_it;
            } else { // compare both
                int rule_preconditions_diff, rule_outcomes_diff;
                double rule_probabilities_diff;
                (*rule_it)->analyze_differences(**comp_rule_it, rule_preconditions_diff, rule_outcomes_diff, rule_probabilities_diff);
                preconditions_diff.push_back(rule_preconditions_diff);
                outcomes_diff.push_back(rule_outcomes_diff);
                probabilities_diff.push_back(rule_probabilities_diff);
                original_changed_rules.push_back(*rule_it);
                compared_changed_rules.push_back(*comp_rule_it);
                ++rule_it;
                ++comp_rule_it;
            }
        } else { // if (**comp_rule_it > **rule_it) {
            if (((comp_rule_it + 1) != compared_ruleset.end()) && (**(comp_rule_it + 1) >= **rule_it)) {     // exists comp_rule_it+1 && comp_rule_it+1 >= rule_it
                added_rules.push_back(*comp_rule_it);
                ++comp_rule_it;
            } else { // compare both
                int rule_preconditions_diff, rule_outcomes_diff;
                double rule_probabilities_diff;
                (*rule_it)->analyze_differences(**comp_rule_it, rule_preconditions_diff, rule_outcomes_diff, rule_probabilities_diff);
                preconditions_diff.push_back(rule_preconditions_diff);
                outcomes_diff.push_back(rule_outcomes_diff);
                probabilities_diff.push_back(rule_probabilities_diff);
                original_changed_rules.push_back(*rule_it);
                compared_changed_rules.push_back(*comp_rule_it);
                ++rule_it;
                ++comp_rule_it;
            }
        }
    }

    // add remaining rules
    while (rule_it != original_rules.end()) {
        removed_rules.push_back(*rule_it);
        ++rule_it;
    }
    while (comp_rule_it != compared_ruleset.end()) {
        added_rules.push_back(*comp_rule_it);
        ++comp_rule_it;
    }
}

void RuleSet::set_identifiers(RuleSetPtr old_rules_ptr)
{
    // Reasons to reverse sort:
    //  - Important rules are those with outcomes
    //  - If rules with outcomes, the identifier will asign better identifiers to rules with outcomes
    //  - Actually, this method could be improved a lot using distances instead of < relations, but not worth the time
    //  as identfiers are used just for gource logging.
    std::sort(rbegin(), rend());  // reverse iterators! -> sorting bigger -> smaller
    std::sort(old_rules_ptr->rbegin(), old_rules_ptr->rend());  // reverse iterators! -> sorting bigger -> smaller

    RuleSet::iterator rule_it = begin();
    RuleSet::iterator old_rule_it = old_rules_ptr->begin();
    while ((rule_it != end()) && (old_rule_it != old_rules_ptr->end())) {
        if (**rule_it == **old_rule_it) {
            (*rule_it)->identifier = (*old_rule_it)->identifier;
            ++rule_it;
            ++old_rule_it;
        } else if (**rule_it > **old_rule_it) {
            if (((rule_it + 1) != end()) && (**(rule_it + 1) >= **old_rule_it)) {     // exists rule_it+1 && rule_it+1 >= old_rule_it
                ++rule_it;
            } else { // just similar
                (*rule_it)->identifier = (*old_rule_it)->identifier;
                ++rule_it;
                ++old_rule_it;
            }
        } else { // if (**old_rule_it > **rule_it) {
            if (((old_rule_it + 1) != old_rules_ptr->end()) && (**(old_rule_it + 1) >= **rule_it)) {     // exists old_rule_it+1 && old_rule_it+1 >= rule_it
                ++old_rule_it;
            } else { // just similar
                (*rule_it)->identifier = (*old_rule_it)->identifier;
                ++rule_it;
                ++old_rule_it;
            }
        }
    }
}

RuleSet RuleSet::find_rules_with_a_precondition(const Predicate<TypedAtom>& precondition) const
{
    RuleSet result;
    for (RuleList::const_iterator rule_it = begin() ; rule_it != end(); ++rule_it) {
        if ((*rule_it)->get_preconditions().has_predicate(precondition)) {
            result.push_back(*rule_it);
        }
    }
    return result;
}

RuleSet RuleSet::find_rules_with_a_precondition(const PredicateGroup<TypedAtom>& possible_preconditions) const
{
    RuleSet result;
    for (PredicateGroup<TypedAtom>::const_iterator pred_it = possible_preconditions.begin() ; pred_it != possible_preconditions.end(); ++pred_it) {
        result.add_rules(find_rules_with_a_precondition(*pred_it));
    }
    return result;
}

std::vector<TypedAtom> RuleSet::get_actions() const
{
    std::vector<TypedAtom> result;
    std::set<TypedAtom> action_atoms;
    BOOST_FOREACH(RulePtr rule_ptr, *this) {
        action_atoms.insert(*rule_ptr);
    }
    // std::set to std::vector
    result.assign(action_atoms.begin(), action_atoms.end());
    return result;
}

std::vector<std::string> RuleSet::get_action_names() const
{
    std::vector<std::string> result;
    std::set<std::string> action_names;
    BOOST_FOREACH(RulePtr rule_ptr, *this) {
        action_names.insert(rule_ptr->name);
    }
    // std::set to std::vector
    result.assign(action_names.begin(), action_names.end());
    return result;
}

double RuleSet::score(const TransitionGroup transitions, const PPDDLObjectManager& object_manager) const
{
    double cutting_threshold = -1. * std::numeric_limits<double>::max();
    return score(transitions, object_manager, cutting_threshold);
}

double RuleSet::score(const TransitionGroup transitions, const PPDDLObjectManager& object_manager, const double cutting_threshold) const
{
    std::vector<float> experience_weights(transitions.size_historial());
    std::fill(experience_weights.begin(), experience_weights.end(), 1);
    return score(transitions, object_manager, cutting_threshold, experience_weights);
}

double RuleSet::score(const TransitionGroup transitions, const PPDDLObjectManager& object_manager, const double cutting_threshold, const std::vector<float> experience_weights) const
{
    uint DEBUG = 0;
    if (DEBUG > 0) {
        std::cout << "SCORE [start]" << std::endl;
    }
    if (DEBUG > 1) {
        std::cout << "Cutting threshold is " << cutting_threshold << std::endl;
    }

    // learning parameters
    double alpha = 0.5;
    double p_min = 1e-9;
    double p_min_noisyDefaultRule = 1e-11;

    // (1) Penalty
    double penalty = 0.0;
    penalty += get_number_literals();
    penalty *= alpha;
    if (DEBUG > 0) {
        std::cout << "alpha: " << alpha << std::endl;
        std::cout << "penalty: " << penalty << std::endl;
    }

    // (2) Log-Likelihood
    double loglikelihood = 0.0, exLik;
    if (DEBUG > 0) {
        std::cout << "Calculating likelihoods..." << std::endl;
    }

    uint i = 0;
    for (TransitionGroup::TransitionHistorial::const_iterator trans_it_it = transitions.begin_historial(); 
         trans_it_it != transitions.end_historial(); 
         ++trans_it_it) {
        if (DEBUG > 1) {
            std::cout << "+++ ex " << i << ": ";
            std::cout << (*trans_it_it)->get_action() << std::endl;
            std::cout << std::endl;
        }

        RuleSet covering_rules = get_covering_rules(**trans_it_it, object_manager);

        if (DEBUG > 1) {
            std::cout << "Covering rules are " << covering_rules.size() << ": " << std::endl;
            std::cout << covering_rules << std::endl;
        }

        // only one non-default rule covers (good case)
        if (covering_rules.size() == 1) {
            RulePtr rule_ptr = covering_rules.front();
            if (DEBUG > 2) {
                std::cout << "Use rule #" << *(covering_rules.front()) << std::endl;
            }

            std::vector<int> exp_per_out = rule_ptr->experiences_per_outcome(**trans_it_it, object_manager);
            if (DEBUG > 2) {
                std::cout << "Experiences per rule outcome " << std::endl;
                for (size_t k = 0; k < exp_per_out.size(); k++)
                    std::cout << exp_per_out[k] << ", ";
                std::cout << std::endl;
            }
            exLik = 0.0;
            OutcomeGroup::iterator out_it = rule_ptr->outcomes.begin();
            for (size_t j = 0; j < exp_per_out.size(); j++) {
                if ((exp_per_out[j] == 1) && (!out_it->noisy)) {
                    exLik += out_it->probability;
                    if (DEBUG > 2) {
                        std::cout << "after o=" << j << ": " << exLik << std::endl;
                    }
                }
                // Noise-outcome:  always covers
                if (out_it->noisy) {
                    exLik += p_min * out_it->probability;
                    if (DEBUG > 2) {
                        std::cout << "after o=" << j << ": " << exLik << std::endl;
                    }
                }
                ++out_it;
            }
            if (exLik <= 0.) {
                if (DEBUG > 0) {
                    std::cerr << "ERROR: " << "bad referencing  exLik=" << exLik << std::endl;
                }
                std::cerr << "!";
                exLik += p_min_noisyDefaultRule * 0.01;
            }
        }
        // only default covers or more than one non-default covers
        // --> apply default rule
        else {
            /*
            ACTION:
            default()
            CONTEXT:
            --
            OUTCOMES:
            0.5 <no-change>
            0.5 <noise>
             */
            if (DEBUG > 2) {
                std::cout << " Using default rule." << std::endl;
            }
            if ((*trans_it_it)->nothing_changed())
                exLik = 0.5 * (100.0 * p_min_noisyDefaultRule);   // Avoid modeling persistence by means of noise default rule.
            else
                exLik = 0.5 * p_min_noisyDefaultRule;
        }

        loglikelihood += experience_weights[i] * log(exLik); // weight the experiences differently
        if (DEBUG > 1) {
            std::cout << " --> lik=" << exLik << std::endl;
        }
        if (DEBUG > 0) {
            std::cout << "log likelihoood is " << loglikelihood << std::endl;
        }


        if (loglikelihood - penalty  <  cutting_threshold) {
            if (DEBUG > 0) {
                std::cout << "Difference between loglikelihood and penalty is " << loglikelihood - penalty << std::endl;
                std::cout << "Cutting threshold is " << cutting_threshold << std::endl;
                std::cout << "Score becomes too small... giving up." << std::endl;
                std::cout << "SCORE [end]" << std::endl;
            }
            return (-1. * std::numeric_limits<double>::max());
        }
        i++;
    }

    double score = loglikelihood - penalty;
    if (DEBUG > 0) {
        std::cout << "score is " << score << std::endl;
        std::cout << "SCORE [end]" << std::endl;
    }

    return score;
}

int RuleSet::get_number_literals() const
{
    int result = 0;
    for (RuleList::const_iterator rule_it = begin() ; rule_it != end(); ++rule_it) {
        result += (*rule_it)->get_number_literals();
    }
    return result;
}

RuleSet RuleSet::get_covering_rules(const Transition& transition, const PPDDLObjectManager& object_manager) const
{
    RuleSet result;
    for (RuleList::const_iterator rule_it = begin() ; rule_it != end(); ++rule_it) {
        if ((*rule_it)->covers_transition_preconditions(transition, object_manager) > 0.0) {
            result.push_back(*rule_it);
        }
    }

    return result;
}

RuleSet RuleSet::get_covering_rules(const PredicateGroup<TypedAtom>& predicates, const PPDDLObjectManager& pddl_object_manager) const
{
    RuleSet result;
    BOOST_FOREACH(const RulePtr& rule_ptr, *this) {
        if (rule_ptr->satisfiesPreconditions(predicates, pddl_object_manager)) {
            result.push_back(rule_ptr);
        }
    }

    return result;
}

RuleSet RuleSet::get_pddl_grounded_rules(const State& state, const bool full) const
{
    RuleSet grounded_rules;

    for (RuleList::const_iterator rule_it = begin() ; rule_it != end(); ++rule_it) {
        RuleList new_grounded_rules = (*rule_it)->get_pddl_grounded_rules(state, full);
        grounded_rules.add_rules(new_grounded_rules);
    }

    return grounded_rules;
}

void RuleSet::set_dangerous_rules(const PredicateGroup<TypedAtom>& dead_end_excuses)
{
    BOOST_FOREACH(const RulePtr& rule_ptr, *this) {
        for (PredicateGroup<TypedAtom>::const_iterator dead_end_excuse_it = dead_end_excuses.begin() ; dead_end_excuse_it != dead_end_excuses.end(); ++dead_end_excuse_it) {
            BOOST_FOREACH(const Outcome& outcome, rule_ptr->outcomes) {
                if (outcome.has_predicate(dead_end_excuse_it->get_negated()) && (outcome.probability > 0.0)) {
                    // outcome.is_dangerous = true
                    Outcome dangerous_outcome(outcome);
                    rule_ptr->outcomes.erase(outcome);
                    dangerous_outcome.is_dangerous = true;
                    rule_ptr->outcomes.insert(dangerous_outcome);
                    
                    rule_ptr->dangerous = true;
                }
            }
        }
    }
}

bool RuleSet::is_dangerous(const TypedAtom action, const State& state) const
{
    RuleSet action_rules = get_grounded_action_rules(action, state);
    RuleSet covering_rules = action_rules.get_covering_rules(state, state.ppddl_object_manager);
    for (RuleList::iterator rule_it = covering_rules.begin() ; rule_it != covering_rules.end(); ++rule_it) {
        if ((*rule_it)->dangerous) {
            return true;
        }
    }
    return false;
}

void RuleSet::add_negated_outcomes_as_preconditions()
{
    for (RuleList::iterator rule_it = begin() ; rule_it != end(); ++rule_it) {
        (*rule_it)->add_negated_outcomes_as_preconditions();
    }
}

void RuleSet::add_predicate_to_preconditions(const Predicate<TypedAtom>& new_predicate)
{
    for (RuleList::iterator rule_it = begin() ; rule_it != end(); ++rule_it) {
        (*rule_it)->add_predicate_to_preconditions(new_predicate);
    }
}

void RuleSet::add_predicate_to_outcomes(const Predicate<TypedAtom>& new_predicate)
{
    BOOST_FOREACH(RulePtr rule, *this) {
        rule->add_predicate_to_outcomes(new_predicate);
    }
}

void RuleSet::add_goal(Goal goal)
{
    for (RuleSet::iterator rule_it = begin(); rule_it != end(); ++rule_it)  {
        (*rule_it)->add_goal(goal);
    }
}

void RuleSet::transform_symbolic_pasula_to_symbolic_PDDL()
{
    BOOST_FOREACH(RulePtr& rule_ptr, *this) {
        rule_ptr->transform_nid_params_to_pddl_params();
    }
}


/**
 * PPDDL
 */

void RuleSet::PPDDL_grounding()
{
    for (RuleList::iterator rule_it = begin(); rule_it != end(); ++rule_it) {
        (*rule_it)->PPDDL_grounding();
    }
}

void Rule::PPDDL_grounding()
{
    preconditions.PPDDL_grounding();
    outcomes.PPDDL_grounding();
    for (GoalGroup::iterator goal_it = goals.begin(); goal_it != goals.end(); ++goal_it) {
        goal_it->PPDDL_grounding();
    }
    (static_cast<TypedAtom*>(this))->PPDDL_grounding();
}

// see TODO below
// bool order_rules_by_name(const RulePtr i, const RulePtr j)
// {
//     return (i->name < j->name);
// }

std::map<std::string, RuleSet> RuleSet::get_rules_per_action() const
{
    using namespace boost::lambda;
    std::map<std::string, RuleSet> result;
    if (empty()) {
        return result;
    }
    RuleSet ordered_rules(*this);
    // TODO check if lambda version works, if not undo
//     std::sort(ordered_rules.begin(), ordered_rules.end(), order_rules_by_name);
    std::sort(ordered_rules.begin(), ordered_rules.end(), 
        boost::lambda::bind(&Rule::name, *_1 ) < boost::lambda::bind(&Rule::name, *_2 )
    );
    
//     std::sort(ordered_rules.begin(), ordered_rules.end(), *_1 < *_2);

    std::string current_action = ordered_rules[0]->name;
    RuleSet current_ruleset;
    for (RuleList::const_iterator rule_it = ordered_rules.begin(); rule_it != ordered_rules.end(); ++rule_it) {
        if ((*rule_it)->name == current_action) {
            current_ruleset.push_back(*rule_it);
        } else {
            result[current_action] = current_ruleset;
            current_action = (*rule_it)->name;
            current_ruleset = RuleSet();
            current_ruleset.push_back(*rule_it);
        }
    }
    result[current_action] = current_ruleset;
    return result;
}
