#include "symbols.h"
#include "symbolic/utils.h"

// OPERATORS
namespace PPDDL {
    bool operator== (const Object &o1, const Object &o2)
    {
        return ((o1.name == o2.name) && (o1.type == o2.type));
    }

    bool operator<(const Object& l, const Object& r)
    {
        if  (l.type < r.type) {
            return true;
        }
        else if (r.type < l.type) {
            return false;
        }
        else {
            return l.name < r.name;
        }
    }
    
    std::ostream& operator<<(std::ostream &out, const PPDDL::Object& o)
    {
        out << o.name;
        out << " - ";
        out << o.type;
        return out;
    }
}

// bool operator== (const Atom &s1, const Atom &s2)
// {
//     return ((s1.name == s2.name) && utils::compare_ordered_vector_strings(s1.param_names, s2.param_names));
// }

bool operator== (const NonTypedAtom &s1, const NonTypedAtom &s2)
{
    return ((s1.name == s2.name) && (s1.params == s2.params));
}

bool operator== (const TypedAtom &s1, const TypedAtom &s2)
{
    return ((s1.name == s2.name) && (s1.params == s2.params));
}

bool operator==(const IndexedAtom& l, const IndexedAtom& r)
{
    return l.index == r.index;
}

// bool operator!= (const Atom &s1, const Atom &s2)
// {
//     return !(s1 == s2);
// }

bool operator!= (const TypedAtom &s1, const TypedAtom &s2)
{
    return !(s1 == s2);
}

bool operator!= (const IndexedAtom &s1, const IndexedAtom &s2)
{
    return s1.index != s2.index;
}

bool TypedAtom::symbolic_compare(const TypedAtom &s) const
{
    bool equal = (s.name == name);
    if (is_symbolic() || s.is_symbolic()) { // if symbolic compare number of params
        equal = (equal && (this->params.size() == s.params.size()));
        if (equal) { // compare types
            for (size_t param_idx = 0; param_idx < params.size(); ++param_idx) {
                equal = (equal && (this->params[param_idx].type == s.params[param_idx].type));
            }
        }
    } else { // if grounded compare params content
        equal = (equal && (this->params == s.params));
    }
    return equal;
}

bool operator<(const NonTypedAtom& l, const NonTypedAtom& r)
{
    if (l.name < r.name) {
        return true;
    } else if (r.name < l.name) {
        return false;
    } else { //equal
//         return l.param_names < r.param_names;
        if (l.params.size() < r.params.size())
            return true;
        else if (r.params.size() < l.params.size())
            return false;
        else { // equal number of parameters
            std::vector< PPDDL::Object >::const_iterator l_it, r_it;
            r_it = r.params.begin();
            for (l_it = l.params.begin(); l_it != l.params.end(); ++l_it) {
                if (l_it->name < r_it->name) {
                    return true;
                } else if (r_it->name < l_it->name) {
                    return false;
                }
                ++r_it;
            }
        }
    }
    return false;
}

bool operator<(const IndexedAtom& l, const IndexedAtom& r)
{
    return l.index < r.index;
}

bool operator<(const TypedAtom& l, const TypedAtom& r)
{
    if (l.name < r.name) {
        return true;
    } else if (r.name < l.name) {
        return false;
    } else { //equal
        // check param types
        return (l.params < r.params);
    }
}

std::ostream& operator<<(std::ostream &out, const Atom &s)
{
    // equivalent to NonTypedAtom
    out << s.name << "(";
    for (size_t i = 0; i < s.params.size(); i++) {
        out << s.params[i].name;
        if ((i + 1) < s.params.size())
            out << " ";
    }
    out << ")";
    return out;
}

std::ostream& operator<<(std::ostream &out, const NonTypedAtom &s)
{
    out << s.name << "(";
    for (size_t i = 0; i < s.params.size(); i++) {
        out << s.params[i].name;
        if ((i + 1) < s.params.size())
            out << " ";
    }
    out << ")";
    return out;
}

std::ostream& operator<<(std::ostream &out, const TypedAtom &s)
{
    out << s.name << "(";
    s.write_typed_parameters(out);
    out << ")";
    return out;
}

std::ostream& operator<<(std::ostream &out, const IndexedAtom &s)
{
    out << s.index << "()";
    return out;
}

std::ostream& operator<<(std::ostream &out, const PPDDLObjectManager &pom)
{
    out << "[";
    BOOST_FOREACH(PPDDL::ObjectTypeMap::value_type object_type_pair, pom.ppddl_objects_types) {
        out << object_type_pair.first << " - " << object_type_pair.second << ", ";
    }
    out << "]";
    return out;
}

void TypedAtom::write_typed_parameters(std::ostream &out) const
{
    for (size_t i = 0; i < params.size(); i++) {
        out << params[i];
        if ((i + 1) < params.size())
            out << " ";
    }
}

void TypedAtom::write_PPDDL(std::ostream &out) const
{
    out << "(" << name << " ";
    write_typed_parameters(out);
    out << ")";
}

void NonTypedAtom::write_PPDDL(std::ostream &out) const
{
    out << "(" << name << " ";
    BOOST_FOREACH(const PPDDL::Object& param, params) {
        out << param.name << " ";
    }
    out << ")";
}

void TypedAtom::write_RDDL(std::ostream &out, const bool future_state) const
{
    out << "(" << name;
    if (future_state) {
        out << "'";
    }
    BOOST_FOREACH(const PPDDL::Object& param, params) {
        out << " " << param.name;
    }
    out << ")";
}


// Atom
void Atom::set_param_names(const std::vector<PPDDL::Object>& new_names) { 
    assert(params.size() == new_names.size()); 
    for(size_t i = 0; i < params.size(); ++i) { 
        params[i].name = new_names[i].name; 
    }  
}

NonTypedAtom::NonTypedAtom(const std::string &name_, const std::vector<std::string> &param_names_)
{
    name = name_;
    BOOST_FOREACH(const std::string& param_name, param_names_) {
        params.push_back(PPDDL::Object(param_name, DEFAULT_OBJECT_TYPE));
    }
}

NonTypedAtom::NonTypedAtom(const std::string &line)
{
    load_atom_name_from_string(line);
    std::vector<std::string> read_param_names = get_atom_param_names_from_string(line);
    BOOST_FOREACH(const std::string& param_name, read_param_names) {
        params.push_back(PPDDL::Object(param_name, DEFAULT_OBJECT_TYPE));
    }
}

bool Atom::is_symbolic() const
{
    BOOST_FOREACH(const PPDDL::Object& param, params) {
        if (symbolic::is_param_symbolic(param.name)) { // found!
            return true;
        }
    }
    return false;
}

bool TypedAtom::has_param(const PPDDL::Object &required_param) const
{
    bool result = false;
    BOOST_FOREACH(const PPDDL::Object& param, params) {
        if (param == required_param) {
            result = true;
        }
    }
    return result;
}

int TypedAtom::get_param_name_idx(const std::string &param) const
{
    int result = -1;
    for (size_t i = 0; i < params.size(); i++) {
        if (params[i].name == param) {
            result = i;
        }
    }
    return result;
}

bool TypedAtom::params_belong_to(const std::vector<PPDDL::Object> &list_of_params) const
{
    bool result = true;
    BOOST_FOREACH(const PPDDL::Object& param, params) {
        bool present = false;
        for (size_t j = 0; j < list_of_params.size(); j++) {
            if (param == list_of_params[j]) {
                present = true;
            }
        }
        if (!present) {
            result = false;
        }
    }
    return result;
}

bool TypedAtom::is_valid() const
{
    return (name != "");
}

void TypedAtom::make_symbolic()
{
    symbolic::make_vector_params_symbolic(params);
}

void symbolic::make_vector_params_symbolic(std::vector<std::string>& vector_params)
{
    for (size_t i = 0; i < vector_params.size(); ++i) {
        vector_params[i] = DEFAULT_PARAM_NAMES[i];
    }
}

void symbolic::make_vector_params_symbolic(std::vector<PPDDL::Object>& vector_params)
{
    for (size_t i = 0; i < vector_params.size(); ++i) {
        vector_params[i].name = DEFAULT_PARAM_NAMES[i];
    }
}

bool NonTypedAtom::is_special() const
{
    bool result = ((*this == MAKE_NONTYPED_ATOM_DEAD) || (*this == MAKE_NONTYPED_ATOM_RETRY) || (*this == MAKE_NONTYPED_ATOM_HELP) || (*this == MAKE_NONTYPED_ATOM_UNKNOWN) || (*this == MAKE_NONTYPED_ATOM_FINISH_ACTION) || (*this == MAKE_NONTYPED_ATOM_TEACHER) || (*this == MAKE_NONTYPED_ATOM_FINISHED) || (*this == MAKE_NONTYPED_ATOM_NOOP_ACTION) || (*this == MAKE_NONTYPED_ATOM_NOACTION) );
    return result;
}

void TypedAtom::transform_nid_params_to_pddl_params()
{
    BOOST_FOREACH(PPDDL::Object& param, params){
        if (param.name == "X") {
            param.name = "?X";
        }
        else if (param.name == "Y") {
            param.name = "?Y";
        }
        else if (param.name == "Z") {
            param.name = "?Z";
        }
        else if (param.name == "V") {
            param.name = "?V";
        }
        else if (param.name == "W") {
            param.name = "?W";
        }
        else if (param.name == "Q") {
            param.name = "?Q";
        }
    }
}

void TypedAtom::change_param_types(const std::vector<std::string>& param_names, const std::vector<std::string>& param_types)
{
    assert(param_names.size() == param_types.size());
    BOOST_FOREACH(PPDDL::Object& param, params) {
        for (size_t i = 0; i < param_names.size(); ++i) {
            if (param_names[i] == param.name) {
                param.type = param_types[i];
            }
        }
    }
}

void TypedAtom::ground(std::vector<std::string> variables)
{
    for (size_t i = 0; i < params.size(); ++i) {
        if (variables.size() > i) {
            params[i].name = variables[i];
        } else {
            std::cout << "\033[1;31m" << "Atoms:: ground - Incorrect number of parameters." << "\033[0m" << std::endl;
            std::cout << "\033[1;31m" << "Atom params are: " << params << "\033[0m" << std::endl;
            std::cout << "\033[1;31m" << "Arguments are: " << variables << "\033[0m" << std::endl;
        }
    }
}

bool TypedAtom::ground(std::vector<PPDDL::Object> variables)
{
    for (size_t i = 0; i < params.size(); ++i) {
        if (variables.size() > i) {
            if (params[i].type == variables[i].type) {
                params[i].name = variables[i].name;
            }
            else {
                LOG(WARNING) << "WARNING probably error!";
                return false;
            }
        } else {
            std::cout << "\033[1;31m" << "Atoms:: ground - Incorrect number of parameters." << "\033[0m" << std::endl;
            std::cout << "\033[1;31m" << "Atom params are: " << params << "\033[0m" << std::endl;
            std::cout << "\033[1;31m" << "Arguments are: " << variables << "\033[0m" << std::endl;
        }
    }
    return true;
}

void TypedAtom::change_param_name(const std::string& orig_var, const std::string& new_var)
{
    BOOST_FOREACH(PPDDL::Object& param, params) {
        if (param.name == orig_var) {
            param.name = new_var;
        }
    }
}

void TypedAtom::change_param_names(const std::vector<std::string>& orig_var_v, const std::vector<std::string>& new_var_v)
{
    BOOST_FOREACH(PPDDL::Object& param, params) {
        for (size_t i = 0; i < orig_var_v.size(); ++i) {
            if (param.name == orig_var_v[i]) {
                param.name = new_var_v[i];
            }
        }
    }
}

void TypedAtom::change_param(const PPDDL::Object& orig_var, const PPDDL::Object& new_var)
{
    BOOST_FOREACH(PPDDL::Object& param, params) {
        if (param == orig_var) {
            param = new_var;
        }
    }
}

void TypedAtom::change_params(const std::vector<PPDDL::Object>& orig_var_v, const std::vector<PPDDL::Object>& new_var_v)
{
    BOOST_FOREACH(PPDDL::Object& param, params) {
        for (size_t i = 0; i < orig_var_v.size(); ++i) {
            if (param == orig_var_v[i]) {
                param = new_var_v[i];
            }
        }
    }
}

void TypedAtom::change_param_names(const std::vector<PPDDL::Object>& orig_var_v, const std::vector<PPDDL::Object>& new_var_v)
{
    BOOST_FOREACH(PPDDL::Object& param, params) {
        for (size_t i = 0; i < orig_var_v.size(); ++i) {
            if (param.name == orig_var_v[i].name) {
                param.name = new_var_v[i].name;
            }
        }
    }
}

std::vector<std::string> Atom::get_param_names() const
{
    std::vector<std::string> param_names;
    BOOST_FOREACH(const PPDDL::Object& param, params) {
        param_names.push_back(param.name);
    }
    return param_names;
}

std::vector<std::string> Atom::get_param_types() const
{
    std::vector<std::string> param_types;
    BOOST_FOREACH(const PPDDL::Object& param, params) {
        param_types.push_back(param.type);
    }
    return param_types;
}

void Atom::load_atom_name_from_string(const std::string &line)
{
    std::stringstream ss(line);
    std::string item;

    if (std::getline(ss, item, '(')) {
        //std::cout << "Atom name: " << item << std::endl;
    }

    name = item;
}

std::vector< std::string > Atom::get_atom_param_names_from_string(const std::string &line)
{
    if (line.find('(') == std::string::npos) { // doesn't have params
        return std::vector< std::string >();
    }
    
    std::stringstream ss(line);
    std::string item;

    // get string between ()
    std::getline(ss, item, '(');
    std::getline(ss, item, ')');
    
    assert((item.find('?') == std::string::npos) && (item.find(" - ") == std::string::npos));
//     if ( (item.find('?') != std::string::npos) && (item.find(" - ") != std::string::npos)) {
        // if has '?', then it should be PPDDL rules
//         return get_param_names_from_ppddl_parameters_string(item);
//     }
//     else {
        return get_param_names_from_nid_parameters_string(item);
//     }
}

std::vector< std::string > Atom::get_param_names_from_nid_parameters_string(const std::string &nid_parameters)
{
    std::stringstream ss;
    std::string item;
    std::vector< std::string> param_names;
    ss.str(nid_parameters);
    
    char separator = utils::choose_variable_separator(nid_parameters);
    while (std::getline(ss, item, separator)) {
        // remove spaces and empty strings
        item.erase(remove_if(item.begin(), item.end(), isspace), item.end());
        if (item.compare("") != 0) {
            param_names.push_back(item);
//             std::cout << "Param names: " << item << std::endl;
        }
    }

    return param_names;
}

void TypedAtom::get_typed_params_from_ppddl_parameters_string(const std::string &ppddl_parameters)
{
    std::stringstream ss;
    std::string item;
    ss.str(ppddl_parameters);
    this->params.clear();
    
    // get elements between ( )
    std::vector<std::string> elements;
    char separator = ' ';
    while (std::getline(ss, item, separator)) {
        // remove spaces and empty strings
        item.erase(remove_if(item.begin(), item.end(), isspace), item.end());
        if (item.compare("") != 0) {
            elements.push_back(item);
            //std::cout << "Param names: " << item << std::endl;
        }
    }
    
    for(size_t i = 0; i + 2 < elements.size(); i=i+3) {
        // elements[i+1] == '-'
        params.push_back(PPDDL::Object(elements[i], elements[i+2]));
    }
}


NonTypedAtom NonTypedAtom::get_atom_from_grounded_name_item(const std::string& grounded_name_atom)
{
    std::stringstream ss(grounded_name_atom);
    std::string item;
    bool has_more_params;

    // notation: add_name[_neg]{_param1}
    // get string between MAKE_GROUNDED_NAME_ATOM_SEPARATOR
    std::getline(ss, item, MAKE_GROUNDED_NAME_ATOM_SEPARATOR);
    NonTypedAtom parsed_atom(item, std::vector<std::string>());
    
    // there are two SEPARATOR after action name
    std::getline(ss, item, MAKE_GROUNDED_NAME_ATOM_SEPARATOR);

    has_more_params = std::getline(ss, item, MAKE_GROUNDED_NAME_ATOM_SEPARATOR);
    while(has_more_params) {
        parsed_atom.params.push_back(PPDDL::Object(item,DEFAULT_OBJECT_TYPE));
        has_more_params = std::getline(ss, item, MAKE_GROUNDED_NAME_ATOM_SEPARATOR);
    }

    return parsed_atom;
}



/**
 * TYPED ATOM
 */

TypedAtom::TypedAtom(const std::string &line)
{
    load_atom_name_from_string(line);
    load_typed_param_names_from_string(line);
}


TypedAtom::TypedAtom(const std::string &name_, const std::vector<std::string> &param_names_, const std::vector<std::string> &param_types_)
{
    name = name_;
    assert(param_names_.size() == param_types_.size());
    for (size_t i = 0; i < param_names_.size(); ++i) {
        params.push_back(PPDDL::Object(param_names_[i], param_types_[i]));
    }
}

TypedAtom::TypedAtom(const std::string &name_, const std::vector<PPDDL::Object> &ppddl_params)
{
    name = name_;
    params = ppddl_params;
}

void TypedAtom::load_typed_param_names_from_string(const std::string &line)
{
    if (line.find('(') == std::string::npos) { // doesn't have params
        return;
    }
    
    std::stringstream ss(line);
    std::string item;

    // get string between ()
    std::getline(ss, item, '(');
    std::getline(ss, item, ')');
    ss.str(item);
    
    if ( (item.find('?') == std::string::npos) && (item.find(" - ") == std::string::npos) ) {
        std::vector<std::string> param_names = get_param_names_from_nid_parameters_string(item);
//         LOG(ERROR) << "Unsupported non typed params!";
//         assert(param_names.empty() && "Unsupported non typed params!");
        BOOST_FOREACH(const std::string& param_name, param_names) {
            params.push_back(PPDDL::Object(param_name, DEFAULT_OBJECT_TYPE));
        }
//         for (size_t i = 0; i < param_names.size(); ++i) {
//             param_types.push_back(DEFAULT_OBJECT_TYPE);
//         }
//         // HACK log once
        static bool hacky_bool = true;
        if (!param_names.empty()) {
            LOG_IF(hacky_bool,WARNING) << "NID compatibility, DEPRECATED. Reading: " << line;
        }
        hacky_bool = false;
    }
    else {
        get_typed_params_from_ppddl_parameters_string(item);
    }
}


bool TypedAtom::is_special() const
{
    bool result = ((*this == MAKE_TYPED_ATOM_DEAD) || 
                   (*this == MAKE_TYPED_ATOM_RETRY) || 
                   (*this == MAKE_TYPED_ATOM_HELP) || 
                   (*this == MAKE_TYPED_ATOM_UNKNOWN) || 
                   (*this == MAKE_TYPED_ATOM_FINISH_ACTION) ||
                   (*this == MAKE_TYPED_ATOM_NOACTION) || 
                   (*this == MAKE_TYPED_ATOM_TEACHER));
    return result;
}


void TypedAtom::PPDDL_grounding()
{
    if (params.size() > 0) // first parameter has two _
        name = name + "_";
    for (std::vector<PPDDL::Object>::iterator param_it = params.begin(); param_it != params.end(); ++param_it) {
        name = name + "_" + param_it->name;
    }
    params.clear();
}

std::string PPDDLObjectManager::get_object_type(const std::string& object) const
{
    PPDDL::ObjectTypeMap::const_iterator iter = ppddl_objects_types.find(object);
    if (iter != ppddl_objects_types.end()) {
        return iter->second;
    }
    else {
        return DEFAULT_OBJECT_TYPE;
    }
}

std::vector<PPDDL::Name> PPDDLObjectManager::get_type_objects(const std::string& type) const
{
    if (type == DEFAULT_OBJECT_TYPE) {
        return get_ppddl_object_names();
    }
    
    PPDDL::TypeObjectsMap::const_iterator iter = ppddl_objects_per_type.find(type);
    if (iter != ppddl_objects_per_type.end()) {
        return iter->second;
    }
    else {
        return std::vector<std::string>();
    }
}

TypedAtom PPDDLObjectManager::get_typed_atom(const Atom& atom) const
{
    std::vector<PPDDL::Object> typed_params;
    BOOST_FOREACH(const std::string& param_name, atom.get_param_names()) {
        typed_params.push_back(PPDDL::Object(param_name, get_object_type(param_name)));
    }
    return TypedAtom(atom.name, typed_params);
}

void PPDDLObjectManager::add_objects(const PPDDLObjectManager& added_object_manager)
{
    BOOST_FOREACH(const PPDDL::ObjectTypeMap::value_type& object_type_pair, added_object_manager.ppddl_objects_types) {
        if (ppddl_objects_types.count(object_type_pair.first) == 0) { // object doesn't exist already
            ppddl_objects_types[object_type_pair.first] = object_type_pair.second;
            ppddl_objects_per_type[object_type_pair.second].push_back(object_type_pair.first);
        }
    }
}

PPDDLObjectManager::CombinationsPPDDLObjectsPtr PPDDLObjectManager::get_combinations_of_objecs_for_given_params(const std::vector<PPDDL::Type>& list_param_types) const
{
    CacheCombinationsObjects::iterator cache_it = cache_combinations_objects.find(list_param_types);
    if (cache_it != cache_combinations_objects.end()) {
        return cache_it->second;
    }
    else {
        std::vector< std::vector< std::string > > combinations_of_params;
        combinations_of_params.push_back(std::vector<std::string>()); // add empty one to begin with
        BOOST_FOREACH(const PPDDL::Type& ppddl_type, list_param_types) { // for each parameter
            std::vector< std::vector< std::string > > new_combinations_of_params;
            std::vector<std::string> pddl_objects = get_type_objects(ppddl_type);
            BOOST_FOREACH(const std::string& ppddl_object, pddl_objects) { // for each object of the parameter type
                BOOST_FOREACH(const std::vector< std::string >& a_combination_of_params, combinations_of_params) { 
                    // create combinations: current_combinations x number_objects
                    // Don't repeat the same object in the arguments
                    if (std::find(a_combination_of_params.begin(), a_combination_of_params.end(), ppddl_object) == a_combination_of_params.end()) {
    //                     a_combination_of_params.push_back(ppddl_object);
                        new_combinations_of_params.push_back(a_combination_of_params);
                        new_combinations_of_params.back().push_back(ppddl_object);
                    }
                }
            }
            combinations_of_params.swap(new_combinations_of_params);
        }
        
        CombinationsPPDDLObjectsPtr combinations_of_params_ptr =boost::make_shared<CombinationsPPDDLObjects>(combinations_of_params);
        cache_combinations_objects[list_param_types] = combinations_of_params_ptr;
        return combinations_of_params_ptr;
    }
}

