/*
 * Class representing a logic atom with parameters.
 * Copyright (C) 2014  David Martínez (dmartinez@iri.upc.edu)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef IRI_RULE_LEARNER_ATOMS_H
#define IRI_RULE_LEARNER_ATOMS_H 1

#include <algorithm>
#include <vector>
#include <set>
#include <map>
#include <fstream>
#include <sstream>
#include <iostream>
#include <iomanip>

#include <boost/foreach.hpp>
#include <boost/optional.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/lexical_cast.hpp>

// Typed objects
#define DEFAULT_OBJECT_TYPE "default"

// special
#define MAKE_GROUNDED_NAME_ATOM_SEPARATOR  '_'

// action names
#define ACTION_NAME_NOACTION "noaction"
#define ACTION_NAME_NOOP "noop"
#define ACTION_NAME_FINISH "finish_action"

// actions
#define MAKE_NONTYPED_ATOM_DEAD          NonTypedAtom("dead()")
#define MAKE_NONTYPED_ATOM_RETRY         NonTypedAtom("retry()")
#define MAKE_NONTYPED_ATOM_HELP          NonTypedAtom("help()")
#define MAKE_NONTYPED_ATOM_UNKNOWN       NonTypedAtom("unknownManipulation()")
#define MAKE_NONTYPED_ATOM_TEACHER       NonTypedAtom("teacher()")
#define MAKE_NONTYPED_ATOM_FINISH_ACTION NonTypedAtom(ACTION_NAME_FINISH)
#define MAKE_NONTYPED_ATOM_NOOP_ACTION   NonTypedAtom(ACTION_NAME_NOOP)
#define MAKE_NONTYPED_ATOM_NOACTION      NonTypedAtom(ACTION_NAME_NOACTION)

#define MAKE_TYPED_ATOM_DEAD          TypedAtom("dead", std::vector<std::string>(), std::vector<std::string>())
#define MAKE_TYPED_ATOM_RETRY         TypedAtom("retry", std::vector<std::string>(), std::vector<std::string>())
#define MAKE_TYPED_ATOM_HELP          TypedAtom("help", std::vector<std::string>(), std::vector<std::string>())
#define MAKE_TYPED_ATOM_UNKNOWN       TypedAtom("unknownManipulation", std::vector<std::string>(), std::vector<std::string>())
#define MAKE_TYPED_ATOM_TEACHER       TypedAtom("teacher", std::vector<std::string>(), std::vector<std::string>())
#define MAKE_TYPED_ATOM_FINISH_ACTION TypedAtom(ACTION_NAME_FINISH, std::vector<std::string>(), std::vector<std::string>())
#define MAKE_TYPED_ATOM_NOOP_ACTION   TypedAtom(ACTION_NAME_NOOP, std::vector<std::string>(), std::vector<std::string>())
#define MAKE_TYPED_ATOM_NOACTION      TypedAtom(ACTION_NAME_NOACTION, std::vector<std::string>(), std::vector<std::string>())
#define MAKE_TYPED_ATOM_DEFAULT       TypedAtom("default", std::vector<std::string>(), std::vector<std::string>())

// predicates
#define MAKE_NONTYPED_ATOM_FINISHED      NonTypedAtom("finished", std::vector<std::string>())
#define MAKE_TYPED_ATOM_FINISHED         TypedAtom("finished", std::vector<std::string>(), std::vector<std::string>())
#define MAKE_TYPED_ATOM_STARTED          TypedAtom("started", std::vector<std::string>(), std::vector<std::string>())
#define MAKE_TYPED_ATOM_EXCUSE_GENERATED TypedAtom("excuseGenerated", std::vector<std::string>(), std::vector<std::string>())

// action_ptr
#define MAKE_ATOM_DEAD_PTR    boost::make_shared<Atom>(MAKE_ATOM_DEAD)
#define MAKE_ATOM_RETRY_PTR   boost::make_shared<Atom>(MAKE_ATOM_RETRY)
#define MAKE_ATOM_HELP_PTR    boost::make_shared<Atom>(MAKE_ATOM_HELP)
#define MAKE_ATOM_UNKNOWN_PTR boost::make_shared<Atom>(MAKE_ATOM_UNKNOWN)
#define MAKE_ATOM_TEACHER_PTR boost::make_shared<Atom>(MAKE_ATOM_TEACHER)

static const std::string DEFAULT_PARAM_NAMES[] = {"?X", "?Y", "?Z", "?V", "?W", "?Q"}; //!< Default param names that are considered symbolic.
static const uint DEFAULT_PARAM_NAMES_NUMBER = 6;
static const std::string DEFAULT_ACTION_NAME = "default";
static std::vector<std::string> DEFAULT_PARAM_NAMES_VECTOR(DEFAULT_PARAM_NAMES, DEFAULT_PARAM_NAMES + DEFAULT_PARAM_NAMES_NUMBER);


namespace PPDDL {
    typedef std::string Type;
    typedef std::string Name;
    
    struct Object {
        Name name;
        Type type;
        
        Object(Name object_name, Type object_type):
            name(object_name),
            type(object_type)
        { }
        
        /*! \brief Operator == */
        friend bool operator== (const Object &s1, const Object &s2);
        /*! \brief Operator <= */
        friend bool operator<(const Object& l, const Object& r);
        /*! \brief Operator << */
        friend std::ostream& operator<<(std::ostream &out, const Object& o);
    };
    
    typedef std::vector< Type > TypeVector;
    typedef std::map< Type, std::vector< Name > > TypeObjectsMap;
    typedef std::map< Name, Type > ObjectTypeMap;
}


namespace symbolic {
    inline bool is_param_symbolic(const std::string& param) { return (param[0] =='?'); };
    void make_vector_params_symbolic(std::vector<std::string>& vector_params);
    void make_vector_params_symbolic(std::vector<PPDDL::Object>& vector_params);
}




/*!
\class Atom
\brief Class representing a atom. Atoms consist of a name and a vector of parameters.
*/
class Atom
{
public:
    std::string name; //!< Name
    std::vector< PPDDL::Object > params; //!< Object-Type pairs of the parameters

    /*!
    \brief Empty constructor
    */
    Atom() { }
    
    /*!
    \brief Checks if atom is symbolic, and updates its is_symbolic member

    If one of the default param names (X, Y, Z, ...) is found,
    is_symbolic is set to true.
    */
    bool is_symbolic() const;
    
    /*!
    \brief Set the param names of the given vector
    
    Used to enhance performance of some routines
    */
    void set_param_names(const std::vector<PPDDL::Object>& new_names);
    
    virtual bool is_special() const =0;
    
    /*! \brief Get vector of param names */
    std::vector<std::string> get_param_names() const;
    
    /*! \brief Get vector of param types */
    std::vector<std::string> get_param_types() const;
    
    friend std::ostream& operator<<(std::ostream &out, const Atom &s);

protected:
    
    void load_atom_name_from_string(const std::string &line);
    static std::vector< std::string > get_atom_param_names_from_string(const std::string &line);
    static std::vector< std::string > get_param_names_from_nid_parameters_string(const std::string &nid_parameters);
    
};


/*!
\class IndexedAtom
\brief Class representing an index that corresponds to a atom.
*/
class IndexedAtom
{   
public:
    uint index; //!< Name

    /*! \brief Empty constructor */
    explicit IndexedAtom(uint atom_index): index(atom_index) {}

    /*! \brief Copy constructor */
    IndexedAtom(const IndexedAtom &indexedatom): index(indexedatom.index) {};

    /*! \brief Operator == */
    friend bool operator== (const IndexedAtom &s1, const IndexedAtom &s2);
    /*! \brief Operator != */
    friend bool operator!= (const IndexedAtom &s1, const IndexedAtom &s2);
    /*! \brief Operator <= */
    friend bool operator<(const IndexedAtom& l, const IndexedAtom& r);
    /*! \brief Operator \<\< */
    friend std::ostream& operator<<(std::ostream &out, const IndexedAtom &s);
};


/*!
\class NonTypedAtom
\brief Class representing a atom. Atoms consist of a name and a vector of parameters.
*/
class NonTypedAtom : public Atom
{
public:
    /*!
    \brief Empty constructor
    */
    NonTypedAtom() { }

    /*!
    \brief constructor

    Parses a string that is assumed to have a atom.
    Examples of the syntax supported:
     "clear(X Y)", "clear(X,Y)", "clear()".

    \param line string having a valid representation of a atom
    */
    explicit NonTypedAtom(const std::string &line);
    
    /*!
    \brief Constructor

    Constructs the atom.

    \param name_ name
    \param param_names_ vector of param names
    */
    NonTypedAtom(const std::string &name_, const std::vector<std::string> &param_names_);

    /*!
    \brief Check if TypedAtom is a special atom

    Checks if atom is a special atom: dead(), retry(), help(), unknown(), finish(), teacher().

    \return true if it is special
    */
    virtual bool is_special() const;
    
    void write_PPDDL(std::ostream &out) const;

    /*! \brief Operator == */
    friend bool operator== (const NonTypedAtom &s1, const NonTypedAtom &s2);
//     /*! \brief Operator != */
//     friend bool operator!= (const NonTypedAtom &s1, const NonTypedAtom &s2);
    /*! \brief Operator <= */
    friend bool operator<(const NonTypedAtom& l, const NonTypedAtom& r);
    /*! \brief Operator \<\< */
    friend std::ostream& operator<<(std::ostream &out, const NonTypedAtom &s);
    
    static NonTypedAtom get_atom_from_grounded_name_item(const std::string& grounded_name_atom);
};


class TypedAtom : public Atom
{
private:
    void get_typed_params_from_ppddl_parameters_string(const std::string &ppddl_parameters);
    
public:
    typedef boost::optional<TypedAtom> Opt;
    typedef boost::shared_ptr<TypedAtom> Ptr;
    typedef boost::shared_ptr<const TypedAtom> ConstPtr;

    /*!
    \brief Empty constructor
    */
    TypedAtom() : Atom() { }

    /*!
    \brief constructor

    Parses a string that is assumed to have a Atom.
    Examples of the syntax supported:
     "clear(?X - location ?Y - thing)", "clear()".

    \param line string having a valid representation of a atom
    */
    explicit TypedAtom(const std::string &line);

    /*!
    \brief Constructor

    Constructs the Atom.

    \param name_ name
    \param param_names_ vector of param names
    \param param_types_ vector of param types
    */
    TypedAtom(const std::string &name_, const std::vector<std::string> &param_names_, const std::vector<std::string> &param_types_);
    
    /*!
    \brief Constructor

    Constructs the Atom.

    \param name_ name
    \param ppddl_params vector of params
    */
    TypedAtom(const std::string &name_, const std::vector<PPDDL::Object> &ppddl_params);
    
    /*!
    \brief Compares if not grounded versions of the atoms are equal

    Compares the name and the number of parameters, but not the parameters content

    \return true if not grounded atoms are equal
    */
    bool symbolic_compare(const TypedAtom &s) const;

    /*!
    \brief Check if TypedAtom is a special atom

    Checks if atom is a special atom: dead(), retry(), help(), unknown(), finish(), teacher().

    \return true if it is special
    */
    virtual bool is_special() const;
    
    /*!
    \brief Creat an atom with a longer name that encodes params
    
    Parameters are removed, and instead are encoded within the name
     */
    void PPDDL_grounding();
    
    /*!
    \brief Check if atom is initialized

    Check if atom is initialized, that is that its name is not empty.

    \return true if initialized, false if empty
    */
    bool is_valid() const;
    
    /*!
    \brief Check if a parameter with the given name exists

    \return true if there is a parameter with the given name
    */
    bool has_param(const PPDDL::Object &required_param) const;

    /*!
    \brief Check if parameter with all the given name exist

    \return true if all of them exist
    */
    bool params_belong_to(const std::vector<PPDDL::Object> &list_of_params) const;

    /*!
    \brief Gets the index of a param

    \param param name of the param to look for its index
    \return true if it is special
    */
    int get_param_name_idx(const std::string &param) const;

    /*!
    \brief Makes the atom symbolic

    Changes the params to be the default symbolic params: X, Y, Z, V, W, Q.
    */
    void make_symbolic();
    
    /*!
     \brief Change parameters
     */
    void change_params(const std::vector<PPDDL::Object>& orig_var_v, const std::vector<PPDDL::Object>& new_var_v);
    
    /*!
     \brief Change parameter
     */
    void change_param(const PPDDL::Object& orig_var, const PPDDL::Object& new_var);
    
    /*!
     \brief Change parameter name
     */
    void change_param_name(const std::string& orig_var, const std::string& new_var);
    
    /*!
     \brief Change parameter names
     */
    void change_param_names(const std::vector<std::string>& orig_var_v, const std::vector<std::string>& new_var_v);
    
    /*!
     \brief Change parameter names
     */
    void change_param_names(const std::vector<PPDDL::Object>& orig_var_v, const std::vector<PPDDL::Object>& new_var_v);
    
    /*!
     * \brief Set param type
     */
    void change_param_types(const std::vector<std::string>& param_names, const std::vector<std::string>& param_types);

    /*!
    \brief Grounds the atom with the given parameters
    */
    void ground(std::vector<std::string> variables);
    
    /*!
    \brief Grounds the atom with the given parameters
    */
    bool ground(std::vector<PPDDL::Object> variables);
    
    /*!
    \brief Substitutes X for ?X, Y for ?Y, etc...
    */
    void transform_nid_params_to_pddl_params();
    
    /*!
    \brief Writes PPDDL typed atom
    */
    void write_PPDDL(std::ostream &out) const;
    
    /*! \brief Writes RDDL typed atom. When defining transitions, future_state means the value for the next time step */
    void write_RDDL(std::ostream &out, const bool future_state = false) const;

    /*! \brief Write typed params */
    void write_typed_parameters(std::ostream &out) const;
    
    /*! \brief Operator == */
    friend bool operator== (const TypedAtom &s1, const TypedAtom &s2);
    /*! \brief Operator != */
    friend bool operator!= (const TypedAtom &s1, const TypedAtom &s2);
    /*! \brief Operator <= */
    friend bool operator<(const TypedAtom& l, const TypedAtom& r);
    /*! \brief Operator \<\< */
    friend std::ostream& operator<<(std::ostream &out, const TypedAtom &s);

private:
    void load_typed_param_names_from_string(const std::string &line);
};


/*!
\class PPDDLObjectManager
\brief Class to manage objects and types.
*/
class PPDDLObjectManager
{
public:
    PPDDLObjectManager() {};
    
    PPDDLObjectManager(const PPDDLObjectManager& other):
        ppddl_objects_per_type(other.ppddl_objects_per_type),
        ppddl_objects_types(other.ppddl_objects_types)
    {};
    
    typedef std::vector<PPDDL::Type> ListPPDDLTypes;
    typedef std::vector< std::vector<PPDDL::Name> > CombinationsPPDDLObjects;
    typedef boost::shared_ptr<CombinationsPPDDLObjects> CombinationsPPDDLObjectsPtr;
    typedef std::map<ListPPDDLTypes, CombinationsPPDDLObjectsPtr> CacheCombinationsObjects;
    
    /**
     * \brief Gets typed atom by searching is corresponding param types
     * 
     * TODO C++11 const method -> use map::at() which throws exceptions
     */
    TypedAtom get_typed_atom(const Atom& atom) const;
    
    /**
     * \brief Gets the type of the object
     * 
     * TODO C++11 const method -> use map::at() which throws exceptions
     */
    std::string get_object_type(const std::string& object) const;
    
    /**
     * \brief Gets the objects with type type
     * 
     * TODO C++11 const method -> use map::at() which throws exceptions
     */
    std::vector<std::string> get_type_objects(const std::string& type) const;
    
    /**
     * \brief Gets vectors of ppddl objects associated to each type
     */
    const PPDDL::TypeObjectsMap& get_ppddl_objects_per_type() const { return ppddl_objects_per_type; };
    
    /**
     * \brief Gets the type associated to each object
     */
    const PPDDL::ObjectTypeMap& get_ppddl_objects_types() const { return ppddl_objects_types; };
    
    /**
     * \brief Add object-type pairs from the added_object_manager
     */
    void add_objects(const PPDDLObjectManager& added_object_manager);
    
    /**
     * \brief Add object-type pairs from a std container
     */
    template <typename Iter>
    void add_objects(Iter added_objects_it, Iter added_objects_end) {
        for (; added_objects_it != added_objects_end; ++added_objects_it) {
            add_object(*added_objects_it);
        }
    }
    
    /**
     * \brief Add a object-type pair
     */
    void add_object(const PPDDL::Object& ppddl_object) {
        if (ppddl_objects_types.count(ppddl_object.name) == 0) { // object doesn't exist
            ppddl_objects_types[ppddl_object.name] = ppddl_object.type;
            ppddl_objects_per_type[ppddl_object.type].push_back(ppddl_object.name);
        }
    }
    
    /**
     * \brief Get list of objects
     */
    std::set<PPDDL::Object> get_ppddl_objects() const { 
        std::set<PPDDL::Object> res; 
        BOOST_FOREACH(PPDDL::ObjectTypeMap::value_type object_type, ppddl_objects_types) {
            res.insert(PPDDL::Object(object_type.first, object_type.second));
        } 
        return res; 
    };
    
    /**
     * \brief Get list of object names
     */
    std::vector<PPDDL::Name> get_ppddl_object_names() const { 
        std::vector<std::string> res; 
        BOOST_FOREACH(PPDDL::ObjectTypeMap::value_type object_type, ppddl_objects_types) {
            res.push_back(object_type.first);
        } 
        return res; 
    };
    
    CombinationsPPDDLObjectsPtr get_combinations_of_objecs_for_given_params(const std::vector<PPDDL::Type>& list_param_types) const;
    
    /*! \brief Operator \<\< */
    friend std::ostream& operator<<(std::ostream &out, const PPDDLObjectManager &pom);
    
private:
    PPDDL::TypeObjectsMap ppddl_objects_per_type;
    PPDDL::ObjectTypeMap ppddl_objects_types;
    mutable CacheCombinationsObjects cache_combinations_objects;
};


#endif
