#include "transitions.h"

#define RDDL

namespace Transitions {
Coverage& Coverage::operator+=(const Coverage& other)
{
    /* addition of other to *this takes place here */
    assert(size() == other.size());
    for (size_t i = 0; i < other.size(); ++i) {
        coverage_vector[i].insert(coverage_vector[i].end(), other.coverage_vector[i].begin(), other.coverage_vector[i].end());
    }
#ifdef DEBUG_COVERAGE_ERRORS
    coverage_errors_vector.insert(coverage_errors_vector.end(), other.coverage_errors_vector.begin(), other.coverage_errors_vector.end());
#else
    add_error(other.coverage_error);
#endif
    return *this;
}
    
Coverage operator+(const Coverage &c1, const Coverage &c2)
{
    assert(c1.size() == c2.size());
    if (c2.is_empty) {
        return c1;
    }
    else if (c1.is_empty) {
        return c2;
    }
    Coverage result(c1);
    for (size_t i = 0; i < c2.size(); ++i) {
        result.coverage_vector[i].insert(result.coverage_vector[i].end(), c2.coverage_vector[i].begin(), c2.coverage_vector[i].end());
    }
#ifdef DEBUG_COVERAGE_ERRORS
    result.coverage_errors_vector.insert(result.coverage_errors_vector.end(), c2.coverage_errors_vector.begin(), c2.coverage_errors_vector.end());
#else
    result.add_error(c2.coverage_error);
#endif
    return result;
}

std::ostream& operator<<(std::ostream &out, const Coverage &c)
{
#ifdef DEBUG_COVERAGE_ERRORS
    out << "Cov: " << c.coverage_vector << ", Err: " << c.coverage_errors_vector;
#else
    out << "Cov: " << c.coverage_vector << ", Err: " << c.coverage_error;
#endif
    return out;
}

float Coverage::get_change_coverage(const index_type change_idx, const float optimistic_value) const
{
    if (coverage_vector[change_idx].size() > 1) { // if two rule groundings cover a transition, there is a conflict, coverage is 0 as a result
        return 0.0;
    }
    else if (coverage_vector[change_idx].empty()) {
        return optimistic_value; // will be 0 in normal case, (0,1] in a more optimistic case
    }
    else {
        assert(coverage_vector[change_idx].size() == 1);
        return coverage_vector[change_idx][0];
    }
}

float Coverage::get_coverage(const bool complete, const float optimistic_value) const
{
    if (coverage_vector.empty()) {
        return 1.0 - get_error_probability();
    }
    
    float result;
    if (complete) {
        result = 1.0;
    }
    else {
        result = 0.0;
    }
    for (size_t change_idx = 0; change_idx < coverage_vector.size(); ++change_idx) {
        if (complete) {
            result *= get_change_coverage(change_idx, optimistic_value);
        }
        else {
            result += get_change_coverage(change_idx, optimistic_value);
        }
    }
    if (complete) {
        result *= (1 - get_error_probability());
    }
    else {
        result /= coverage_vector.size();
    }
    return result;
}

float Coverage::get_error_probability() const {
#ifdef DEBUG_COVERAGE_ERRORS
    float result = 0.0;
    BOOST_FOREACH(const float coverage_error ,coverage_errors_vector) { 
        result += (1.0 - result) * coverage_error;
    }
    return result;
#else
    return coverage_error;
#endif
}


float Coverage::get_likelihood(const bool complete, const float optimistic_value) const
{
    return get_coverage(complete, optimistic_value); // if complete == true, already removing error probability from result
}

}

std::ostream& operator<<(std::ostream &out, const TransitionGroup &t)
{
    BOOST_FOREACH(const std::set<Transition>::iterator& transition_it, t.transition_historial) {
        out << *transition_it;
    }
    return out;
}

std::ostream& operator<<(std::ostream &out, const Transition &t)
{
    t.prev_state.write_PPDDL(out, t.get_action().params);
    out << std::endl;
    t.action.write_PPDDL(out);
    out << std::endl;
    t.next_state.write_PPDDL(out, t.get_action().params);
    out << std::endl;
    out << t.reward << std::endl;
    return out;
}

bool operator<(const Transition &l, const Transition &r)
{
    if (l.action < r.action)
        return true;
    else if (r.action < l.action)
        return false;
    else if (l.prev_state < r.prev_state)
        return true;
    else if (r.prev_state < l.prev_state)
        return false;
    else if (l.next_state < r.next_state)
        return true;
    else if (r.next_state < l.next_state)
        return false;
    else if (l.reward < r.reward)
        return true;
    else if (l.reward < r.reward)
        return false;
    else
        return false;
}

TransitionGroup read_transitions_from_file(const std::string& file_path, const PPDDLObjectManager& object_manager, const PredicateGroup<TypedAtom>& constant_preds, const bool& variable_maintains_to_next_state)
{
    TransitionGroup transitions(object_manager);
    std::ifstream transitions_file(file_path.c_str());
    if (transitions_file.is_open()) {
        std::string line_prev, line_action, line_next, line_reward;

        while (transitions_file.good()) {
            getline(transitions_file, line_prev);
            getline(transitions_file, line_action);
            getline(transitions_file, line_next);
            getline(transitions_file, line_reward);

            // avoid adding twice last transition
            if (transitions_file.good()) {
                try {
                    float transition_reward = boost::lexical_cast<float>(line_reward);
                    PredicateGroup<NonTypedAtom> prev_s(line_prev);
                    PredicateGroup<TypedAtom> prev_s_typed = Predicates::transform_to_typed(prev_s, object_manager);
                    prev_s_typed.set_constants(constant_preds);
                    
                    NonTypedAtom action(line_action);
                    PredicateGroup<NonTypedAtom> next_s(line_next);
                    PredicateGroup<TypedAtom> next_s_typed = Predicates::transform_to_typed(next_s, object_manager);
                    next_s_typed.set_constants(constant_preds);
                    
                    Transition transition(prev_s_typed,
                                          object_manager.get_typed_atom(action),
                                          next_s_typed,
                                          transition_reward,
                                          variable_maintains_to_next_state
                                         );
                    transitions.add_transition(transition);
                } catch (boost::bad_lexical_cast &) {
                    LOG(WARNING) << "Error reading transition " << transitions.size_historial() + 1;
                }
            }
        }
    } else {
        LOG(ERROR) << "Couldn't read transitions from " << file_path;
    }
    transitions_file.close();
    return transitions;
}

/**
 * Transition
 */

void Transition::remove_unused_predicates()
{
    prev_state.remove_predicates_with_params_not_in(action.params);
    next_state.remove_predicates_with_params_not_in(action.params);
}

void Transition::remove_negated_predicates()
{
    prev_state.remove_negated_predicates();
    next_state.remove_negated_predicates();
}

bool Transition::nothing_changed() const
{
    return prev_state == next_state;
}

void Transition::update_changes(const bool& variable_maintains_to_next_state)
{
    if (variable_maintains_to_next_state) {
        changes = prev_state.get_differences(next_state);
    }
    else {
        changes.clear();
        BOOST_FOREACH(const Predicate<TypedAtom>& next_state_pred, next_state) {
            if ( next_state_pred.positive && !next_state_pred.is_constant() ) {
                changes.insert(next_state_pred);
            }
        }
    }
}


/**
 * TransitionGroup
 */

// reference TransitionGroup::operator[] (size_t n)
// {
//     return *transition_historial[n];
// }
// 
// const_reference TransitionGroup::operator[] (size_t n) const
// {
//     return *transition_historial[n];
// }

void TransitionGroup::write_transitions_to_file(const std::string& file_path) const
{
    std::ofstream transitions_file(file_path.c_str());
    if (transitions_file.is_open()) {
        transitions_file << *this;
    } else
        std::cerr << "Couldn't write transitions to " << file_path << std::endl;
    transitions_file.close();
}

void TransitionGroup::write_pasula_transitions_to_file(const std::string& file_path) const
{
    std::ofstream transitions_file(file_path.c_str());
    if (transitions_file.is_open()) {
        BOOST_FOREACH(const TransitionSet::const_iterator& transition_it, transition_historial) {
            transitions_file << Predicates::transform_to_nontyped(transition_it->get_prev_state()) << std::endl;
            transitions_file << NonTypedAtom(transition_it->get_action().name, transition_it->get_action().get_param_names()) << std::endl;
            transitions_file << Predicates::transform_to_nontyped(transition_it->get_post_state()) << std::endl;
            transitions_file << transition_it->get_reward() << std::endl;
        }
    } else
        std::cerr << "Couldn't write transitions to " << file_path << std::endl;
    transitions_file.close();
}

void TransitionGroup::write_lfit_transitions_to_file(const std::string& file_path, const std::string& domain_types, const std::vector<NonTypedAtom>& list_of_actions) const
{
    std::ofstream transitions_file(file_path.c_str());
    if (transitions_file.is_open()) {
        transitions_file << domain_types << std::endl;
        
        for (TransitionSet::const_iterator transition_it = begin_set() ; transition_it != end_set(); ++transition_it) {
            transitions_file << std::endl;
            transition_it->get_prev_state().write_LFIT(transitions_file);
            transitions_file << " action=" << transition_it->get_action().name;
            transitions_file << " : ";
            transition_it->get_post_state().write_LFIT(transitions_file);
            transitions_file << " action=" << transition_it->get_action().name;
            if (transition_it->get_number_repetitions() > 1) {
                transitions_file << std::endl;
                transitions_file << "# " << transition_it->get_number_repetitions();
            }
        }
    } else
        std::cerr << "Couldn't write transitions to " << file_path << std::endl;
    transitions_file.close();
}

TransitionGroup TransitionGroup::get_action_list_transitions(std::vector<TypedAtom> actions) const
{
    TransitionGroup important_transitions(ppddl_object_manager);

    // remove duplicated actions to don't duplicate transitions
    std::sort(actions.begin(), actions.end());
    std::vector< TypedAtom >::iterator it = std::unique(actions.begin(), actions.end());
    actions.resize(std::distance(actions.begin(), it));

    for (std::vector<TypedAtom>::iterator action = actions.begin(); action != actions.end(); ++action) {
        TransitionGroup aux_transitions = get_action_transitions(*action);
        important_transitions.add_transitions(aux_transitions);
    }

    return important_transitions;
}

TransitionGroup TransitionGroup::get_action_transitions(const TypedAtom& action) const
{
    TransitionGroup important_transitions(ppddl_object_manager);
    for(TransitionHistorial::const_iterator transition_it_it = begin_historial(); transition_it_it != end_historial(); ++transition_it_it) {
        if ((*transition_it_it)->get_action().symbolic_compare(action) || (action == MAKE_TYPED_ATOM_NOACTION)) {
            important_transitions.add_transition(**transition_it_it);
        }
    }

    return important_transitions;
}

TransitionGroup Transition::get_symbolic_transitions(const PPDDLObjectManager& object_manager, const int max_action_variables) const
{
    TransitionGroup result(object_manager);
    
    std::vector<PPDDL::Object> action_params = action.params;
    // possible deictic vars
    std::vector<PPDDL::Object> other_vars = prev_state.get_params();
    BOOST_FOREACH(const PPDDL::Object& action_param, action_params) {
        std::vector<PPDDL::Object>::iterator it = std::find (other_vars.begin(), other_vars.end(), action_param);
        if (it != other_vars.end()) {
            other_vars.erase(it);
        }
    }
    
    // other noop transitions
    int n_deictic_vars = std::max(0, max_action_variables - (int)action_params.size());
    std::vector< std::vector<PPDDL::Object> > other_vars_combinations = utils::get_all_permutations<PPDDL::Object>(other_vars, n_deictic_vars);
    
    BOOST_FOREACH(const std::vector<PPDDL::Object>& deictic_vars_combination, other_vars_combinations) {
        std::vector<PPDDL::Object> vars_combination(action_params);
        vars_combination.insert(vars_combination.end(), deictic_vars_combination.begin(), deictic_vars_combination.end());
        Transition new_transition(*this);
        new_transition.prev_state.remove_predicates_with_params_not_in(vars_combination);
        new_transition.next_state.remove_predicates_with_params_not_in(vars_combination);
        for (size_t i = 0; i < vars_combination.size(); ++i) {
            new_transition.prev_state.change_param_name(vars_combination[i].name, DEFAULT_PARAM_NAMES[i]);
            new_transition.next_state.change_param_name(vars_combination[i].name, DEFAULT_PARAM_NAMES[i]);
        }
        new_transition.action.make_symbolic();
        
        // only add predicates if prev state is not empty (i.e. if we can obtain any rule)
        if (!new_transition.prev_state.empty()) {
            result.add_transition(new_transition);
        }
    }
    
    return result;
}

void TransitionGroup::add_transition(const Transition& transition)
{
    // Add new objects (if there are any new)
    std::vector<PPDDL::Object> transition_objects = transition.get_prev_state().get_params();
    ppddl_object_manager.add_objects(transition_objects.begin(), transition_objects.end());
    transition_objects = transition.get_post_state().get_params();
    ppddl_object_manager.add_objects(transition_objects.begin(), transition_objects.end());
    
    // Add transition
    std::pair<TransitionSet::iterator, bool> pair_setit_bool;
    pair_setit_bool = transition_set.insert(transition);
    if (!pair_setit_bool.second) { // transition already existed
        pair_setit_bool.first->increase_number_repetitions();
    }
    transition_historial.push_back(pair_setit_bool.first);
}

void TransitionGroup::add_transitions(const TransitionGroup& transitions)
{
    for(TransitionHistorial::const_iterator transition_it_it = transitions.begin_historial(); transition_it_it != transitions.end_historial(); ++transition_it_it) {
        add_transition(**transition_it_it);
    }
}

const Transition& TransitionGroup::get_transition_from_set(const int index) const
{
    TransitionSet::iterator transition_it = transition_set.begin();
    std::advance(transition_it, index);
    return *transition_it;
}

TransitionGroup TransitionGroup::get_symbolic_transitions(const int max_action_variables) const
{
    TransitionGroup result(ppddl_object_manager);
    for(TransitionHistorial::const_iterator transition_it_it = begin_historial(); transition_it_it != end_historial(); ++transition_it_it) {
        result.add_transitions( (*transition_it_it)->get_symbolic_transitions(ppddl_object_manager, max_action_variables) );
    }
    return result;
}

void TransitionGroup::remove_negated_predicates()
{
    TransitionSet original_transition_set;
    original_transition_set.swap(transition_set);
    TransitionHistorial original_transition_historial;
    original_transition_historial.swap(transition_historial);
    for (TransitionHistorial::const_iterator transition_it_it = original_transition_historial.begin(); 
         transition_it_it != original_transition_historial.end(); 
         ++transition_it_it) 
    {
        Transition new_transition(**transition_it_it);
        new_transition.remove_negated_predicates();
        add_transition(new_transition);
    }
}

void TransitionGroup::name_grounding()
{
    TransitionSet original_transition_set;
    original_transition_set.swap(transition_set);
    TransitionHistorial original_transition_historial;
    original_transition_historial.swap(transition_historial);
    for(TransitionHistorial::const_iterator transition_it_it = original_transition_historial.begin(); 
        transition_it_it != original_transition_historial.end(); 
        ++transition_it_it) {
        Transition name_grounded_transition(**transition_it_it);
        name_grounded_transition.name_grounding();
        add_transition(name_grounded_transition);
    }
}

void Transition::name_grounding()
{
    prev_state.PPDDL_grounding();
    next_state.PPDDL_grounding();
    action.PPDDL_grounding();
}

