/*
 * Clases representing transitions (previous experiences).
 * Copyright (C) 2014  David Martínez (dmartinez@iri.upc.edu)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#ifndef IRI_RULE_LEARNER_TRANSITIONS_H
#define IRI_RULE_LEARNER_TRANSITIONS_H 1

// #define DEBUG_COVERAGE_ERRORS

#include <vector>
#include <fstream>
#include <algorithm>    // std::min

#include <symbolic/symbols.h>
#include <symbolic/predicates.h>

namespace Transitions {
struct Coverage {
  private:
    typedef size_t size_type;
    typedef size_t index_type;
    
    std::vector< std::vector<float> > coverage_vector;
    
#ifdef DEBUG_COVERAGE_ERRORS
    std::vector<float> coverage_errors_vector;
#else
    float coverage_error;
#endif
    bool is_empty;
    
  public:
    Coverage() : is_empty(true) {}
    
    explicit Coverage(const PredicateGroup<TypedAtom>& changes):
        coverage_vector(changes.size()),
#ifndef DEBUG_COVERAGE_ERRORS
        coverage_error(0.0),
#endif
        is_empty(true)
    { }
    
    size_type size() const { return coverage_vector.size(); }

#ifdef DEBUG_COVERAGE_ERRORS
    void add_error(const float new_error_prob) {coverage_errors_vector.push_back(new_error_prob); is_empty = false; }
#else
    void add_error(const float new_error_prob) {coverage_error += (1.0 - coverage_error) * new_error_prob; is_empty = false; }
#endif

    void add_change_coverage(const index_type change_idx, const float new_coverage) 
        { coverage_vector[change_idx].push_back(new_coverage); is_empty = false; };    
    float get_error_probability() const;
    float get_change_coverage(const index_type change_idx, const float optimistic_value = 0.0) const;
    float get_coverage(const bool complete, const float optimistic_value = 0.0) const;
    float get_likelihood(const bool complete, const float optimistic_value = 0.0) const;
    
    Coverage& operator+=(const Coverage& rhs);
    friend Coverage operator+(const Coverage &c1, const Coverage &c2);
    friend std::ostream& operator<<(std::ostream &out, const Coverage &c);
};
}

// required by Transition method template
class TransitionGroup;

/*!
\class Transition
\brief Class representing previous experiences of the form: state_t, action_t, state_t+1, obtained reward
*/
class Transition
{
private:
    PredicateGroup<TypedAtom> prev_state, next_state;
    TypedAtom action;
    float reward;
    PredicateGroup<TypedAtom> changes;
    mutable int number_repetitions; //! Number of times that the transition has appeared
    
public:
    Transition(const Transition& transition) :
        prev_state(transition.prev_state),
        next_state(transition.next_state),
        action(transition.action),
        reward(transition.reward),
        changes(transition.changes),
        number_repetitions(transition.number_repetitions)
    { }

    Transition(const PredicateGroup<TypedAtom>& prev_state, 
               const TypedAtom& action, 
               const PredicateGroup<TypedAtom>& next_state, 
               const float& reward, 
               const bool& variable_maintains_to_next_state) : // True if as default variable's value maintains in the next state | False if it resets to false
        prev_state(prev_state),
        next_state(next_state),
        action(action),
        reward(reward),
        number_repetitions(1)
    {
        update_changes(variable_maintains_to_next_state);
    }

    // get info
    const PredicateGroup<TypedAtom>& get_prev_state() const { return prev_state; }
    const PredicateGroup<TypedAtom>& get_post_state() const { return next_state; }
    const TypedAtom& get_action() const { return action; }
    const float& get_reward() const { return reward; }
    const PredicateGroup<TypedAtom>& get_changes() const { return changes; };
    
    void increase_number_repetitions() const { number_repetitions++; /* is mutable */ }
    int get_number_repetitions() const { return number_repetitions; }
    
    void update_changes(const bool& variable_maintains_to_next_state);
    
    /**
     * \brief Get symbolic transitions
     * 
     * Gets the symbolic transitions, taking as many symbolic variables as action params
     * When a predicate is modified, related variables are grouped into the same symbolic transition,
     * otherwise a new transition is generated for every combination of params.
     * \n
     * When using action params, the action is maintained.
     * When using other params, the noop action is set.
     */
    TransitionGroup get_symbolic_transitions(const PPDDLObjectManager& object_manager, const int max_action_variables) const;

    void name_grounding();

    bool nothing_changed() const;

    /**
     * \brief Removes predicates whose parameters are not in the action parameters
     */
    void remove_unused_predicates();
    
    /**
     * \brief As we are using closed-world assumption, negated predicates don't have to be specified directly
     */
    void remove_negated_predicates();

    friend std::ostream& operator<<(std::ostream &out, const Transition &s);
    friend bool operator<(const Transition &l, const Transition &r);
};

typedef std::vector<Transition> TransitionVector;
typedef std::set<Transition> TransitionSet;

/*!
\class TransitionGroup
\brief Vector of transitions.

Its transitions can be accessed in two different ways:
 - Historial: All the transitions ordered by time.
 - Set: Set of non-repeated transitions (no time ordering).
*/
class TransitionGroup
{
public:
    typedef std::vector<TransitionSet::iterator> TransitionHistorial;
    
    typedef Transition& reference;
    typedef const Transition& const_reference;
    
    PPDDLObjectManager ppddl_object_manager;
    
    explicit TransitionGroup(const PPDDLObjectManager& object_manager) : ppddl_object_manager(object_manager) {};
    
//     TransitionGroup(const TransitionGroup& transition_group):
//         TransitionList(transition_group),
//         ppddl_object_manager(transition_group.ppddl_object_manager) 
//         {};
    
    void name_grounding();
    
    void add_transition(const Transition& transition);
    void add_transitions(const TransitionGroup& transitions);
    
    size_t size_set() const { return transition_set.size(); }
    size_t size_historial() const { return transition_historial.size(); }
    
    const Transition& get_transition_from_set(const int index) const;
    
    /*
    reference operator[] (size_t n);
    const_reference operator[] (size_t n) const;*/
    
    TransitionHistorial::iterator begin_historial() { return transition_historial.begin(); };
    TransitionHistorial::const_iterator begin_historial() const { return transition_historial.begin(); };
    TransitionHistorial::iterator end_historial() { return transition_historial.end(); };
    TransitionHistorial::const_iterator end_historial() const { return transition_historial.end(); };
    
    TransitionSet::iterator begin_set() { return transition_set.begin(); };
    TransitionSet::const_iterator begin_set() const { return transition_set.begin(); };
    TransitionSet::iterator end_set() { return transition_set.end(); };
    TransitionSet::const_iterator end_set() const { return transition_set.end(); };
    
    /**
     * \brief Write transitions to file with all the important information
     */
    void write_transitions_to_file(const std::string& file_path) const;
    
    /**
     * \brief Write transitions to file in a format compatible with Pasula learner
     */
    void write_pasula_transitions_to_file(const std::string& transitions_filename) const;
    
    /**
     * \brief Write transitions to file in a format compatible with LFIT learner
     */
    void write_lfit_transitions_to_file(const std::string& file_path, 
                                        const std::string& domain_types, 
                                        const std::vector<NonTypedAtom>& list_of_actions) const;

    /**
     * \brief Get transitions that include any of the actions in the given list.
     * 
     * \param actions list of actions whose transitions are returned
     * \return list of transitions with the given actions
     */
    TransitionGroup get_action_list_transitions(std::vector<TypedAtom> actions) const;
    
    /**
     * \brief Get transitions that include the given action (symbolic comparison done to check if the action is the same as the transitiona action).
     * 
     * \param action Action (TypedAtom) whose transitions are returned
     * \return list of transitions with the given action name
     */
    TransitionGroup get_action_transitions(const TypedAtom& action) const;
        
    /**
     * \brief Get symbolic transitions
     * 
     * Gets the symbolic transitions for each transition, taking as many symbolic variables as action params
     * When a predicate is modified, related variables are grouped into the same symbolic transition,
     * otherwise a new transition is generated for every combination of params.
     * \n
     * When using action params, the action is maintained.
     * When using other params, the noop action is set.
     */
    TransitionGroup get_symbolic_transitions(const int max_action_variables) const;
    
    /**
     * \brief As we are using closed-world assumption, negated predicates don't have to be specified directly
     */
    void remove_negated_predicates();

    friend std::ostream& operator<<(std::ostream &out, const TransitionGroup &s);
    
private:
    TransitionSet transition_set;
    TransitionHistorial transition_historial;
};

typedef boost::shared_ptr<TransitionGroup> TransitionGroupPtr;
typedef boost::shared_ptr<const TransitionGroup> TransitionGroupConstPtr;

// other


// write/read to file
TransitionGroup read_transitions_from_file(const std::string& file_path, const PPDDLObjectManager& object_manager, const PredicateGroup<TypedAtom>& constant_preds, const bool& variable_maintains_to_next_state);

#endif
